﻿/* CubicSpline.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.Numerics
{
  /// <summary>三次スプライン補間処理クラス</summary>
  public static class CubicSpline
  {

    /// <summary>三次スプライン補間のための係数を計算する</summary>
    /// <param name="x">X</param>
    /// <param name="y">Y</param>
    /// <returns>三次スプライン補間のための係数</returns>
    public static double[] GetParameters(double[] x, double[] y)
    {
      double[] h = new double[y.Length - 1];
      double[] a = new double[y.Length - 2];
      double[,] hm = new double[3, y.Length - 2];
      for (int i = 0; i < y.Length - 1; i++) h[i] = x[i + 1] - x[i];
      for (int i = 0; i < y.Length - 2; i++)
      {
        if (i != 0) hm[0, i] = h[i];
        hm[1, i] = 2 * (h[i] + h[i + 1]);
        if (i != y.Length - 3) hm[2, i] = h[i + 1];
        a[i] = 3 * ((y[i + 2] - y[i + 1]) / h[i + 1] - (y[i + 1] - y[i]) / h[i]);
      }
      LinearAlgebra.SolveTridiagonalMatrix(hm, ref a);
      double[] c = new double[y.Length];
      a.CopyTo(c, 1);
      c[0] = c[c.Length - 1] = 0;
      return c;
    }

    /// <summary>補間処理を行う</summary>
    /// <param name="x">X</param>
    /// <param name="y">Y</param>
    /// <param name="c">係数</param>
    /// <param name="x2">補完処理を行う位置</param>
    /// <returns>補間値</returns>
    /// <remarks>x2が順序よく並ぶことを想定した処理</remarks>
    public static double[] Interpolate(double[] x, double[] y, double[] c, double[] x2)
    {
      double[] y2 = new double[x2.Length];
      int num = 0;
      for (int i = 0; i < x2.Length; i++)
      {
        //範囲確認
        if (x2[i] < x[0] || x[x.Length - 1] < x2[i])
          throw new Exception("CubicSpline: out of range");
        //補間範囲を特定して補間処理実行
        while (x[num + 1] < x2[i]) num++;
        y2[i] = interPolate(x, y, c, x2[i], num);
      }
      return y2;
    }

    /// <summary>補間処理を行う</summary>
    /// <param name="x">X</param>
    /// <param name="y">Y</param>
    /// <param name="c">係数</param>
    /// <param name="x2">補完処理を行う位置</param>
    /// <returns>補間値</returns>
    public static double Interpolate(double[] x, double[] y, double[] c, double x2)
    {
      int low = 0;
      int high = x.Length - 1;
      //二分法で補間範囲を求める
      while (1 < high - low)
      {
        int mid = (low + high) >> 1;
        if (x2 < x[mid]) high = mid;
        else low = mid;
      }
      return interPolate(x, y, c, x2, low);
    }

    /// <summary>x2の位置での補間値を計算する</summary>
    /// <param name="x">X</param>
    /// <param name="y">Y</param>
    /// <param name="cf">係数</param>
    /// <param name="x2">補間を行う位置</param>
    /// <param name="num">x2の補間式の番号</param>
    /// <returns>補間値</returns>
    private static double interPolate(double[] x, double[] y, double[] cf, double x2, int num)
    {
      double dx = x2 - x[num];
      double h = x[num + 1] - x[num];
      double a = y[num];
      double b = (y[num + 1] - y[num]) / h - h * (cf[num + 1] + 2 * cf[num]) / 3d;
      double c = cf[num];
      double d = (cf[num + 1] - cf[num]) / (3d * h);
      return a + dx * (b + dx * (c + dx * d));
    }

  }
}
