﻿/* LinearAlgebra.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.Numerics
{
  /// <summary>
  /// 線形代数処理クラス
  /// 参考文献:
  /// 小国力　新数値計算法
  /// 北川源四郎　時系列解析入門
  /// ニューメリカルレシピ in C
  /// </summary>
  public static class LinearAlgebra
  {

    #region LU分解

    /// <summary>
    /// Crout法によりLU分解を行う
    /// 入力配列であるmatrixにはLU分解済の係数が記録される
    /// </summary>
    /// <param name="matrix">N×Nの行列</param>
    /// <param name="permutationVector">pivot選択による行置換リスト</param>
    /// <remarks>Newmerical Recipiesより移植</remarks>
    public static void LUDecomposition
        (double[,] matrix, int[] permutationVector)
    {
      const double TINY = 1.0e-20;
      //行列の行数・列数を取得
      int number = matrix.GetLength(0);

      //スケーリングを記憶する配列
      double[] vv = new double[number];

      for (int i = 0; i < number; i++)
      {
        double big = 0.0;
        for (uint j = 0; j < number; j++)
        {
          big = Math.Max(big, Math.Abs(matrix[i, j]));
        }
        if (big == 0.0) throw new Exception("特異行列です");
        vv[i] = 1.0 / big;
      }

      for (int j = 0; j < number; j++)
      {
        double sum = 0;
        double big = 0.0d;
        int imax = 0;

        //Crout法を適用
        //i~jまでの繰り返し計算
        for (int i = 0; i < j; i++)
        {
          sum = matrix[i, j];
          for (uint k = 0; k < i; k++)
            sum -= matrix[i, k] * matrix[k, j];
          matrix[i, j] = sum;
        }

        //j~Nまでの繰り返し計算
        for (int i = j; i < number; i++)
        {
          sum = matrix[i, j];
          for (uint k = 0; k < j; k++)
            sum -= matrix[i, k] * matrix[k, j];
          matrix[i, j] = sum;

          //スケーリングを考慮して最大のpivot選択を行う
          double dum = vv[i] * Math.Abs(sum);
          if (dum >= big)
          {
            big = dum;
            imax = i;
          }
        }

        //行交換の必要判定
        if (j != imax)
        {
          //行の交換
          for (uint k = 0; k < number; k++)
          {
            double dum = matrix[imax, k];
            matrix[imax, k] = matrix[j, k];
            matrix[j, k] = dum;
          }
          vv[imax] = vv[j];   //スケーリング係数の交換
        }
        //行の置換を記憶
        permutationVector[j] = imax;

        if (matrix[j, j] == 0.0) matrix[j, j] = TINY;

        if (j != number)
        {
          for (int i = j + 1; i < number; i++)
            matrix[i, j] = matrix[i, j] / matrix[j, j];
        }
      }
    }

    /// <summary>LU行列に基き前進・後退代入処理を行う</summary>
    /// <param name="luMatrix">LU分解済の行列</param>
    /// <param name="permutationVector">置換ベクトル</param>
    /// <param name="b">bベクトル：解が上書きされる</param>
    public static void ForwardAndBackSubstitution
        (double[,] luMatrix, int[] permutationVector, double[] b)
    {
      //行列の行数・列数を取得
      int number = luMatrix.GetLength(0);

      //ベクトルbが0以外の数値をとる位置
      int ii = 0;

      for (int i = 0; i < number; i++)
      {
        //置換ベクトルに従ってbベクトルを入替え
        int ip = permutationVector[i];
        double sum = b[ip];
        b[ip] = b[i];

        //前進代入処理
        if (ii != 0)
        {
          for (int j = ii - 1; j < i; j++)
            sum -= luMatrix[i, j] * b[j];
        }
        else if (sum != 0) ii = i + 1;
        b[i] = sum;
      }
      //後退代入処理
      for (int i = number - 1; 0 <= i; i--)
      {
        double sum = b[i];
        for (int j = i + 1; j < number; j++)
          sum -= luMatrix[i, j] * b[j];
        b[i] = sum / luMatrix[i, i];
      }
    }

    /// <summary>mAの逆行列を計算する</summary>
    /// <param name="mA">逆行列を求める行列</param>
    /// <param name="mB">出力:逆行列</param>
    public static void GetInverse(double[,] mA, ref double[,] mB)
    {
      int[] perm = new int[mA.GetLength(0)];
      double[] col = new double[mA.GetLength(0)];
      LUDecomposition(mA, perm);
      for (int i = 0; i < mA.GetLength(0); i++)
      {
        for (int j = 0; j < col.Length; j++) col[j] = 0;
        col[i] = 1;
        ForwardAndBackSubstitution(mA, perm, col);
        for (int j = 0; j < col.Length; j++) mB[j, i] = col[j];
      }
    }

    /// <summary>mAの逆行列を計算する</summary>
    /// <param name="mA">逆行列を求める行列</param>
    /// <param name="mB">出力:逆行列</param>
    public static void GetInverse(MatrixView mA, ref MatrixView mB)
    {
      double[,] mAView = mA.GetView();
      double[,] mBView = mB.GetView();
      GetInverse(mAView, ref mBView);
      mB.SetView(mBView);
    }

    #endregion

    #region 帯行列関連

    /// <summary>
    /// Thomas algorithmで三重対角行列連立一次方程式を解く
    /// abc(0,i)*nx(i-1)+abc(1,i)*nx(i)+abc(2,i)*nx(i+1)=x(i)
    /// </summary>
    /// <param name="abc">係数行列</param>
    /// <param name="x">解で上書きされる</param>
    public static void SolveTridiagonalMatrix
        (double[,] abc, ref double[] x)
    {
      //行列の数を取得
      int num = abc.GetLength(1) - 1;

      abc[2, 0] /= abc[1, 0];
      x[0] /= abc[1, 0];

      for (int i = 1; i < num; i++)
      {
        abc[2, i] /= abc[1, i] - abc[0, i] * abc[2, i - 1];
        x[i] = (x[i] - abc[0, i] * x[i - 1]) /
            (abc[1, i] - abc[0, i] * abc[2, i - 1]);
      }

      x[num] = (x[num] - abc[0, num] * x[num - 1]) /
          (abc[1, num] - abc[0, num] * abc[2, num - 1]);

      for (int i = num - 1; 0 <= i; i--)
      {
        x[i] -= abc[2, i] * x[i + 1];
      }
    }

    #endregion

    #region 最小二乗法

    /// <summary>最小二乗法で回帰係数を計算する</summary>
    /// <param name="y">目的変数ベクトル</param>
    /// <param name="x">説明変数行列</param>
    /// <returns>回帰係数ベクトル</returns>
    public static double[] LeastSquareFit
      (double[] y, double[,] x)
    {
      double sig, aic;
      return LeastSquareFit(y, x, out sig, out aic);
    }

    /// <summary>最小二乗法で回帰係数を計算する</summary>
    /// <param name="y">目的変数ベクトル</param>
    /// <param name="x">説明変数行列</param>
    /// <param name="sigma2">残差分散σ^2</param>
    /// <param name="aic">赤池情報量規準</param>
    /// <returns>回帰係数ベクトル</returns>
    public static double[] LeastSquareFit
      (double[] y, double[,] x, out double sigma2, out double aic)
    {
      int col = y.Length;
      int row = x.GetLength(1);

      double[,] s = new double[col, row + 1];
      for (int i = 0; i < col; i++)
      {
        s[i, row] = y[i];
        for (int j = 0; j < row; j++)
          s[i, j] = x[i, j];
      }

      double[,] mQ;
      HouseholderTransform(ref s, out mQ);
      //残差分散
      sigma2 = s[row, row] * s[row, row] / col;

      double[] a = new double[row];
      for (int i = row - 1; 0 <= i; i--)
      {
        double ss = s[i, row];
        for (int j = row - 1; i < j; j--)
        {
          ss -= s[i, j] * a[j];
        }
        a[i] = ss / s[i, i];
      }

      //赤池情報量規準
      aic = col * (Math.Log(2 * Math.PI * sigma2) + 1) + 2 * (row + 1);
      return a;
    }

    /// <summary>ハウスホルダ変換により上三角行列を作成する</summary>
    /// <param name="mA">変換対象行列(上書きされる)</param>
    /// <param name="mQ">ハウスホルダ変換行列</param>
    public static void HouseholderTransform
      (ref double[,] mA, out double[,] mQ)
    {
      int n = mA.GetLength(0);
      int m = mA.GetLength(1);
      double[,] qi = new double[n, n];
      mQ = new double[n, n];
      for (int i = 0; i < n; i++) mQ[i, i] = 1;
      double[] wi = new double[n];

      for (int i = 0; i < m; i++)//DEBUG
      {
        double sig2 = 0;
        for (int j = 0; j < i; j++) wi[j] = 0;
        for (int j = i; j < n; j++)
        {
          sig2 += mA[j, i] * mA[j, i];
          wi[j] = mA[j, i];
        }
        double sig = Math.Sqrt(sig2);
        double sn = Math.Sign(wi[i]);
        wi[i] += sn * sig;

        double alpha = 1d / (sig2 + sn * sig * mA[i, i]);

        for (int j = 0; j < n; j++)
        {
          for (int k = 0; k < n; k++)
          {
            qi[j, k] = -alpha * wi[j] * wi[k];
            if (j == k) qi[j, k] += 1;
          }
        }
        Multiplicate(qi, mQ, out mQ);
        Multiplicate(qi, mA, out mA);
      }
    }

    #endregion

    #region 行列・ベクトル演算

    /// <summary>行列の積(C=AB)を計算する</summary>
    /// <param name="mA">A行列</param>
    /// <param name="mB">B行列</param>
    /// <param name="mC">C行列</param>
    public static void Multiplicate
      (double[,] mA, double[,] mB, out double[,] mC)
    {
      int m = mA.GetLength(0);
      int n = mA.GetLength(1);
      int p = mB.GetLength(1);
      mC = new double[m, p];

      for (int i = 0; i < m; i++)
        for (int j = 0; j < p; j++)
          for (int k = 0; k < n; k++)
            mC[i, j] += mA[i, k] * mB[k, j];
    }

    /// <summary>行列とベクトルの積和を計算する（vC = α mA vB + β vC）</summary>
    /// <param name="mA">行列A</param>
    /// <param name="vB">ベクトルB</param>
    /// <param name="vC">ベクトルC（解が上書きされる）</param>
    /// <param name="alpha">第一項の係数</param>
    /// <param name="beta">第二項の係数</param>
    public static void Multiplicate
      (double[,] mA, double[] vB, ref double[] vC, double alpha, double beta)
    {
      int m = mA.GetLength(0);
      int n = mA.GetLength(1);
      for (uint i = 0; i < m; i++)
      {
        vC[i] *= beta;
        for (uint j = 0; j < n; j++)
          vC[i] += alpha * mA[i, j] * vB[j];
      }
    }

    /// <summary>行列とベクトルの積和を計算する（vC = α mA vB + β vC）</summary>
    /// <param name="mA">行列A</param>
    /// <param name="vB">ベクトルB</param>
    /// <param name="vC">ベクトルC（解が上書きされる）</param>
    /// <param name="alpha">第一項の係数</param>
    /// <param name="beta">第二項の係数</param>
    public static void Multiplicate
      (MatrixView mA, VectorView vB, ref VectorView vC, double alpha, double beta)
    {
      double[,] mAView = mA.GetView();
      double[] vBView = vB.GetView();
      double[] vCView = vC.GetView();
      Multiplicate(mAView, vBView, ref vCView, alpha, beta);
      vC.SetView(vCView);
    }

    #endregion

    #region 部分行列・部分ベクトル

    /// <summary>部分行列クラス</summary>
    public class MatrixView
    {

      /// <summary>要素の値を設定・取得する</summary>
      /// <param name="column">行番号</param>
      /// <param name="row">列番号</param>
      /// <returns>要素の値</returns>
      public double this[int column, int row]
      {
        get { return originalMatrix[columnStartNumber + column, rowStartNumber + row]; }
        set { originalMatrix[columnStartNumber + column, rowStartNumber + row] = value; }
      }

      /// <summary>行数を取得する</summary>
      public int Column { private set; get; }

      /// <summary>列数を取得する</summary>
      public int Row { private set; get; }

      /// <summary>元の行列</summary>
      private double[,] originalMatrix;

      /// <summary>列の開始番号</summary>
      private int rowStartNumber;

      /// <summary>行の開始番号</summary>
      private int columnStartNumber;
      
      /// <summary>コンストラクタ</summary>
      /// <param name="rowSize">列数</param>
      /// <param name="columnSize">行数</param>
      /// <param name="rowStartNumber">列の開始番号</param>
      /// <param name="columnStartNumber">行の開始番号</param>
      /// <param name="originalMatrix">元の行列</param>
      public MatrixView(int rowSize, int columnSize,
        int rowStartNumber, int columnStartNumber, double[,] originalMatrix)
      {
        Column = columnSize;
        Row = rowSize;
        this.originalMatrix = originalMatrix;
        this.rowStartNumber = rowStartNumber;
        this.columnStartNumber = columnStartNumber;
      }

      /// <summary>部分マトリクスを取得する</summary>
      /// <returns>部分マトリクス</returns>
      public double[,] GetView()
      {
        double[,] view = new double[Column, Row];
        for (int i = 0; i < Column; i++)
          for (int j = 0; j < Row; j++)
            view[i, j] = originalMatrix[columnStartNumber + i, rowStartNumber + j];
        return view;
      }

      /// <summary>部分マトリクスを設定する</summary>
      /// <param name="view">部分マトリクス</param>
      public void SetView(double[,] view)
      {
        for (int i = 0; i < Column; i++)
          for (int j = 0; j < Row; j++)
            originalMatrix[columnStartNumber + i, rowStartNumber + j] = view[i, j];
      }
    }

    /// <summary>部分ベクトルクラス</summary>
    public class VectorView
    {

      /// <summary>要素の値を設定・取得する</summary>
      /// <param name="index">要素番号</param>
      /// <returns>要素の値</returns>
      public double this[int index]
      {
        get { return originalVector[startNumber + index]; }
        set { originalVector[startNumber + index] = value; }
      }
      
      /// <summary>ベクトル長を取得する</summary>
      public int Length { private set; get; }

      /// <summary>元のベクトル</summary>
      private double[] originalVector;

      /// <summary>部分ベクトル開始番号</summary>
      private int startNumber;

      /// <summary>コンストラクタ</summary>
      /// <param name="length">ベクトル長</param>
      /// <param name="startNumber">部分ベクトル開始番号</param>
      /// <param name="originalVector">元のベクトル</param>
      public VectorView(int length, int startNumber, double[] originalVector)
      {
        Length = length;
        this.originalVector = originalVector;
        this.startNumber = startNumber;
      }

      /// <summary>部分ベクトルを取得する</summary>
      /// <returns>部分ベクトル</returns>
      public double[] GetView()
      {
        double[] view = new double[Length];
        for (int i = 0; i < Length; i++)view[i] = originalVector[startNumber + i];
        return view;
      }

      /// <summary>部分ベクトルを設定する</summary>
      /// <param name="view">部分ベクトル</param>
      public void SetView(double[] view)
      {
        for (int i = 0; i < Length; i++)originalVector[startNumber + i] = view[i];
      }

    }

    #endregion

  }
}
