/* MoistAir.cs
 * 
 * Copyright (C) 2007 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;
using System.Runtime.Serialization;

namespace Popolo.ThermophysicalProperty
{

  /// <summary>湿り空気クラス</summary>
  /// <remarks>
  /// HVACSIM+(J)
  /// 宇田川光弘:パソコンによる空気調和計算法
  /// ASHRAE Fundamentals Handbook 1997 Psychrometrics
  /// </remarks>
  [Serializable]
  public class MoistAir : ImmutableMoistAir
  {

    #region Constants

    /// <summary>シリアライズ用バージョン情報</summary>
    private double S_VERSION = 1.3;

    /// <summary>絶対温度と摂氏との変換定数</summary>
    private const double CONVERT_C_TO_K = 273.15;

    /// <summary>乾き空気の定圧比熱[kJ/kg-K]</summary>
    private const double DRYAIR_ISOBARIC_SPECIFIC_HEAT = 1.005;

    /// <summary>水蒸気の定圧比熱[kJ/kg-K]</summary>
    private const double VAPOR_ISOBARIC_SPECIFIC_HEAT = 1.805;

    /// <summary>0℃の水の定圧比熱[kJ/kg-K]</summary>
    private const double WATER_ISOBARIC_SPECIFIC_HEAT = 4.186;

    /// <summary>0℃の水の蒸発潜熱[kJ/kg]</summary>
    private const double VAPORIZATION_LATENT_HEAT = 2501;

    /// <summary>海抜0mでの大気圧=101.325[kPa]</summary>
    private const double ATMOSPHERIC_PRESSURE_AT_SEALEVEL = 101.325;

    /// <summary>乾き空気のガス定数[kJ/(kg K)]</summary>
    private const double DRYAIR_GAS_CONSTANT = 0.287042;

    #endregion

    #region enumerators

    /// <summary>湿り空気物性種類</summary>
    public enum Property
    {
      /// <summary>乾球温度[CDB]</summary>
      DryBulbTemperature = 0,
      /// <summary>湿球温度[CWB]</summary>
      WetBulbTemperature = 1,
      /// <summary>絶対湿度[kg/kg]</summary>
      HumidityRatio = 2,
      /// <summary>相対湿度[%]</summary>
      RelativeHumidity = 3,
      /// <summary>エンタルピー[kJ/kg]</summary>
      Enthalpy = 4,
      /// <summary>水蒸気分圧[kPa]</summary>
      WaterPartialPressure = 5,
      /// <summary>比容積[m3/kg]</summary>
      SpecificVolume = 6,
      /// <summary>飽和温度[C]</summary>
      SaturatedTemperature = 7
    }

    #endregion
    
    #region Properties

    /// <summary>乾き空気の比熱[kJ/kg-K]を取得する</summary>
    public static double IsobaricSpecificHeatOfDryAir
    {
      get { return DRYAIR_ISOBARIC_SPECIFIC_HEAT; }
    }

    /// <summary>蒸気の比熱[kJ/kg-K]を取得する</summary>
    public static double IsobaricSpecificHeatOfVapor
    {
      get { return VAPOR_ISOBARIC_SPECIFIC_HEAT; }
    }

    /// <summary>水の蒸発潜熱[kJ/kg]を取得する</summary>
    public static double LatentHeatOfVaporization
    {
      get { return VAPORIZATION_LATENT_HEAT; }
    }

    /// <summary>乾球温度[C]を設定・取得する</summary>
    public double DrybulbTemperature { get; set; }

    /// <summary>湿球温度[C]を設定・取得する</summary>
    public double WetbulbTemperature { get; set; }

    /// <summary>絶対湿度[kg/kg(DA)]を設定・取得する</summary>
    public double HumidityRatio { get; set; }

    /// <summary>相対湿度[%]を設定・取得する</summary>
    public double RelativeHumidity { get; set; }

    /// <summary>エンタルピー[kJ/kg]を設定・取得する</summary>
    public double Enthalpy { get; set; }

    /// <summary>比容積[m3/kg]を設定・取得する</summary>
    public double SpecificVolume { get; set; }

    /// <summary>大気圧[kPa]を設定・取得する</summary>
    public double AtmosphericPressure { get; set; }
    
    #endregion

    #region Constructor

    /// <summary>コンストラクタ</summary>
    public MoistAir()
    {
      //24C/50%で初期化
      this.AtmosphericPressure = ATMOSPHERIC_PRESSURE_AT_SEALEVEL;
      this.DrybulbTemperature = 24;
      this.HumidityRatio = 0.0093;
      this.RelativeHumidity = 50;
      this.Enthalpy = 47.8;
      this.WetbulbTemperature = 17.1;
      this.SpecificVolume = 0.85;
    }

    /// <summary>コンストラクタ</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    public MoistAir(double drybulbTemperature, double humidityRatio)
    {
      this.AtmosphericPressure = ATMOSPHERIC_PRESSURE_AT_SEALEVEL;
      this.DrybulbTemperature = drybulbTemperature;
      this.HumidityRatio = humidityRatio;
      this.RelativeHumidity = GetRelativeHumidityFromDryBulbTemperatureAndHumidityRatio
        (drybulbTemperature, humidityRatio, ATMOSPHERIC_PRESSURE_AT_SEALEVEL);
      this.Enthalpy = GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (drybulbTemperature, humidityRatio);
      this.WetbulbTemperature = GetWetBulbTemperatureFromDryBulbTemperatureAndHumidityRatio
        (drybulbTemperature, humidityRatio, ATMOSPHERIC_PRESSURE_AT_SEALEVEL);
      this.SpecificVolume = GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (drybulbTemperature, humidityRatio, ATMOSPHERIC_PRESSURE_AT_SEALEVEL);
    }

    /// <summary>コピーコンストラクタ</summary>
    /// <param name="moistAir">コピーする湿り空気状態</param>
    public MoistAir(ImmutableMoistAir moistAir)
    {
      this.AtmosphericPressure = moistAir.AtmosphericPressure;
      this.DrybulbTemperature = moistAir.DrybulbTemperature;
      this.HumidityRatio = moistAir.HumidityRatio;
      this.RelativeHumidity = moistAir.RelativeHumidity;
      this.Enthalpy = moistAir.Enthalpy;
      this.WetbulbTemperature = moistAir.WetbulbTemperature;
      this.SpecificVolume = moistAir.SpecificVolume;
    }

    #endregion

    #region プログラム1

    /// <summary>絶対湿度[kg/kg]と大気圧[kPa]から水蒸気分圧[kPa]を求める</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]：1気圧は101.325[kPa]</param>
    /// <returns>水蒸気分圧[kPa]</returns>
    public static double GetWaterVaporPartialPressureFromHumidityRatio
        (double humidityRatio, double atmosphericPressure)
    {
      return atmosphericPressure * humidityRatio / (0.62198 + humidityRatio);
    }

    /// <summary>水蒸気分圧[kPa]と大気圧[kPa]から絶対湿度[kg/kg]を求める</summary>
    /// <param name="waterVaporPartialPressure">水蒸気分圧[kPa]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]：1気圧は101.325[kPa]</param>
    /// <returns>絶対湿度[kg/kg]</returns>
    public static double GetHumidityRatioFromWaterVaporPartialPressure
        (double waterVaporPartialPressure, double atmosphericPressure)
    {
      return 0.62198 * waterVaporPartialPressure /
          (atmosphericPressure - waterVaporPartialPressure);
    }

    #endregion

    #region プログラム2

    /// <summary>乾球温度[℃]と絶対湿度[kg/kg]からエンタルピー[kJ/kg]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[℃]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <returns>エンタルピー[kJ/kg]</returns>
    public static double GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (double dryBulbTemperature, double humidityRatio)
    {
      return DRYAIR_ISOBARIC_SPECIFIC_HEAT * dryBulbTemperature +
          humidityRatio * (VAPOR_ISOBARIC_SPECIFIC_HEAT * dryBulbTemperature +
          VAPORIZATION_LATENT_HEAT);
    }

    /// <summary>乾球温度[℃]とエンタルピー[kJ/kg]から絶対湿度[kg/kg]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[℃]</param>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <returns>絶対湿度[kg/kg]</returns>
    public static double GetHumidityRatioFromDryBulbTemperatureAndEnthalpy
        (double dryBulbTemperature, double enthalpy)
    {
      return (enthalpy - DRYAIR_ISOBARIC_SPECIFIC_HEAT * dryBulbTemperature) /
          (VAPOR_ISOBARIC_SPECIFIC_HEAT * dryBulbTemperature +
          VAPORIZATION_LATENT_HEAT);
    }

    /// <summary>絶対湿度[kg/kg]とエンタルピー[kJ/kg]から乾球温度[℃]を求める</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <returns>乾球温度[℃]</returns>
    public static double GetDryBulbTemperatureFromHumidityRatioAndEnthalpy
        (double humidityRatio, double enthalpy)
    {
      return (enthalpy - VAPORIZATION_LATENT_HEAT * humidityRatio) /
          (DRYAIR_ISOBARIC_SPECIFIC_HEAT +
          VAPOR_ISOBARIC_SPECIFIC_HEAT * humidityRatio);
    }

    #endregion

    #region プログラム3

    /// <summary>乾球温度[℃]と湿球温度[℃]から絶対湿度[kg/kg]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[℃]</param>
    /// <param name="wetBulbTemperature">湿球温度[℃]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]：1気圧は101.325[kPa]</param>
    /// <returns>絶対湿度[kg/kg]</returns>
    public static double GetHumidityRatioFromDryBulbTemperatureAndWetBulbTemperature
        (double dryBulbTemperature, double wetBulbTemperature, double atmosphericPressure)
    {
      double ps = Water.GetSaturationPressure(wetBulbTemperature);
      double ws = GetHumidityRatioFromWaterVaporPartialPressure(ps, atmosphericPressure);
      double a = ws * (VAPORIZATION_LATENT_HEAT +
          (VAPOR_ISOBARIC_SPECIFIC_HEAT - WATER_ISOBARIC_SPECIFIC_HEAT)
          * wetBulbTemperature) + DRYAIR_ISOBARIC_SPECIFIC_HEAT *
          (wetBulbTemperature - dryBulbTemperature);
      double b = VAPORIZATION_LATENT_HEAT + VAPOR_ISOBARIC_SPECIFIC_HEAT *
          dryBulbTemperature - WATER_ISOBARIC_SPECIFIC_HEAT *
          wetBulbTemperature;
      return a / b;
    }

    /// <summary>絶対湿度[kg/kg]と湿球温度[C]から乾球温度[C]を計算する</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>乾球温度[C]</returns>
    public static double GetDryBulbTemperatureFromHumidityRatioAndWetBulbTemperature
        (double humidityRatio, double wetBulbTemperature, double atmosphericPressure)
    {
      double ps = Water.GetSaturationPressure(wetBulbTemperature);
      double ws = GetHumidityRatioFromWaterVaporPartialPressure(ps, atmosphericPressure);
      double a = ws * (VAPORIZATION_LATENT_HEAT +
          (VAPOR_ISOBARIC_SPECIFIC_HEAT - WATER_ISOBARIC_SPECIFIC_HEAT)
          * wetBulbTemperature);
      a += DRYAIR_ISOBARIC_SPECIFIC_HEAT * wetBulbTemperature;
      a += (WATER_ISOBARIC_SPECIFIC_HEAT * wetBulbTemperature -
          VAPORIZATION_LATENT_HEAT) * humidityRatio;
      double b = VAPOR_ISOBARIC_SPECIFIC_HEAT * humidityRatio +
          DRYAIR_ISOBARIC_SPECIFIC_HEAT;
      return a / b;
    }

    /// <summary>乾球温度[℃]と絶対湿度[kg/kg]から湿球温度[℃]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[℃]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]：1気圧は101.325[kPa]</param>
    /// <returns>湿球温度[℃]</returns>
    public static double GetWetBulbTemperatureFromDryBulbTemperatureAndHumidityRatio
        (double dryBulbTemperature, double humidityRatio, double atmosphericPressure)
    {
      const double DELTA = 0.00001;
      const double ERR_TOL = 0.0001;
      const int MAX_ITERATION = 20;

      //湿球温度を仮定して絶対湿度を計算する
      double wbt = dryBulbTemperature;
      double hrt = GetHumidityRatioFromDryBulbTemperatureAndWetBulbTemperature
          (dryBulbTemperature, wbt, atmosphericPressure);
      //誤差を評価
      double err1 = humidityRatio - hrt;

      //ニュートン法で反復計算
      int iterNum = 0;    //反復回数
      while (ERR_TOL < Math.Abs(err1) && iterNum < MAX_ITERATION)
      {
        //誤差関数の数値微分を計算
        hrt = GetHumidityRatioFromDryBulbTemperatureAndWetBulbTemperature
            (dryBulbTemperature, wbt + DELTA, atmosphericPressure);
        double err2 = humidityRatio - hrt;
        double errDx = (err2 - err1) / DELTA;

        //湿球温度を更新
        wbt -= err1 / errDx;

        //更新された湿球温度で誤差を再評価
        hrt = GetHumidityRatioFromDryBulbTemperatureAndWetBulbTemperature
        (dryBulbTemperature, wbt, atmosphericPressure);
        err1 = humidityRatio - hrt;
      }

      return wbt;
    }

    #endregion

    #region プログラム4

    /// <summary>エンタルピー[kJ/kg]と湿球温度[C]から乾球温度[C]を計算する</summary>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>乾球温度[C]</returns>
    public static double GetDryBulbTemperatureFromWetBulbTemperatureAndEnthalpy
        (double wetBulbTemperature, double enthalpy, double atmosphericPressure)
    {
      const double DELTA = 0.00001;
      const double ERR_TOL = 0.0001;
      const int MAX_ITERATION = 20;

      //乾球温度を仮定して絶対湿度を計算する
      double dbt = wetBulbTemperature;
      double hrt = GetHumidityRatioFromDryBulbTemperatureAndEnthalpy(dbt, enthalpy);
      //誤差を評価
      double err1 = dbt - GetDryBulbTemperatureFromHumidityRatioAndWetBulbTemperature
          (hrt, wetBulbTemperature, atmosphericPressure);

      //ニュートン法で反復計算
      int iterNum = 0;    //反復回数
      while (ERR_TOL < Math.Abs(err1) && iterNum < MAX_ITERATION)
      {
        //誤差関数の数値微分を計算
        hrt = GetHumidityRatioFromDryBulbTemperatureAndEnthalpy(dbt + DELTA, enthalpy);
        double err2 = (dbt + DELTA) - GetDryBulbTemperatureFromHumidityRatioAndWetBulbTemperature
            (hrt, wetBulbTemperature, atmosphericPressure);
        double errDx = (err2 - err1) / DELTA;

        //乾球温度を更新
        dbt -= err1 / errDx;

        //更新された乾球温度で誤差を再評価
        hrt = GetHumidityRatioFromDryBulbTemperatureAndEnthalpy(dbt, enthalpy);
        err1 = dbt - GetDryBulbTemperatureFromHumidityRatioAndWetBulbTemperature
            (hrt, wetBulbTemperature, atmosphericPressure);
      }

      return dbt;
    }

    #endregion

    #region プログラム5

    /// <summary>エンタルピー[kJ/kg]と湿球温度[C]から絶対湿度[kg/kg]を計算する</summary>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>絶対湿度[kg/kg]</returns>
    public static double GetHumidityRatioFromWetBulbTemperatureAndEnthalpy
        (double wetBulbTemperature, double enthalpy, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromWetBulbTemperatureAndEnthalpy
          (wetBulbTemperature, enthalpy, atmosphericPressure);
      return GetHumidityRatioFromDryBulbTemperatureAndEnthalpy(dbt, enthalpy);
    }

    /// <summary>乾球温度[C]とエンタルピー[kJ/kg]から湿球温度[C]を計算する</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>湿球温度[C]</returns>
    public static double GetWetBulbTemperatureFromDryBulbTemperatureAndEnthalpy
        (double dryBulbTemperature, double enthalpy, double atmosphericPressure)
    {
      double hrt = GetHumidityRatioFromDryBulbTemperatureAndEnthalpy
          (dryBulbTemperature, enthalpy);
      return GetWetBulbTemperatureFromDryBulbTemperatureAndHumidityRatio
          (dryBulbTemperature, hrt, atmosphericPressure);
    }

    /// <summary>絶対湿度[kg/kg]とエンタルピー[kJ/kg]から湿球温度[C]を計算する</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>湿球温度[C]</returns>
    public static double GetWetBulbTemperatureFromHumidityRatioAndEnthalpy
        (double humidityRatio, double enthalpy, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromHumidityRatioAndEnthalpy
          (humidityRatio, enthalpy);
      return GetWetBulbTemperatureFromDryBulbTemperatureAndHumidityRatio
          (dbt, humidityRatio, atmosphericPressure);
    }

    /// <summary>乾球温度[C]と湿球温度[C]からエンタルピー[kJ/kg]を計算する</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>エンタルピー[kJ/kg]</returns>
    public static double GetEnthalpyFromDryBulbTemperatureAndWetBulbTemperature
        (double dryBulbTemperature, double wetBulbTemperature, double atmosphericPressure)
    {
      double hrt = GetHumidityRatioFromDryBulbTemperatureAndWetBulbTemperature
          (dryBulbTemperature, wetBulbTemperature, atmosphericPressure);
      return GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
          (dryBulbTemperature, hrt);
    }

    /// <summary>絶対湿度[kg/kg]と湿球温度[C]からエンタルピー[kJ/kg]を計算する</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>エンタルピー[kJ/kg]</returns>
    public static double GetEnthalpyFromHumidityRatioAndWetBulbTemperature
        (double humidityRatio, double wetBulbTemperature, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromHumidityRatioAndWetBulbTemperature
          (humidityRatio, wetBulbTemperature, atmosphericPressure);
      return GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
          (dbt, humidityRatio);
    }

    #endregion

    #region プログラム6

    /// <summary>乾球温度[C]と相対湿度[%]から絶対湿度[kg/kg]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>絶対湿度[kg/kg]</returns>
    public static double GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
        (double dryBulbTemperature, double relativeHumidity, double atmosphericPressure)
    {
      //飽和水蒸気分圧[kPa]の計算
      double ps = Water.GetSaturationPressure(dryBulbTemperature);
      //水蒸気分圧[kPa]の計算
      double pw = 0.01 * relativeHumidity * ps;
      return GetHumidityRatioFromWaterVaporPartialPressure
          (pw, atmosphericPressure);
    }

    /// <summary>乾球温度[C]と絶対湿度[kg/kg]から相対湿度[%]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>相対湿度[%]</returns>
    public static double GetRelativeHumidityFromDryBulbTemperatureAndHumidityRatio
        (double dryBulbTemperature, double humidityRatio, double atmosphericPressure)
    {
      //水蒸気分圧[kPa]の計算
      double pw = GetWaterVaporPartialPressureFromHumidityRatio
          (humidityRatio, atmosphericPressure);
      //飽和水蒸気圧[kPa]の計算
      double ps = Water.GetSaturationPressure(dryBulbTemperature);
      if (ps <= 0.0) return 0.0;
      else return 100.0 * pw / ps;
    }

    /// <summary>絶対湿度[kg/kg]と相対湿度[%]から乾球温度[C]を求める</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>乾球温度[C]</returns>
    public static double GetDryBulbTemperatureFromHumidityRatioAndRelativeHumidity
        (double humidityRatio, double relativeHumidity, double atmosphericPressure)
    {
      double ps = GetWaterVaporPartialPressureFromHumidityRatio
          (humidityRatio, atmosphericPressure) / relativeHumidity * 100;
      return Water.GetSaturationTemperature(ps);
    }

    #endregion

    #region プログラム7

    /// <summary>乾球温度[C]と湿球温度[C]から相対湿度[%]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>相対湿度[%]</returns>
    public static double GetRelativeHumidityFromDryBulbTemperatureAndWetBulbTemperature
        (double dryBulbTemperature, double wetBulbTemperature, double atmosphericPressure)
    {
      double hrt = GetHumidityRatioFromDryBulbTemperatureAndWetBulbTemperature
          (dryBulbTemperature, wetBulbTemperature, atmosphericPressure);
      return GetRelativeHumidityFromDryBulbTemperatureAndHumidityRatio
          (dryBulbTemperature, hrt, atmosphericPressure);
    }

    /// <summary>湿球温度[C]と絶対湿度[kg/kg]から相対湿度[%]を求める</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>相対湿度[%]</returns>
    public static double GetRelativeHumidityFromWetBulbTemperatureAndHumidityRatio
        (double humidityRatio, double wetBulbTemperature, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromHumidityRatioAndWetBulbTemperature
          (humidityRatio, wetBulbTemperature, atmosphericPressure);
      return GetRelativeHumidityFromDryBulbTemperatureAndWetBulbTemperature
          (dbt, wetBulbTemperature, atmosphericPressure);
    }

    /// <summary>乾球温度[C]とエンタルピー[kJ/kg]から相対湿度[%]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>相対湿度[%]</returns>
    public static double GetRelativeHumidityFromDryBulbTemperatureAndEnthalpy
        (double dryBulbTemperature, double enthalpy, double atmosphericPressure)
    {
      double hrt = GetHumidityRatioFromDryBulbTemperatureAndEnthalpy
          (dryBulbTemperature, enthalpy);
      return GetRelativeHumidityFromDryBulbTemperatureAndHumidityRatio
          (dryBulbTemperature, hrt, atmosphericPressure);
    }

    /// <summary>絶対湿度[kg/kg]とエンタルピー[kJ/kg]から相対湿度[%]を求める</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>相対湿度[%]</returns>
    public static double GetRelativeHumidityFromHumidityRatioAndEnthalpy
        (double humidityRatio, double enthalpy, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromHumidityRatioAndEnthalpy
          (humidityRatio, enthalpy);
      return GetRelativeHumidityFromDryBulbTemperatureAndHumidityRatio
          (dbt, humidityRatio, atmosphericPressure);
    }

    /// <summary>乾球温度[C]と相対湿度[%]から湿球温度[C]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>湿球温度[C]</returns>
    public static double GetWetBulbTemperatureFromDryBulbTemperatureAndRelativeHumidity
        (double dryBulbTemperature, double relativeHumidity, double atmosphericPressure)
    {
      double hrt = GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (dryBulbTemperature, relativeHumidity, atmosphericPressure);
      return GetWetBulbTemperatureFromDryBulbTemperatureAndHumidityRatio
          (dryBulbTemperature, hrt, atmosphericPressure);
    }

    /// <summary>絶対湿度[kg/kg]と相対湿度[%]から湿球温度[C]を求める</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>湿球温度[C]</returns>
    public static double GetWetBulbTemperatureFromHumidityRatioAndRelativeHumidity
        (double humidityRatio, double relativeHumidity, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromHumidityRatioAndRelativeHumidity
          (humidityRatio, relativeHumidity, atmosphericPressure);
      return GetWetBulbTemperatureFromDryBulbTemperatureAndHumidityRatio
          (dbt, humidityRatio, atmosphericPressure);
    }

    /// <summary>乾球温度[C]と相対湿度[%]からエンタルピー[kJ/kg]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>エンタルピー[kJ/kg]</returns>
    public static double GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
        (double dryBulbTemperature, double relativeHumidity, double atmosphericPressure)
    {
      double hrt = GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (dryBulbTemperature, relativeHumidity, atmosphericPressure);
      return GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
          (dryBulbTemperature, hrt);
    }

    /// <summary>絶対湿度[kg/kg]と相対湿度[%]からエンタルピー[kJ/kg]を求める</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>エンタルピー[kJ/kg]</returns>
    public static double GetEnthalpyFromHumidityRatioAndRelativeHumidity
        (double humidityRatio, double relativeHumidity, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromHumidityRatioAndRelativeHumidity
          (humidityRatio, relativeHumidity, atmosphericPressure);
      return GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
          (dbt, humidityRatio);
    }


    #endregion

    #region プログラム8

    /// <summary>湿球温度[C]と相対湿度[%]から乾球温度[C]を計算する</summary>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>乾球温度[C]</returns>
    public static double GetDryBulbTemperatureFromWetBulbTemperatureAndRelativeHumidity
        (double wetBulbTemperature, double relativeHumidity, double atmosphericPressure)
    {
      const double DELTA = 0.00001;
      const double ERR_TOL = 0.0001;
      const int MAX_ITERATION = 20;

      //乾球温度を仮定して絶対湿度を計算する
      double dbt = wetBulbTemperature;
      double hrt = GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (dbt, relativeHumidity, atmosphericPressure);
      //誤差を評価
      double err1 = dbt - GetDryBulbTemperatureFromHumidityRatioAndWetBulbTemperature
          (hrt, wetBulbTemperature, atmosphericPressure);

      //ニュートン法で反復計算
      int iterNum = 0;    //反復回数
      while (ERR_TOL < Math.Abs(err1) && iterNum < MAX_ITERATION)
      {
        //誤差関数の数値微分を計算
        hrt = GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
            (dbt + DELTA, relativeHumidity, atmosphericPressure);
        double err2 = (dbt + DELTA) - GetDryBulbTemperatureFromHumidityRatioAndWetBulbTemperature
            (hrt, wetBulbTemperature, atmosphericPressure);
        double errDx = (err2 - err1) / DELTA;

        //乾球温度を更新
        dbt -= err1 / errDx;

        //更新された乾球温度で誤差を再評価
        hrt = GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
            (dbt, relativeHumidity, atmosphericPressure);
        err1 = dbt - GetDryBulbTemperatureFromHumidityRatioAndWetBulbTemperature
        (hrt, wetBulbTemperature, atmosphericPressure);
      }

      return dbt;
    }

    /// <summary>湿球温度[C]と相対湿度[%]から絶対湿度[kg/kg]を計算する</summary>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>絶対湿度[kg/kg]</returns>
    public static double GetHumidityRatioFromWetBulbTemperatureAndRelativeHumidity
        (double wetBulbTemperature, double relativeHumidity, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromWetBulbTemperatureAndRelativeHumidity
          (wetBulbTemperature, relativeHumidity, atmosphericPressure);
      return GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (dbt, relativeHumidity, atmosphericPressure);
    }

    /// <summary>湿球温度[C]と相対湿度[%]から比エンタルピー[kJ/kg]を計算する</summary>
    /// <param name="wetBulbTemperature">湿球温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>比エンタルピー[kJ/kg]</returns>
    public static double GetEnthalpyFromWetBulbTemperatureAndRelativeHumidity
        (double wetBulbTemperature, double relativeHumidity, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromWetBulbTemperatureAndRelativeHumidity
          (wetBulbTemperature, relativeHumidity, atmosphericPressure);
      return GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
          (dbt, relativeHumidity, atmosphericPressure);
    }

    #endregion

    #region プログラム9

    /// <summary>比エンタルピー[kJ/kg]と相対湿度[%]から乾球温度[C]を計算する</summary>
    /// <param name="enthalpy">比エンタルピー[kJ/kg]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>乾球温度[C]</returns>
    private static double GetDryBulbTemperatureFromEnthalpyAndRelativeHumidity
        (double enthalpy, double relativeHumidity, double atmosphericPressure)
    {
      const double DELTA = 0.00001;
      const double ERR_TOL = 0.0001;
      const int MAX_ITERATION = 20;

      //乾球温度を仮定して比エンタルピーを計算する
      double dbt = 25;
      double ent = GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
          (dbt, relativeHumidity, atmosphericPressure);
      //誤差を評価
      double err1 = enthalpy - ent;

      //ニュートン法で反復計算
      int iterNum = 0;    //反復回数
      while (ERR_TOL < Math.Abs(err1) && iterNum < MAX_ITERATION)
      {
        //誤差関数の数値微分を計算
        ent = GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
            (dbt + DELTA, relativeHumidity, atmosphericPressure);
        double err2 = enthalpy - ent;
        double errDx = (err2 - err1) / DELTA;

        //乾球温度を更新
        dbt -= err1 / errDx;

        //更新された乾球温度で誤差を再評価
        ent = GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
            (dbt, relativeHumidity, atmosphericPressure);
        err1 = enthalpy - ent;
      }

      return dbt;
    }

    /// <summary>比エンタルピー[kJ/kg]と相対湿度[%]から絶対湿度[kg/kg]を計算する</summary>
    /// <param name="enthalpy">比エンタルピー[kJ/kg]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>絶対湿度[kg/kg]</returns>
    public static double GetHumidityRatioFromEnthalpyAndRelativeHumidity
        (double enthalpy, double relativeHumidity, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromEnthalpyAndRelativeHumidity
          (enthalpy, relativeHumidity, atmosphericPressure);
      return GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
          (dbt, relativeHumidity, atmosphericPressure);
    }

    /// <summary>比エンタルピー[kJ/kg]と相対湿度[%]から湿球温度[C]を計算する</summary>
    /// <param name="enthalpy">比エンタルピー[kJ/kg]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>湿球温度[C]</returns>
    public static double GetWetBulbTemperatureFromEnthalpyAndRelativeHumidity
        (double enthalpy, double relativeHumidity, double atmosphericPressure)
    {
      double dbt = GetDryBulbTemperatureFromEnthalpyAndRelativeHumidity
          (enthalpy, relativeHumidity, atmosphericPressure);
      return GetWetBulbTemperatureFromDryBulbTemperatureAndRelativeHumidity
          (dbt, relativeHumidity, atmosphericPressure);
    }

    #endregion

    #region プログラム10

    /// <summary>相対湿度[%]および比体積[m3/kg]から乾球温度[C]を求める</summary>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="specificVolume">比体積[m3/kg]</param>
    /// <param name="atm">大気圧[kPa]</param>
    /// <returns>乾球温度[C]</returns>
    public static double GetDryBulbTemperatureFromRelativeHumidityAndSpecificVolume
        (double relativeHumidity, double specificVolume, double atm)
    {
      //収束計算
      const double DELTA = 1.0e-10;
      const double TOL = 1.0e-9;
      int iterNum = 0;
      double dbt = 25;
      while (true)
      {
        double err1 = GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity(dbt, relativeHumidity, atm) - GetHumidityRatioFromDryBulbTemperatureAndSpecificVolume(dbt, specificVolume, atm);
        if (Math.Abs(err1) < TOL) break;
        double err2 = GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity(dbt + DELTA, relativeHumidity, atm) - GetHumidityRatioFromDryBulbTemperatureAndSpecificVolume(dbt + DELTA, specificVolume, atm);
        dbt -= err1 / ((err2 - err1) / DELTA);
        iterNum++;
        if (20 < iterNum) throw new Exception("Iteration Error");
      }
      return dbt;
    }

    /// <summary>湿球温度[C]および比体積[m3/kg]から乾球温度[C]を求める</summary>
    /// <param name="wetbulbTemperature">湿球温度[C]</param>
    /// <param name="specificVolume">比体積[m3/kg]</param>
    /// <param name="atm">大気圧[kPa]</param>
    /// <returns>乾球温度[C]</returns>
    public static double GetDryBulbTemperatureFromWetBulbTemperatureAndSpecificVolume
        (double wetbulbTemperature, double specificVolume, double atm)
    {
      //収束計算
      const double DELTA = 1.0e-10;
      const double TOL = 1.0e-9;
      int iterNum = 0;
      double dbt = 25;
      while (true)
      {
        double err1 = GetHumidityRatioFromDryBulbTemperatureAndWetBulbTemperature(dbt, wetbulbTemperature, atm) - GetHumidityRatioFromDryBulbTemperatureAndSpecificVolume(dbt, specificVolume, atm);
        if (Math.Abs(err1) < TOL) break;
        double err2 = GetHumidityRatioFromDryBulbTemperatureAndWetBulbTemperature(dbt + DELTA, wetbulbTemperature, atm) - GetHumidityRatioFromDryBulbTemperatureAndSpecificVolume(dbt + DELTA, specificVolume, atm);
        dbt -= err1 / ((err2 - err1) / DELTA);
        iterNum++;
        if (20 < iterNum) throw new Exception("Iteration Error");
      }
      return dbt;
    }

    /// <summary>乾球温度[C]および絶対湿度[kg/kg]から比体積[m3/kg]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>比体積[m3/kg]</returns>
    public static double GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
        (double dryBulbTemperature, double humidityRatio, double atmosphericPressure)
    {
      return ((dryBulbTemperature + CONVERT_C_TO_K) * DRYAIR_GAS_CONSTANT) /
          atmosphericPressure * (1.0 + 1.6078 * humidityRatio);
    }

    /// <summary>比体積[m3/kg]および絶対湿度[kg/kg]から乾球温度[C]を求める</summary>
    /// <param name="specificVolume">比体積[m3/kg]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>乾球温度[C]</returns>
    public static double GetDryBulbTemperatureFromSpecificVolumeAndHumidityRatio
        (double specificVolume, double humidityRatio, double atmosphericPressure)
    {
      return specificVolume / (1.0 + 1.6078 * humidityRatio) *
          atmosphericPressure / DRYAIR_GAS_CONSTANT - CONVERT_C_TO_K;
    }

    /// <summary>乾球温度[C]および比体積[m3/kg]から絶対湿度[kg/kg]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[C]</param>
    /// <param name="specificVolume">比体積[m3/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>絶対湿度[kg/kg]</returns>
    public static double GetHumidityRatioFromDryBulbTemperatureAndSpecificVolume
        (double dryBulbTemperature, double specificVolume, double atmosphericPressure)
    {
      return (specificVolume / (dryBulbTemperature + CONVERT_C_TO_K) /
          DRYAIR_GAS_CONSTANT * atmosphericPressure - 1.0) / 1.6078;
    }

    #endregion

    #region その他の物性値

    /// <summary>標高[m]に応じた大気圧[kPa]を取得する</summary>
    /// <param name="altitude">標高[m]</param>
    /// <returns>大気圧[kPa]</returns>
    public static double GetAtmosphericPressure(double altitude)
    {
      return ATMOSPHERIC_PRESSURE_AT_SEALEVEL * Math.Pow(1.0 - 2.25577e-5 * altitude, 5.2559);
    }

    /// <summary>湿り空気比熱[kJ/kg-K]を計算する</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg(DA)]</param>
    /// <returns>湿り空気比熱[kJ/kg-K]</returns>
    public static double GetSpecificHeat(double humidityRatio)
    {
      return DRYAIR_ISOBARIC_SPECIFIC_HEAT + VAPOR_ISOBARIC_SPECIFIC_HEAT * humidityRatio;
    }

    /// <summary>粘性係数[Pa s]を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <returns>粘性係数[Pa s]</returns>
    public static double GetViscosity(double drybulbTemperature)
    {
      return (0.0074237 / (drybulbTemperature + 390.15)) *
          Math.Pow((drybulbTemperature + CONVERT_C_TO_K) / 293.15, 1.5);
    }

    /// <summary>動粘性係数[m2/s]を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>動粘性係数[m2/s]</returns>
    public static double GetDynamicViscosity
        (double drybulbTemperature, double humidityRatio, double atmosphericPressure)
    {
      return GetViscosity(drybulbTemperature) *
          GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
          (drybulbTemperature, humidityRatio, atmosphericPressure);
    }

    /// <summary>熱伝導率[W/(mK)]を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <returns>熱伝導率[W/(mK)]</returns>
    public static double GetThermalConductivity(double drybulbTemperature)
    {
      return 0.0241 + 0.000077 * drybulbTemperature;
    }

    /// <summary>膨張率[1/K]を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <returns>膨張率[1/K]</returns>
    public static double GetExpansionCoefficient(double drybulbTemperature)
    {
      return 1 / (drybulbTemperature + CONVERT_C_TO_K);
    }

    /// <summary>空気中における蒸気の拡散係数[m2/s]を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="pressure">圧力[kPa]</param>
    /// <returns>空気中における蒸気の拡散係数[m2/s]</returns>
    public static double GetDiffusionCoefficient
        (double drybulbTemperature, double pressure)
    {
      const double cf = 1.894775e-8;
      return cf / pressure *
          Math.Exp(2.072 * Math.Log(drybulbTemperature + CONVERT_C_TO_K));
    }

    /// <summary>熱拡散率[m2/s]を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>熱拡散率[m2/s]</returns>
    public static double GetThermalDiffusivity
        (double drybulbTemperature, double humidityRatio, double atmosphericPressure)
    {
      double lambda = GetThermalConductivity(drybulbTemperature);
      double cp = GetSpecificHeat(humidityRatio);
      double rho = GetSpecificVolumeFromDryBulbTemperatureAndHumidityRatio
          (drybulbTemperature, humidityRatio, atmosphericPressure);
      return lambda / (1000 * cp * rho);
    }

    #endregion

    #region 飽和状態関連

    /// <summary>絶対湿度[kg/kg]から飽和乾球温度（露点温度）[℃]を求める</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>飽和乾球温度（露点温度）[℃]</returns>
    public static double GetSaturationDryBulbTemperatureFromHumidityRatio
      (double humidityRatio, double atmosphericPressure)
    {
      double ps = GetWaterVaporPartialPressureFromHumidityRatio
        (humidityRatio, atmosphericPressure);
      return Water.GetSaturationTemperature(ps);
    }

    /// <summary>エンタルピー[kJ/kg]から飽和乾球温度（露点温度）[℃]を求める</summary>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>飽和乾球温度（露点温度）[℃]</returns>
    public static double GetSaturationDryBulbTemperatureFromEnthalpy
      (double enthalpy, double atmosphericPressure)
    {
      return GetDryBulbTemperatureFromEnthalpyAndRelativeHumidity
        (enthalpy, 100, atmosphericPressure);
    }

    /// <summary>乾球温度[℃]から飽和絶対湿度[kg/kg]を求める</summary>
    /// <param name="dryBulbTemperature">乾球温度[℃]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>飽和絶対湿度[kg/kg]</returns>
    public static double GetSaturationHumidityRatioFromDryBulbTemperature
      (double dryBulbTemperature, double atmosphericPressure)
    {
      double ps = Water.GetSaturationPressure(dryBulbTemperature);
      return GetHumidityRatioFromWaterVaporPartialPressure(ps, atmosphericPressure);
    }

    /// <summary>エンタルピー[kJ/kg]から飽和絶対湿度[kg/kg]を求める</summary>
    /// <param name="enthalpy">エンタルピー[kJ/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>飽和絶対湿度[kg/kg]</returns>
    public static double GetSaturationHumidityRatioFromEnthalpy
      (double enthalpy, double atmosphericPressure)
    {
      return GetHumidityRatioFromEnthalpyAndRelativeHumidity
        (enthalpy, 100, atmosphericPressure);
    }

    /// <summary>乾球温度[℃]から飽和エンタルピー[kJ/kg]を求める</summary>
    /// <param name="drybulbTemperature">飽和温度[℃]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>飽和エンタルピー[kJ/kg]</returns>
    public static double GetSaturationEnthalpyFromDrybulbTemperature
      (double drybulbTemperature, double atmosphericPressure)
    {
      double ps = Water.GetSaturationPressure(drybulbTemperature);
      double ws = GetHumidityRatioFromWaterVaporPartialPressure
        (ps, atmosphericPressure);
      return GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (drybulbTemperature, ws);
    }

    /// <summary>絶対湿度[kg/kg]から飽和エンタルピー[kJ/kg]を求める</summary>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmosphericPressure">大気圧[kPa]</param>
    /// <returns>飽和エンタルピー[kJ/kg]</returns>
    public static double GetSaturationEnthalpyFromHuidityRatio
      (double humidityRatio, double atmosphericPressure)
    {
      double ts = GetSaturationDryBulbTemperatureFromHumidityRatio
        (humidityRatio, atmosphericPressure);
      return GetEnthalpyFromDryBulbTemperatureAndHumidityRatio
        (ts, humidityRatio);
    }

    #endregion

    #region 空気混合処理

    /// <summary>空気を混合する</summary>
    /// <param name="air">混合する空気の配列</param>
    /// <param name="volume">混合空気の体積</param>
    /// <returns>混合空気</returns>
    public static MoistAir BlendAir(ImmutableMoistAir[] air, double[] volume)
    {
      int airNum = air.Length;
      if (airNum != volume.Length) throw new Exception
        ("MoistAir Class: BlendAir: 混合する空気の数と割合の数が一致しません");

      double rSum = 0.0d;
      double tSum = 0.0d;
      double trSum = 0.0d;
      double hSum = 0.0d;
      double hrSum = 0.0d;
      for (int i = 0; i < airNum; i++)
      {
        //割合を積算
        volume[i] = Math.Max(0, volume[i]);
        rSum += volume[i];
        //温度[K]を積算
        tSum += air[i].DrybulbTemperature;
        //温度[K]×割合を積算
        trSum += air[i].DrybulbTemperature * volume[i];
        //絶対湿度[kg/kg]を積算
        hSum += air[i].HumidityRatio;
        //絶対湿度[kg/kg]×割合を積算
        hrSum += air[i].HumidityRatio * volume[i];
      }
      //混合後の乾球温度および絶対湿度を計算
      double drybulbTempOut;
      double absHumidOut;
      if (rSum >= 1.0e-5d)
      {
        drybulbTempOut = trSum / rSum;
        absHumidOut = hrSum / rSum;
      }
      //割合の積算が小さい場合は発散を防ぐために混合空気の数で割る
      else
      {
        drybulbTempOut = tSum / airNum;
        absHumidOut = hSum / airNum;
      }
      //出口空気状態を計算**飽和した場合の処理が必要**
      return new MoistAir(drybulbTempOut, absHumidOut);
    }

    /// <summary>空気を混合する</summary>
    /// <param name="air1">混合空気1</param>
    /// <param name="air2">混合空気2</param>
    /// <param name="air1Volume">空気1の体積</param>
    /// <param name="air2Volume">空気2の体積</param>
    /// <returns>混合空気</returns>
    public static MoistAir BlendAir
      (ImmutableMoistAir air1, ImmutableMoistAir air2, double air1Volume, double air2Volume)
    {
      if (air1Volume <= 0) return new MoistAir
        (air2.DrybulbTemperature, air2.HumidityRatio);
      else if (air2Volume <= 0) return new MoistAir
        (air1.DrybulbTemperature, air1.HumidityRatio);
      else
      {
        double vSum = air1Volume + air2Volume;
        double dbt = (air1.DrybulbTemperature * air1Volume + air2.DrybulbTemperature * air2Volume) / vSum;
        double ahd = (air1.HumidityRatio * air1Volume + air2.HumidityRatio * air2Volume) / vSum;
        return new MoistAir(dbt, ahd);
      }
    }

    #endregion

    #region 空気状態コピー処理

    /// <summary>空気状態をコピーする</summary>
    /// <param name="air">コピー先の空気</param>
    public void CopyTo(MoistAir air)
    {
      air.AtmosphericPressure = this.AtmosphericPressure;
      air.DrybulbTemperature = this.DrybulbTemperature;
      air.WetbulbTemperature = this.WetbulbTemperature;
      air.HumidityRatio = this.HumidityRatio;
      air.RelativeHumidity = this.RelativeHumidity;
      air.Enthalpy = this.Enthalpy;
      air.SpecificVolume = this.SpecificVolume;
    }

    #endregion

    #region ISerializableインターフェース実装

    /// <summary>デシリアライズ用Constructor</summary>
    /// <param name="sInfo"></param>
    /// <param name="context"></param>
    protected MoistAir(SerializationInfo sInfo, StreamingContext context)
    {
      //バージョン情報
      double version = sInfo.GetDouble("S_Version");

      //乾球温度[℃]
      DrybulbTemperature = sInfo.GetDouble("drybulbTemperature");
      //湿球温度[℃]
      WetbulbTemperature = sInfo.GetDouble("wetbulbTemperature");
      //絶対湿度[kg/kg]
      HumidityRatio = sInfo.GetDouble("humidityRatio");
      //相対湿度[%]
      RelativeHumidity = sInfo.GetDouble("relativeHumidity");
      //エンタルピー[kJ/kg]
      Enthalpy = sInfo.GetDouble("enthalpy");
      //比容積[m3/kg]
      if (1.1 <= version) SpecificVolume = sInfo.GetDouble("specificVolume");
      //大気圧[kPa]
      if (1.2 <= version) AtmosphericPressure = sInfo.GetDouble("atmosphericPressure");
    }

    /// <summary>MoistAirシリアル化処理</summary>
    /// <param name="info"></param>
    /// <param name="context"></param>
    public virtual void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      //バージョン情報
      info.AddValue("S_Version", S_VERSION);

      //乾球温度[℃]
      info.AddValue("drybulbTemperature", DrybulbTemperature);
      //湿球温度[℃]
      info.AddValue("wetbulbTemperature", WetbulbTemperature);
      //絶対湿度[kg/kg]
      info.AddValue("humidityRatio", HumidityRatio);
      //相対湿度[%]
      info.AddValue("relativeHumidity", RelativeHumidity);
      //エンタルピー[kJ/kg]
      info.AddValue("enthalpy", Enthalpy);
      //比容積[m3/kg]
      info.AddValue("specificVolume", SpecificVolume);
      //大気圧[kPa]
      info.AddValue("atmosphericPressure", AtmosphericPressure);
    }

    #endregion

  }

  #region ImmutableMoistAir

  /// <summary>読み取り専用湿り空気</summary>
  public interface ImmutableMoistAir : ISerializable
  {
    /// <summary>大気圧[kPa]を取得する</summary>
    double AtmosphericPressure { get; }

    /// <summary>乾球温度[C]を取得する</summary>
    double DrybulbTemperature { get; }

    /// <summary>湿球温度[C]を取得する</summary>
    double WetbulbTemperature { get; }

    /// <summary>絶対湿度[kg/kg(DA)]を取得する</summary>
    double HumidityRatio { get; }

    /// <summary>相対湿度[%]を取得する</summary>
    double RelativeHumidity { get; }

    /// <summary>エンタルピー[kJ/kg]を取得する</summary>
    double Enthalpy { get; }

    /// <summary>比容積[m3/kg]を取得する</summary>
    double SpecificVolume { get; }

    /// <summary>空気状態をコピーする</summary>
    /// <param name="air">コピー先の湿り空気オブジェクト</param>
    void CopyTo(MoistAir air);
  }

  #endregion
  
}
