﻿/* NormalRandom.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.Numerics
{
  /// <summary>正規分布に従う乱数系列を作成するクラス</summary>
  public class NormalRandom
  {

    #region インスタンス変数

    /// <summary>一様乱数製造機</summary>
    private MersenneTwister rnd;

    /// <summary>乱数ストック</summary>
    private double rndStock;

    /// <summary>乱数ストックを持つか否か</summary>
    private bool hasStock = false;

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタ</summary>
    /// <param name="seed">乱数シード</param>
    public NormalRandom(int seed)
    {
      rnd = new MersenneTwister(seed);
    }

    #endregion

    #region メソッド

    /// <summary>平均0分散1に従う正規乱数系列を取得する</summary>
    /// <returns>平均0分散1に従う正規乱数</returns>
    public double NextDouble()
    {
      if (hasStock)
      {
        hasStock = false;
        return rndStock;
      }
      else
      {
        hasStock = true;
        double rtn;
        makeNormalRandomNumbers(rnd, out rtn, out rndStock);
        return rtn;
      }
    }

    /// <summary>BoxMuller法で平均0分散1の正規乱数を計算する</summary>
    /// <param name="rnd">一様乱数製造機</param>
    /// <param name="nrnd1">正規乱数1</param>
    /// <param name="nrnd2">正規乱数2</param>
    private static void makeNormalRandomNumbers
      (MersenneTwister rnd, out double nrnd1, out double nrnd2)
    {
      double v = rnd.NextDouble() * 2 - 1;
      double u = rnd.NextDouble() * 2 - 1;
      double r2 = v * v + u * u;
      while (1 < r2)
      {
        v = rnd.NextDouble() * 2 - 1;
        u = rnd.NextDouble() * 2 - 1;
        r2 = v * v + u * u;
      }
      double w = Math.Sqrt(-2.0 * Math.Log(r2) / r2);
      nrnd1 = w * v;
      nrnd2 = w * u;
    }

    #endregion

  }
}
