﻿/* RandomWeather.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;
using Popolo.ThermophysicalProperty;

namespace Popolo.Weather
{
  /// <summary>確率的気象モデル</summary>
  public static class RandomWeather
  {

    #region 列挙型定義

    /// <summary>地点</summary>
    public enum Location
    {
      /// <summary>東京</summary>
      Tokyo,
      /// <summary>大阪</summary>
      Osaka,
      /// <summary>福岡</summary>
      Fukuoka,
      /// <summary>札幌</summary>
      Sapporo,
      /// <summary>仙台</summary>
      Sendai,
      /// <summary>那覇</summary>
      Naha
    }

    #endregion

    #region 変数

    /// <summary>計算地点の太陽</summary>
    private static Sun sun;

    /// <summary>トレンド成分標準偏差</summary>
    private static double sdevTDT, sdevTHR, sdevTAT;

    /// <summary>ARモデル係数</summary>
    private static double[] iDTCof, iHRCof;

    /// <summary>ARモデル係数</summary>
    private static double iATCof, iDTSD, iHRSD, iATSD;

    /// <summary>振幅変調・シフト係数</summary>
    private static double shiftDT, swingDTa, swingDTb, shiftHR;

    /// <summary>年周期フーリエ係数</summary>
    private static readonly double[] acaDT, bcaDT, acaHR, bcaHR, acaAT, bcaAT,
      acFF, bcFF, acCC, bcCC, acDTSD, bcDTSD, acHRSD, bcHRSD;

    /// <summary>日周期フーリエ係数</summary>
    private static readonly double[] accDT, bccDT, 
      accHR, bccHR, acCLD, bcCLD, acFAR, bcFAR;

    /// <summary>相対湿度下限値</summary>
    private static double minimumRelativeHumidity = 10;

    #region 東京

    //トレンド成分標準偏差
    private static double sdevTDT_Tokyo = 0.5686;
    private static double sdevTHR_Tokyo = 0.3078;
    private static double sdevTAT_Tokyo = 0.0241;

    //ARモデル係数
    private static double[] iDTCof_Tokyo = new double[] { 1.1366, 0.0836, -0.0915, 0.0017, -0.1000, -0.0253 };
    private static double[] iHRCof_Tokyo = new double[] { 0.0212, 1.0330, 0.0186, -0.0618, -0.0262, -0.0197 };
    private static double iATCof_Tokyo = 0.8306;
    private static double iDTSD_Tokyo = 0.4572;
    private static double iHRSD_Tokyo = 0.3636;
    private static double iATSD_Tokyo = 0.0289;

    //振幅変調・シフト係数
    private static double shiftDT_Tokyo = 2.584;
    private static double swingDTa_Tokyo = 1.118;
    private static double swingDTb_Tokyo = 1.249;
    private static double shiftHR_Tokyo = -2.806;

    //年周期フーリエ係数
    private static readonly double[] acaDT_Tokyo = new double[] { 16.1553, -9.1951, -0.5818, 0.6386, -0.4619 };
    private static readonly double[] bcaDT_Tokyo = new double[] { 0.0000, 4.5427, -0.3686, 0.5093, -0.2657 };
    private static readonly double[] acaHR_Tokyo = new double[] { 8.4347, -5.8197, 0.2694, 0.3759, -0.1934 };
    private static readonly double[] bcaHR_Tokyo = new double[] { 0.0000, 2.9745, -1.2801, 0.4618, -0.0777 };
    private static readonly double[] acaAT_Tokyo = new double[] { 0.3797, 0.1610, 0.0332, 0.0231, -0.0285 };
    private static readonly double[] bcaAT_Tokyo = new double[] { 0.0000, -0.0061, 0.0052, 0.0085, -0.0225 };
    private static readonly double[] acFF_Tokyo = new double[] { 0.8132, 0.0770, 0.0077, 0.0152, -0.0207 };
    private static readonly double[] bcFF_Tokyo = new double[] { 0.0000, 0.0002, -0.0101, 0.0192, -0.0217 };
    private static readonly double[] acCC_Tokyo = new double[] { 0.7975, -0.0427, -0.0228, -0.0255, 0.0199 };
    private static readonly double[] bcCC_Tokyo = new double[] { 0.0000, -0.0055, 0.0156, -0.0112, 0.0037 };
    private static readonly double[] acDTSD_Tokyo = new double[] { 1.2483, 0.0497, -0.0513, 0, 0 };
    private static readonly double[] bcDTSD_Tokyo = new double[] { 0.0000, -0.1111, 0.0172, 0, 0 };
    private static readonly double[] acHRSD_Tokyo = new double[] { 1.3130, -0.4424, -0.1839, 0, 0 };
    private static readonly double[] bcHRSD_Tokyo = new double[] { 0.0000, 0.2039, 0.0424, 0, 0 };

    //日周期フーリエ係数
    private static readonly double[] accDT_Tokyo = new double[] { 0.0000, -1.5906, 0.5391, -0.0523 };
    private static readonly double[] bccDT_Tokyo = new double[] { 0.0000, 1.7772, -0.2794, -0.0602 };
    private static readonly double[] accHR_Tokyo = new double[] { 0.0000, 0.1410, -0.0029, -0.0107 };
    private static readonly double[] bccHR_Tokyo = new double[] { 0.0000, 0.1681, 0.0217, -0.0091 };
    private static readonly double[] acCLD_Tokyo = new double[] { 0.0450, 0.3672, 0.0000, -0.0390 };
    private static readonly double[] bcCLD_Tokyo = new double[] { 0.0000, 0.0230, 0.0000, -0.0037 };
    private static readonly double[] acFAR_Tokyo = new double[] { 0.3807, 0.2269, 0.0000, -0.0008 };
    private static readonly double[] bcFAR_Tokyo = new double[] { 0.0000, 0.0062, 0.0000, -0.0019 };

    //最小相対湿度
    private static double minimumRelativeHumidity_Tokyo = 10;

    #endregion

    #region 大阪

    //トレンド成分標準偏差
    private static double sdevTDT_Osaka = 0.5154;
    private static double sdevTHR_Osaka = 0.2984;
    private static double sdevTAT_Osaka = 0.0323;

    //ARモデル係数
    private static double[] iDTCof_Osaka = new double[] { 1.1584	,0.0931	,-0.0837,	-0.0088,	-0.1388,	-0.0284 };
    private static double[] iHRCof_Osaka = new double[] { -0.0086, 1.0589, 0.0413, -0.0643, -0.0146, -0.0403 };
    private static double iATCof_Osaka = 0.7394;
    private static double iDTSD_Osaka = 0.4360;
    private static double iHRSD_Osaka = 0.3315;
    private static double iATSD_Osaka = 0.0906;

    //振幅変調・シフト係数
    private static double shiftDT_Osaka = 0;
    private static double swingDTa_Osaka = 1.486;
    private static double swingDTb_Osaka = 1.337;
    private static double shiftHR_Osaka = -4.173;

    //年周期フーリエ係数
    private static readonly double[] acaDT_Osaka = new double[] { 16.8268, -9.3956, -0.3928, 0.3107, -0.2057 };
    private static readonly double[] bcaDT_Osaka = new double[] { 0.0000, 4.4420, -0.5665, 0.5255, -0.3072 };
    private static readonly double[] acaHR_Osaka = new double[] { 9.4201, -6.0004, 0.5700, 0.1399, -0.0428 };
    private static readonly double[] bcaHR_Osaka = new double[] { 0.0000, 2.8796, -1.3209, 0.5317, -0.0812 };
    private static readonly double[] acaAT_Osaka = new double[] { 0.3703, 0.0655, -0.0209, 0.0046, -0.0212 };
    private static readonly double[] bcaAT_Osaka = new double[] { 0.0000, 0.0317, 0.0183, 0.0068, -0.0205 };
    private static readonly double[] acFF_Osaka = new double[] { 0.8090, 0.0109, -0.0331, 0.0088, -0.0146 };
    private static readonly double[] bcFF_Osaka = new double[] { 0.0000, 0.0223, 0.0094, 0.0152, -0.0122 };
    private static readonly double[] acCC_Osaka = new double[] { 0.7729, -0.0411, -0.0171, -0.0087, 0.0242 };
    private static readonly double[] bcCC_Osaka = new double[] { 0.0000, -0.0324, 0.0147, -0.0103, 0.0210 };
    private static readonly double[] acDTSD_Osaka = new double[] { 1.3801, 0.1494, 0.0147, 0, 0 };
    private static readonly double[] bcDTSD_Osaka = new double[] { 0.0000, -0.1642, 0.0041, 0, 0 };
    private static readonly double[] acHRSD_Osaka = new double[] { 1.2499, -0.3970, -0.1582, 0, 0 };
    private static readonly double[] bcHRSD_Osaka = new double[] { 0.0000, 0.1771, 0.1389, 0, 0 };

    //日周期フーリエ係数
    private static readonly double[] accDT_Osaka = new double[] { 0.0000, -1.7793, 0.6353, -0.0995 };
    private static readonly double[] bccDT_Osaka = new double[] { 0.0000, 1.6188, -0.2374, -0.1190 };
    private static readonly double[] accHR_Osaka = new double[] { 0.0000, -0.1133, 0.0500, 0.0032 };
    private static readonly double[] bccHR_Osaka = new double[] { 0.0000, 0.2190, -0.0122, 0.0042 };
    private static readonly double[] acCLD_Osaka = new double[] { 0.0040, 0.3535, 0.0000, -0.0401 };
    private static readonly double[] bcCLD_Osaka = new double[] { 0.0000, -0.0277, 0.0000, 0.0049 };
    private static readonly double[] acFAR_Osaka = new double[] { 0.3579, 0.1904, 0.0000, 0.0053 };
    private static readonly double[] bcFAR_Osaka = new double[] { 0.0000, -0.0113, 0.0000, 0.0056 };

    //最小相対湿度
    private static readonly double minimumRelativeHumidity_Osaka = 10;

    #endregion

    #region 札幌

    //トレンド成分標準偏差
    private static double sdevTDT_Sapporo = 0.5686;
    private static double sdevTHR_Sapporo = 0.3078;
    private static double sdevTAT_Sapporo = 0.0241;

    //ARモデル係数
    private static double[] iDTCof_Sapporo = new double[] { 1.1456, 0.1611, -0.1402, -0.0552, -0.0701, -0.0481 };
    private static double[] iHRCof_Sapporo = new double[] { 0.0625, 0.9749, -0.0017, -0.0038, -0.0330, -0.0330 };
    private static double iATCof_Sapporo = 0.7219;
    private static double iDTSD_Sapporo = 0.4664;
    private static double iHRSD_Sapporo = 0.3704;
    private static double iATSD_Sapporo = 0.1025;

    //振幅変調・シフト係数
    private static double shiftDT_Sapporo = 0.723;
    private static double swingDTa_Sapporo = 1.346;
    private static double swingDTb_Sapporo = 1.343;
    private static double shiftHR_Sapporo = -1.613;
    
    //年周期フーリエ係数
    private static readonly double[] acaDT_Sapporo = new double[] { 8.8674, -11.5445, -0.7236, 0.5122, -0.2416 };
    private static readonly double[] bcaDT_Sapporo = new double[] { 0.0000, 5.3594, -0.4775, 0.5302, -0.2817 };
    private static readonly double[] acaHR_Sapporo = new double[] { 5.8837, -4.3882, 0.5599, 0.2440, -0.2687 };
    private static readonly double[] bcaHR_Sapporo = new double[] { 0.0000, 2.3810, -1.3414, 0.4957, -0.1201 };
    private static readonly double[] acaAT_Sapporo = new double[] { 0.3778, 0.0967, -0.0072, -0.0008, 0.0109 };
    private static readonly double[] bcaAT_Sapporo = new double[] { 0.0000, 0.0038, 0.0028, -0.0116, 0.0003 };
    private static readonly double[] acFF_Sapporo = new double[] { 0.7343, -0.0526, -0.0393, 0.0037, 0.0138 };
    private static readonly double[] bcFF_Sapporo = new double[] { 0.0000, 0.0256, 0.0114, -0.0118, -0.0041 };
    private static readonly double[] acCC_Sapporo = new double[] { 0.7694, -0.0497, 0.0083, 0.0141, 0.0022 };
    private static readonly double[] bcCC_Sapporo = new double[] { 0.0000, -0.0067, -0.0044, 0.0022, 0.0001 };
    private static readonly double[] acDTSD_Sapporo = new double[] { 1.5794, 0.4146, 0.0894, 0, 0 };
    private static readonly double[] bcDTSD_Sapporo = new double[] { 0.0000, -0.0883, 0.1445, 0, 0 };
    private static readonly double[] acHRSD_Sapporo = new double[] { 0.9420, -0.3845, -0.0669, 0, 0 };
    private static readonly double[] bcHRSD_Sapporo = new double[] { 0.0000, 0.3613, -0.0884, 0, 0 };

    //日周期フーリエ係数
    private static readonly double[] accDT_Sapporo = new double[] { 0.0000, -2.0102, 0.6600, -0.0203 };
    private static readonly double[] bccDT_Sapporo = new double[] { 0.0000, 1.2000, -0.0492, -0.1270 };
    private static readonly double[] accHR_Sapporo = new double[] { 0.0000, -0.1166, 0.0264, 0.0067 };
    private static readonly double[] bccHR_Sapporo = new double[] { 0.0000, 0.1300, 0.0315, -0.0058 };
    private static readonly double[] acCLD_Sapporo = new double[] { 0.0772, 0.4520, 0.0662, -0.0249 };
    private static readonly double[] bcCLD_Sapporo = new double[] { 0.0000, 0.0431, 0.0163, -0.0144 };
    private static readonly double[] acFAR_Sapporo = new double[] { 0.3787, 0.2029, 0.0115, 0.0138 };
    private static readonly double[] bcFAR_Sapporo = new double[] { 0.0000, 0.0036, 0.0034, -0.0043 };

    //最小相対湿度
    private static readonly double minimumRelativeHumidity_Sapporo = 10;

    #endregion

    #region 仙台

    //トレンド成分標準偏差
    private static double sdevTDT_Sendai = 0.6328;
    private static double sdevTHR_Sendai = 0.3540;
    private static double sdevTAT_Sendai = 0.0200;

    //ARモデル係数
    private static double[] iDTCof_Sendai = new double[] { 1.0883, 0.1604, -0.0766, -0.0292, -0.0789, -0.0603 };
    private static double[] iHRCof_Sendai = new double[] { 0.0353, 1.0105, 0.0100, -0.0091, -0.0327, -0.0448 };
    private static double iATCof_Sendai = 0.7464;
    private static double iDTSD_Sendai = 0.4915;
    private static double iHRSD_Sendai = 0.3376;
    private static double iATSD_Sendai = 0.0964;

    //振幅変調・シフト係数
    private static double shiftDT_Sendai = 1.794;
    private static double swingDTa_Sendai = 1.284;
    private static double swingDTb_Sendai = 1.304;
    private static double shiftHR_Sendai = -2.182;

    //年周期フーリエ係数
    private static readonly double[] acaDT_Sendai = new double[] { 12.3043, -9.6166, -0.6660, 0.6683, -0.4191 };
    private static readonly double[] bcaDT_Sendai = new double[] { 0.0000, 5.0537, -0.4351, 0.4763, -0.1837 };
    private static readonly double[] acaHR_Sendai = new double[] { 7.4230, -5.1868, 0.4997, 0.4007, -0.3268 };
    private static readonly double[] bcaHR_Sendai = new double[] { 0.0000, 2.9537, -1.4521, 0.4731, -0.0609 };
    private static readonly double[] acaAT_Sendai = new double[] { 0.3801, 0.1609, -0.0026, 0.0128, -0.0174 };
    private static readonly double[] bcaAT_Sendai = new double[] { 0.0000, -0.0168, 0.0156, 0.0043, -0.0167 };
    private static readonly double[] acFF_Sendai = new double[] { 0.7986, 0.0491, -0.0205, 0.0078, -0.0144 };
    private static readonly double[] bcFF_Sendai = new double[] { 0.0000, -0.0169, 0.0182, -0.0022, -0.0056 };
    private static readonly double[] acCC_Sendai = new double[] { 0.7758, -0.1076, -0.0254, -0.0058, 0.0141 };
    private static readonly double[] bcCC_Sendai = new double[] { 0.0000, 0.0066, 0.0007, -0.0130, 0.0040 };
    private static readonly double[] acDTSD_Sendai = new double[] { 1.4371, 0.0498, 0.0356, 0, 0 };
    private static readonly double[] bcDTSD_Sendai = new double[] { 0.0000, -0.1038, 0.1049, 0, 0 };
    private static readonly double[] acHRSD_Sendai = new double[] { 1.1640, -0.5124, -0.1287, 0, 0 };
    private static readonly double[] bcHRSD_Sendai = new double[] { 0.0000, 0.3059, -0.0090, 0, 0 };
    //日周期フーリエ係数
    private static readonly double[] accDT_Sendai = new double[] { 0.0000, -1.9449, 0.7034, -0.0493 };
    private static readonly double[] bccDT_Sendai = new double[] { 0.0000, 1.3603, -0.1132, -0.1271 };
    private static readonly double[] accHR_Sendai = new double[] { 0.0000, -0.0205, 0.0410, -0.0030 };
    private static readonly double[] bccHR_Sendai = new double[] { 0.0000, 0.1717, 0.0334, -0.0134 };
    private static readonly double[] acCLD_Sendai = new double[] { 0.0180, 0.3333, 0.0000, -0.0400 };
    private static readonly double[] bcCLD_Sendai = new double[] { 0.0000, 0.0282, 0.0000, -0.0099 };
    private static readonly double[] acFAR_Sendai = new double[] { 0.3852, 0.2452, 0.0000, 0.0117 };
    private static readonly double[] bcFAR_Sendai = new double[] { 0.0000, 0.0005, 0.0000, 0.0002 };

    //最小相対湿度
    private static readonly double minimumRelativeHumidity_Sendai = 10;

    #endregion

    #region 福岡

    //トレンド成分標準偏差
    private static double sdevTDT_Fukuoka = 0.4559;
    private static double sdevTHR_Fukuoka = 0.2169;
    private static double sdevTAT_Fukuoka = 0.0221;

    //ARモデル係数
    private static double[] iDTCof_Fukuoka = new double[] { 1.1385, 0.0996, -0.1107, -0.0029, -0.0907, -0.0418 };
    private static double[] iHRCof_Fukuoka = new double[] { 0.0264, 0.9199, 0.0160, 0.0400, -0.0145, -0.0172 };
    private static double iATCof_Fukuoka = 0.7344;
    private static double iDTSD_Fukuoka = 0.4749;
    private static double iHRSD_Fukuoka = 0.3547;
    private static double iATSD_Fukuoka = 0.0871;

    //振幅変調・シフト係数
    private static double shiftDT_Fukuoka = -0.162;
    private static double swingDTa_Fukuoka = 1.324;
    private static double swingDTb_Fukuoka = 1.290;
    private static double shiftHR_Fukuoka = -3.323;

    //年周期フーリエ係数
    private static readonly double[] acaDT_Fukuoka = new double[] { 16.8268, -9.3956, -0.3928, 0.3107, -0.2057 };
    private static readonly double[] bcaDT_Fukuoka = new double[] { 0.0000, 4.4420, -0.5665, 0.5255, -0.3072 };
    private static readonly double[] acaHR_Fukuoka = new double[] { 9.4201, -6.0004, 0.5700, 0.1399, -0.0428 };
    private static readonly double[] bcaHR_Fukuoka = new double[] { 0.0000, 2.8796, -1.3209, 0.5317, -0.0812 };
    private static readonly double[] acaAT_Fukuoka = new double[] { 0.3703, 0.0655, -0.0209, 0.0046, -0.0212 };
    private static readonly double[] bcaAT_Fukuoka = new double[] { 0.0000, 0.0317, 0.0183, 0.0068, -0.0205 };
    private static readonly double[] acFF_Fukuoka = new double[] { 0.8090, 0.0109, -0.0331, 0.0088, -0.0146 };
    private static readonly double[] bcFF_Fukuoka = new double[] { 0.0000, 0.0223, 0.0094, 0.0152, -0.0122 };
    private static readonly double[] acCC_Fukuoka = new double[] { 0.7729, -0.0411, -0.0171, -0.0087, 0.0242 };
    private static readonly double[] bcCC_Fukuoka = new double[] { 0.0000, -0.0324, 0.0147, -0.0103, 0.0210 };
    private static readonly double[] acDTSD_Fukuoka = new double[] { 1.3338, 0.1684, 0.1002, 0, 0 };
    private static readonly double[] bcDTSD_Fukuoka = new double[] { 0.0000, -0.2047, 0.0312, 0, 0 };
    private static readonly double[] acHRSD_Fukuoka = new double[] { 1.2744, -0.1366, -0.1652, 0, 0 };
    private static readonly double[] bcHRSD_Fukuoka = new double[] { 0.0000, 0.1132, 0.1945, 0, 0 };

    //日周期フーリエ係数
    private static readonly double[] accDT_Fukuoka = new double[] { 0.0000, -1.7793, 0.6353, -0.0995 };
    private static readonly double[] bccDT_Fukuoka = new double[] { 0.0000, 1.6188, -0.2374, -0.1190 };
    private static readonly double[] accHR_Fukuoka = new double[] { 0.0000, -0.1133, 0.0500, 0.0032 };
    private static readonly double[] bccHR_Fukuoka = new double[] { 0.0000, 0.2190, -0.0122, 0.0042 };
    private static readonly double[] acCLD_Fukuoka = new double[] { 0.0040, 0.3535, 0.0000, -0.0401 };
    private static readonly double[] bcCLD_Fukuoka = new double[] { 0.0000, -0.0277, 0.0000, 0.0049 };
    private static readonly double[] acFAR_Fukuoka = new double[] { 0.3579, 0.1904, 0.0000, 0.0053 };
    private static readonly double[] bcFAR_Fukuoka = new double[] { 0.0000, -0.0113, 0.0000, 0.0056 };

    //最小相対湿度
    private static readonly double minimumRelativeHumidity_Fukuoka = 10;

    #endregion

    #region 那覇

    //トレンド成分標準偏差
    private static double sdevTDT_Naha = 0.4332;
    private static double sdevTHR_Naha = 0.3588;
    private static double sdevTAT_Naha = 0.0210;

    //ARモデル係数
    private static double[] iDTCof_Naha = new double[] { 0.9343, 0.0676, -0.0137, 0.0114, -0.0049, 0.0145 };
    private static double[] iHRCof_Naha = new double[] { -0.0082, 0.7498, 0.0229, 0.1372, 0.0045, 0.0637 };
    private static double iATCof_Naha = 0.7364;
    private static double iDTSD_Naha = 0.6282;
    private static double iHRSD_Naha = 0.3404;
    private static double iATSD_Naha = 0.1009;

    //振幅変調・シフト係数
    private static double shiftDT_Naha = 1.899;
    private static double swingDTa_Naha = 0.582;
    private static double swingDTb_Naha = 0.811;
    private static double shiftHR_Naha = -3.168;

    //年周期フーリエ係数
    private static readonly double[] acaDT_Naha = new double[] { 22.9623, -5.3214, -0.2766, -0.1536, 0.1923 };
    private static readonly double[] bcaDT_Naha = new double[] { 0.0000, 2.9532, 0.0588, 0.1686, -0.0156 };
    private static readonly double[] acaHR_Naha = new double[] { 13.8874, -5.6110, 0.0933, -0.0870, 0.1813 };
    private static readonly double[] bcaHR_Naha = new double[] { 0.0000, 2.1287, -0.3409, 0.0453, 0.1980 };
    private static readonly double[] acaAT_Naha = new double[] { 0.3407, 0.0017, 0.0121, -0.0082, 0.0056 };
    private static readonly double[] bcaAT_Naha = new double[] { 0.0000, 0.0724, 0.0061, 0.0043, -0.0122 };
    private static readonly double[] acFF_Naha = new double[] { 0.7991, -0.0288, -0.0036, 0.0057, 0.0024 };
    private static readonly double[] bcFF_Naha = new double[] { 0.0000, 0.0727, -0.0097, 0.0093, -0.0065 };
    private static readonly double[] acCC_Naha = new double[] { 0.7882, 0.0221, -0.0010, 0.0026, -0.0179 };
    private static readonly double[] bcCC_Naha = new double[] { 0.0000, -0.0490, 0.0134, -0.0110, 0.0071 };
    private static readonly double[] acDTSD_Naha = new double[] { 0.8363, 0.2361, 0.0041, 0, 0 };
    private static readonly double[] bcDTSD_Naha = new double[] { 0.0000, -0.1779, -0.0364, 0, 0 };
    private static readonly double[] acHRSD_Naha = new double[] { 1.9009, 0.3594, -0.4965, 0, 0 };
    private static readonly double[] bcHRSD_Naha = new double[] { 0.0000, -0.3534, 0.4481, 0, 0 };

    //日周期フーリエ係数
    private static readonly double[] accDT_Naha = new double[] { 0.0000, -1.1943, 0.4402, -0.0188 };
    private static readonly double[] bccDT_Naha = new double[] { 0.0000, 0.7588, -0.1384, -0.0821 };
    private static readonly double[] accHR_Naha = new double[] { 0.0000, -0.1003, 0.0246, 0.0117 };
    private static readonly double[] bccHR_Naha = new double[] { 0.0000, -0.0056, 0.0006, 0.0013 };
    private static readonly double[] acCLD_Naha = new double[] { 0.0626, 0.3517, 0.0000, -0.0314 };
    private static readonly double[] bcCLD_Naha = new double[] { 0.0000, -0.0491, 0.0000, 0.0113 };
    private static readonly double[] acFAR_Naha = new double[] { 0.3428, 0.2665, 0.0000, 0.0165 };
    private static readonly double[] bcFAR_Naha = new double[] { 0.0000, -0.0426, 0.0000, -0.0061 };

    //最小相対湿度
    private static readonly double minimumRelativeHumidity_Naha = 30;

    #endregion

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタ</summary>
    static RandomWeather()
    {
      iDTCof = new double[6];
      iHRCof = new double[6];
      acaDT = new double[5];
      bcaDT = new double[5];
      acaHR = new double[5];
      bcaHR = new double[5];
      acaAT = new double[5];
      bcaAT = new double[5];
      acFF = new double[5];
      bcFF = new double[5];
      acCC = new double[5];
      bcCC = new double[5];
      acDTSD = new double[5];
      bcDTSD = new double[5];
      acHRSD = new double[5];
      bcHRSD = new double[5];

      accDT = new double[4];
      bccDT = new double[4]; 
      accHR = new double[4];
      bccHR = new double[4];
      acCLD = new double[4];
      bcCLD = new double[4];
      acFAR = new double[4];
      bcFAR = new double[4];
    }

    #endregion

    #region メイン処理

    /// <summary>確率的に気象データを発生させる</summary>
    /// <param name="seed">乱数シード</param>
    /// <param name="location">地点</param>
    /// <param name="year">発生させる年数</param>
    /// <param name="drybulbTemperature">出力:乾球温度[C]</param>
    /// <param name="humidityRatio">出力:絶対湿度[g/kg]</param>
    /// <param name="radiation">出力:水平面全天日射[W/m2]</param>
    /// <param name="isFair">出力:晴れか否か</param>
    public static void MakeWeather
      (int seed, Location location, int year, out double[] drybulbTemperature,
      out double[] humidityRatio, out double[] radiation, out bool[] isFair)
    {
      //係数初期化
      initParametor(location);

      MersenneTwister rnd = new MersenneTwister(seed);
      NormalRandom nRnd = new NormalRandom(seed);
      DateTime dt = new DateTime(2001, 1, 1, 0, 0, 0);

      int totalDay = year * 365;
      int totalHour = totalDay * 24;
      drybulbTemperature = new double[totalHour];
      humidityRatio = new double[totalHour];
      radiation = new double[totalHour];
      double[] swing = new double[totalDay];
      double[] dbtRnd = new double[totalDay];
      double[] hrtRnd = new double[totalDay];
      isFair = new bool[totalHour];

      //トレンド成分を計算
      double[] trendDT = new double[year];
      double[] trendHR = new double[year];
      double[] trendAT = new double[year];
      for (int i = 0; i < year; i++)
      {
        trendDT[i] = nRnd.NextDouble() * sdevTDT;
        trendHR[i] = nRnd.NextDouble() * sdevTHR;
        trendAT[i] = nRnd.NextDouble() * sdevTAT;
      }

      //確定的年周期成分を計算
      double[] caDBT, caHRT, caATM, caFTF, caCTC, caDTSIG, caHRSIG;
      makeAnnualData
        (out caDBT, out caHRT, out caATM, out caFTF, out caCTC, out caDTSIG, out caHRSIG);
      caATM = interpolateDailyData(caATM);
      caFTF = interpolateDailyData(caFTF);
      caCTC = interpolateDailyData(caCTC);
      caDTSIG = interpolateDailyData(caDTSIG);
      caHRSIG = interpolateDailyData(caHRSIG);

      //確定的日周期成分を計算
      double[] ccDBT, ccHRT, ccATMF, ccATMC;
      makeCircadianData(out ccDBT, out ccHRT, out ccATMF, out ccATMC);

      //水平面全天日射の計算
      int tHour = 0;
      for (int i = 0; i < totalDay; i++)
      {
        int yHour = tHour % 8760;
        int cYear = i / 365;

        //晴れ曇りの状態をマルコフ連鎖で計算
        for (int j = 0; j < 8; j++)
        {
          int ch = tHour + 3 * j;
          //0時点の状態は不変分布から計算
          if (ch == 0)
            isFair[0] = isFair[1] = isFair[2] = 
              ((1 - caCTC[yHour]) / (2 - caFTF[yHour] - caCTC[yHour])) < rnd.NextDouble();
          //その他の時点の状態は推移確率から計算
          else
          {
            bool curF = isFair[ch - 1];
            if (curF) isFair[ch] = rnd.NextDouble() < caFTF[yHour];
            else isFair[ch] = caCTC[yHour] < rnd.NextDouble();
          }
          isFair[ch + 2] = isFair[ch + 1] = isFair[ch];
        }

        //不規則変動成分の計算
        double[] iATM = new double[24];
        //時点0の場合には助走計算
        if (i == 0)
        {
          for (int j = 0; j < 50; j++)
          {
            iATM[23] = iATM[23] * iATCof + nRnd.NextDouble() * iATSD;
            iATM[23] = iATM[23] * iATCof + nRnd.NextDouble() * iATSD;
          }
        }
        iATM[0] = iATM[23] * iATCof + nRnd.NextDouble() * iATSD;
        iATM[1] = iATM[0] * iATCof + nRnd.NextDouble() * iATSD;
        for (int j = 1; j < 12; j++)
        {
          iATM[j * 2] = iATM[j * 2 - 1] * iATCof + nRnd.NextDouble() * iATSD;
          iATM[j * 2 + 1] = iATM[j * 2] * iATCof + nRnd.NextDouble() * iATSD;
        }

        //確定成分と不規則変動成分を集計
        int nSum = 0;
        double rSum = 0;
        for (int j = 0; j < 24; j++)
        {
          int ch = tHour + j;
          sun.Update(dt.AddHours(j));
          if (0 < sun.Altitude)
          {
            nSum++;
            double idisF = (1 - caCTC[yHour]) / (2 - caFTF[yHour] - caCTC[yHour]);
            rSum -= idisF * ccATMF[j] + (1 - idisF) * ccATMC[j];
            if (isFair[ch])
            {
              radiation[ch] = trendAT[cYear] + caATM[yHour] + iATM[j] + ccATMF[j];
              rSum += iATM[j] + ccATMF[j];
            }
            else
            {
              radiation[ch] = trendAT[cYear] + caATM[yHour] + iATM[j] + ccATMC[j];
              rSum += iATM[j] + ccATMC[j];
            }

            //大気透過率から水平面全天日射を計算（渡辺の式）
            radiation[ch] = Math.Min(1.0, Math.Max(0.001, radiation[ch]));
            radiation[ch] = getGlobalHorizontalRadiation(radiation[ch], sun);
          }
          else radiation[ch] = 0;
        }
        //不規則成分で振幅変調および絶対値シフト
        rSum /= nSum;
        dbtRnd[i] = caDBT[i % 365] +rSum * shiftDT;
        hrtRnd[i] = caHRT[i % 365] +rSum * shiftHR;
        swing[i] = rSum * swingDTa + swingDTb;

        dt = dt.AddDays(1);
        tHour += 24;
      }

      //振幅変調・絶対値シフトデータを時刻別データにSpline補間
      dbtRnd = interpolateDailyData(dbtRnd);
      hrtRnd = interpolateDailyData(hrtRnd);
      swing = interpolateDailyData(swing);

      //VARモデル助走計算
      double[] idbt = new double[3];
      double[] ihrt = new double[3];
      for (int i = 0; i < 100; i++)
        updateRandomComponent(nRnd, ref idbt, ref ihrt);
      
      //確定成分と不規則変動成分を合成
      tHour = 0;
      for (int i = 0; i < totalDay; i++)
      {
        int yHour = tHour % 8760;
        int cYear = i / 365;

        for (int j = 0; j < 24; j++)
        {
          int ch = tHour + j;
          updateRandomComponent(nRnd, ref idbt, ref ihrt);
          drybulbTemperature[ch] = trendDT[cYear] + 
            idbt[idbt.Length - 1] * caDTSIG[yHour] + dbtRnd[ch] + ccDBT[j] * swing[ch];
          humidityRatio[ch] = trendHR[cYear] + hrtRnd[ch] + ccHRT[j];
          double hrtMax = MoistAir.GetSaturationHumidityRatioFromDryBulbTemperature
            (drybulbTemperature[ch], 101.325) * 1000 - humidityRatio[ch];
          double hrtMin =
            MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
            (drybulbTemperature[ch], minimumRelativeHumidity, 101.325) * 1000;
          hrtMin -= humidityRatio[ch];
          ihrt[2] = (Math.Max(Math.Min(ihrt[2] * caHRSIG[yHour], hrtMax), hrtMin));
          ihrt[2] /= caHRSIG[yHour];
          humidityRatio[ch] += ihrt[2] * caHRSIG[yHour];
        }
        tHour += 24;
      }
    }

    #endregion
    
    #region 周期成分の計算

    /// <summary>年周期成分を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmTransmissivity">大気透過率[-]</param>
    /// <param name="fairToFair">晴れから晴れへの推移確率</param>
    /// <param name="cloudToCloud">曇りから曇りへの推移確率</param>
    /// <param name="dbtSigma">乾球温度不規則変動成分の標準偏差</param>
    /// <param name="hrtSigma">絶対湿度不規則変動成分の標準偏差</param>
    private static void makeAnnualData
      (out double[] drybulbTemperature, out double[] humidityRatio,
      out double[] atmTransmissivity, out double[] fairToFair, 
      out double[] cloudToCloud, out double[] dbtSigma, out double[] hrtSigma)
    {
      //日別年周期データを作成
      drybulbTemperature = new double[365];
      humidityRatio = new double[365];
      atmTransmissivity = new double[365];
      fairToFair = new double[365];
      cloudToCloud = new double[365];
      dbtSigma = new double[365];
      hrtSigma = new double[365];
      for (int i = 0; i < acaDT.Length; i++)
      {
        double wk = 2d * Math.PI * i / 365d;
        for (int j = 0; j < 365; j++)
        {
          double cwk = Math.Cos(wk * j);
          double swk = Math.Sin(wk * j);
          drybulbTemperature[j] += acaDT[i] * cwk - bcaDT[i] * swk;
          humidityRatio[j] += acaHR[i] * cwk - bcaHR[i] * swk;
          atmTransmissivity[j] += acaAT[i] * cwk - bcaAT[i] * swk;
          fairToFair[j] += acFF[i] * cwk - bcFF[i] * swk;
          cloudToCloud[j] += acCC[i] * cwk - bcCC[i] * swk;
          dbtSigma[j] += acDTSD[i] * cwk - bcDTSD[i] * swk;
          hrtSigma[j] += acHRSD[i] * cwk - bcHRSD[i] * swk;
        }
      }
    }

    /// <summary>日周期成分を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    /// <param name="atmTransFair">晴れの日の大気透過率[-]</param>
    /// <param name="atmTransCloudy">曇の日の大気透過率[-]</param>
    private static void makeCircadianData
      (out double[] drybulbTemperature, out double[] humidityRatio,
       out double[] atmTransFair, out double[] atmTransCloudy)
    {
      //時刻別データを作成
      drybulbTemperature = new double[24];
      humidityRatio = new double[24];
      atmTransCloudy = new double[24];
      atmTransFair = new double[24];
      for (int i = 0; i < accDT.Length; i++)
      {
        double wk = 2d * Math.PI * i / 24d;
        for (int j = 0; j < 24; j++)
        {
          double cwk = Math.Cos(wk * j);
          double swk = Math.Sin(wk * j);
          drybulbTemperature[j] += accDT[i] * cwk - bccDT[i] * swk;
          humidityRatio[j] += accHR[i] * cwk - bccHR[i] * swk;
          atmTransCloudy[j] += acCLD[i] * cwk - bcCLD[i] * swk;
          atmTransFair[j] += acFAR[i] * cwk - bcFAR[i] * swk;
        }
      }
    }

    #endregion

    #region 不規則変動成分の計算

    /// <summary>不規則変動成分を更新する</summary>
    /// <param name="nrnd">正規乱数製造機</param>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[g/kg]</param>
    private static void updateRandomComponent
      (NormalRandom nrnd, ref double[] drybulbTemperature, ref double[] humidityRatio)
    {
      double nrnd1 = nrnd.NextDouble();
      double nrnd2 = nrnd.NextDouble();
      double dbt = nrnd1 * iDTSD +
        drybulbTemperature[2] * iDTCof[0] + humidityRatio[2] * iDTCof[1] +
        drybulbTemperature[1] * iDTCof[2] + humidityRatio[1] * iDTCof[3] +
        drybulbTemperature[0] * iDTCof[4] + humidityRatio[0] * iDTCof[5];
      double hrt = nrnd2 * iHRSD +
        drybulbTemperature[2] * iHRCof[0] + humidityRatio[2] * iHRCof[1] +
        drybulbTemperature[1] * iHRCof[2] + humidityRatio[1] * iHRCof[3] +
        drybulbTemperature[0] * iHRCof[4] + humidityRatio[0] * iHRCof[5];
      drybulbTemperature[0] = drybulbTemperature[1];
      drybulbTemperature[1] = drybulbTemperature[2];
      drybulbTemperature[2] = dbt;
      humidityRatio[0] = humidityRatio[1];
      humidityRatio[1] = humidityRatio[2];
      humidityRatio[2] = hrt;
    }

    #endregion

    #region その他の計算

    /// <summary>初期化処理</summary>
    /// <param name="location">地点</param>
    private static void initParametor(Location location)
    {
      switch (location)
      {
        case Location.Osaka:
          sun = new Sun(34, 40.7, 0, 135, 31.3, 0, 135, 0, 0);

          //トレンド成分標準偏差
          sdevTDT = sdevTDT_Osaka;
          sdevTHR = sdevTHR_Osaka;
          sdevTAT = sdevTAT_Osaka;

          //ARモデル係数
          iDTCof_Osaka.CopyTo(iDTCof, 0);
          iHRCof_Osaka.CopyTo(iHRCof, 0);
          iATCof = iATCof_Osaka;
          iDTSD = iDTSD_Osaka;
          iHRSD = iHRSD_Osaka;
          iATSD = iATSD_Osaka;

          //振幅変調・シフト係数
          shiftDT = shiftDT_Osaka;
          swingDTa = swingDTa_Osaka;
          swingDTb = swingDTb_Osaka;
          shiftHR = -shiftHR_Osaka;

          //年周期フーリエ係数
          acaDT_Osaka.CopyTo(acaDT, 0);
          bcaDT_Osaka.CopyTo(bcaDT, 0);
          acaHR_Osaka.CopyTo(acaHR, 0);
          bcaHR_Osaka.CopyTo(bcaHR, 0);
          acaAT_Osaka.CopyTo(acaAT, 0);
          bcaAT_Osaka.CopyTo(bcaAT, 0);
          acFF_Osaka.CopyTo(acFF, 0);
          bcFF_Osaka.CopyTo(bcFF, 0);
          acCC_Osaka.CopyTo(acCC, 0);
          bcCC_Osaka.CopyTo(bcCC, 0);
          acDTSD_Osaka.CopyTo(acDTSD, 0);
          bcDTSD_Osaka.CopyTo(bcDTSD, 0);
          acHRSD_Osaka.CopyTo(acHRSD, 0);
          bcHRSD_Osaka.CopyTo(bcHRSD, 0);

          //日周期フーリエ係数
          accDT_Osaka.CopyTo(accDT, 0);
          bccDT_Osaka.CopyTo(bccDT, 0);
          accHR_Osaka.CopyTo(accHR, 0);
          bccHR_Osaka.CopyTo(bccHR, 0);
          acCLD_Osaka.CopyTo(acCLD, 0);
          bcCLD_Osaka.CopyTo(bcCLD, 0);
          acFAR_Osaka.CopyTo(acFAR, 0);
          bcFAR_Osaka.CopyTo(bcFAR, 0);

          //最小相対湿度
          minimumRelativeHumidity = minimumRelativeHumidity_Osaka;
          break;
        case Location.Sapporo:
          sun = new Sun(43, 3.5, 0, 141, 19.9, 0, 135, 0, 0);

          //トレンド成分標準偏差
          sdevTDT = sdevTDT_Sapporo;
          sdevTHR = sdevTHR_Sapporo;
          sdevTAT = sdevTAT_Sapporo;

          //ARモデル係数
          iDTCof_Sapporo.CopyTo(iDTCof, 0);
          iHRCof_Sapporo.CopyTo(iHRCof, 0);
          iATCof = iATCof_Sapporo;
          iDTSD = iDTSD_Sapporo;
          iHRSD = iHRSD_Sapporo;
          iATSD = iATSD_Sapporo;

          //振幅変調・シフト係数
          shiftDT = shiftDT_Sapporo;
          swingDTa = swingDTa_Sapporo;
          swingDTb = swingDTb_Sapporo;
          shiftHR = -shiftHR_Sapporo;

          //年周期フーリエ係数
          acaDT_Sapporo.CopyTo(acaDT, 0);
          bcaDT_Sapporo.CopyTo(bcaDT, 0);
          acaHR_Sapporo.CopyTo(acaHR, 0);
          bcaHR_Sapporo.CopyTo(bcaHR, 0);
          acaAT_Sapporo.CopyTo(acaAT, 0);
          bcaAT_Sapporo.CopyTo(bcaAT, 0);
          acFF_Sapporo.CopyTo(acFF, 0);
          bcFF_Sapporo.CopyTo(bcFF, 0);
          acCC_Sapporo.CopyTo(acCC, 0);
          bcCC_Sapporo.CopyTo(bcCC, 0);
          acDTSD_Sapporo.CopyTo(acDTSD, 0);
          bcDTSD_Sapporo.CopyTo(bcDTSD, 0);
          acHRSD_Sapporo.CopyTo(acHRSD, 0);
          bcHRSD_Sapporo.CopyTo(bcHRSD, 0);

          //日周期フーリエ係数
          accDT_Sapporo.CopyTo(accDT, 0);
          bccDT_Sapporo.CopyTo(bccDT, 0);
          accHR_Sapporo.CopyTo(accHR, 0);
          bccHR_Sapporo.CopyTo(bccHR, 0);
          acCLD_Sapporo.CopyTo(acCLD, 0);
          bcCLD_Sapporo.CopyTo(bcCLD, 0);
          acFAR_Sapporo.CopyTo(acFAR, 0);
          bcFAR_Sapporo.CopyTo(bcFAR, 0);

          //最小相対湿度
          minimumRelativeHumidity = minimumRelativeHumidity_Sapporo;
          break;
        case Location.Sendai:
          sun = new Sun(38, 15.5, 0, 140, 54.0, 0, 135, 0, 0);

          //トレンド成分標準偏差
          sdevTDT = sdevTDT_Sendai;
          sdevTHR = sdevTHR_Sendai;
          sdevTAT = sdevTAT_Sendai;

          //ARモデル係数
          iDTCof_Sendai.CopyTo(iDTCof, 0);
          iHRCof_Sendai.CopyTo(iHRCof, 0);
          iATCof = iATCof_Sendai;
          iDTSD = iDTSD_Sendai;
          iHRSD = iHRSD_Sendai;
          iATSD = iATSD_Sendai;

          //振幅変調・シフト係数
          shiftDT = shiftDT_Sendai;
          swingDTa = swingDTa_Sendai;
          swingDTb = swingDTb_Sendai;
          shiftHR = -shiftHR_Sendai;

          //年周期フーリエ係数
          acaDT_Sendai.CopyTo(acaDT, 0);
          bcaDT_Sendai.CopyTo(bcaDT, 0);
          acaHR_Sendai.CopyTo(acaHR, 0);
          bcaHR_Sendai.CopyTo(bcaHR, 0);
          acaAT_Sendai.CopyTo(acaAT, 0);
          bcaAT_Sendai.CopyTo(bcaAT, 0);
          acFF_Sendai.CopyTo(acFF, 0);
          bcFF_Sendai.CopyTo(bcFF, 0);
          acCC_Sendai.CopyTo(acCC, 0);
          bcCC_Sendai.CopyTo(bcCC, 0);
          acDTSD_Sendai.CopyTo(acDTSD, 0);
          bcDTSD_Sendai.CopyTo(bcDTSD, 0);
          acHRSD_Sendai.CopyTo(acHRSD, 0);
          bcHRSD_Sendai.CopyTo(bcHRSD, 0);

          //日周期フーリエ係数
          accDT_Sendai.CopyTo(accDT, 0);
          bccDT_Sendai.CopyTo(bccDT, 0);
          accHR_Sendai.CopyTo(accHR, 0);
          bccHR_Sendai.CopyTo(bccHR, 0);
          acCLD_Sendai.CopyTo(acCLD, 0);
          bcCLD_Sendai.CopyTo(bcCLD, 0);
          acFAR_Sendai.CopyTo(acFAR, 0);
          bcFAR_Sendai.CopyTo(bcFAR, 0);

          //最小相対湿度
          minimumRelativeHumidity = minimumRelativeHumidity_Sendai;
          break;
        case Location.Fukuoka:
          sun = new Sun(33, 34.8, 0, 130, 22.6, 0, 135, 0, 0);

          //トレンド成分標準偏差
          sdevTDT = sdevTDT_Fukuoka;
          sdevTHR = sdevTHR_Fukuoka;
          sdevTAT = sdevTAT_Fukuoka;

          //ARモデル係数
          iDTCof_Fukuoka.CopyTo(iDTCof, 0);
          iHRCof_Fukuoka.CopyTo(iHRCof, 0);
          iATCof = iATCof_Fukuoka;
          iDTSD = iDTSD_Fukuoka;
          iHRSD = iHRSD_Fukuoka;
          iATSD = iATSD_Fukuoka;

          //振幅変調・シフト係数
          shiftDT = shiftDT_Fukuoka;
          swingDTa = swingDTa_Fukuoka;
          swingDTb = swingDTb_Fukuoka;
          shiftHR = -shiftHR_Fukuoka;

          //年周期フーリエ係数
          acaDT_Fukuoka.CopyTo(acaDT, 0);
          bcaDT_Fukuoka.CopyTo(bcaDT, 0);
          acaHR_Fukuoka.CopyTo(acaHR, 0);
          bcaHR_Fukuoka.CopyTo(bcaHR, 0);
          acaAT_Fukuoka.CopyTo(acaAT, 0);
          bcaAT_Fukuoka.CopyTo(bcaAT, 0);
          acFF_Fukuoka.CopyTo(acFF, 0);
          bcFF_Fukuoka.CopyTo(bcFF, 0);
          acCC_Fukuoka.CopyTo(acCC, 0);
          bcCC_Fukuoka.CopyTo(bcCC, 0);
          acDTSD_Fukuoka.CopyTo(acDTSD, 0);
          bcDTSD_Fukuoka.CopyTo(bcDTSD, 0);
          acHRSD_Fukuoka.CopyTo(acHRSD, 0);
          bcHRSD_Fukuoka.CopyTo(bcHRSD, 0);

          //日周期フーリエ係数
          accDT_Fukuoka.CopyTo(accDT, 0);
          bccDT_Fukuoka.CopyTo(bccDT, 0);
          accHR_Fukuoka.CopyTo(accHR, 0);
          bccHR_Fukuoka.CopyTo(bccHR, 0);
          acCLD_Fukuoka.CopyTo(acCLD, 0);
          bcCLD_Fukuoka.CopyTo(bcCLD, 0);
          acFAR_Fukuoka.CopyTo(acFAR, 0);
          bcFAR_Fukuoka.CopyTo(bcFAR, 0);

          //最小相対湿度
          minimumRelativeHumidity = minimumRelativeHumidity_Fukuoka;
          break;
        case Location.Naha:
          sun = new Sun(26, 12.2, 0, 127, 41.3, 0, 135, 0, 0);

          //トレンド成分標準偏差
          sdevTDT = sdevTDT_Naha;
          sdevTHR = sdevTHR_Naha;
          sdevTAT = sdevTAT_Naha;

          //ARモデル係数
          iDTCof_Naha.CopyTo(iDTCof, 0);
          iHRCof_Naha.CopyTo(iHRCof, 0);
          iATCof = iATCof_Naha;
          iDTSD = iDTSD_Naha;
          iHRSD = iHRSD_Naha;
          iATSD = iATSD_Naha;

          //振幅変調・シフト係数
          shiftDT = shiftDT_Naha;
          swingDTa = swingDTa_Naha;
          swingDTb = swingDTb_Naha;
          shiftHR = -shiftHR_Naha;

          //年周期フーリエ係数
          acaDT_Naha.CopyTo(acaDT, 0);
          bcaDT_Naha.CopyTo(bcaDT, 0);
          acaHR_Naha.CopyTo(acaHR, 0);
          bcaHR_Naha.CopyTo(bcaHR, 0);
          acaAT_Naha.CopyTo(acaAT, 0);
          bcaAT_Naha.CopyTo(bcaAT, 0);
          acFF_Naha.CopyTo(acFF, 0);
          bcFF_Naha.CopyTo(bcFF, 0);
          acCC_Naha.CopyTo(acCC, 0);
          bcCC_Naha.CopyTo(bcCC, 0);
          acDTSD_Naha.CopyTo(acDTSD, 0);
          bcDTSD_Naha.CopyTo(bcDTSD, 0);
          acHRSD_Naha.CopyTo(acHRSD, 0);
          bcHRSD_Naha.CopyTo(bcHRSD, 0);

          //日周期フーリエ係数
          accDT_Naha.CopyTo(accDT, 0);
          bccDT_Naha.CopyTo(bccDT, 0);
          accHR_Naha.CopyTo(accHR, 0);
          bccHR_Naha.CopyTo(bccHR, 0);
          acCLD_Naha.CopyTo(acCLD, 0);
          bcCLD_Naha.CopyTo(bcCLD, 0);
          acFAR_Naha.CopyTo(acFAR, 0);
          bcFAR_Naha.CopyTo(bcFAR, 0);

          //最小相対湿度
          minimumRelativeHumidity = minimumRelativeHumidity_Naha;
          break;
        default://東京
          sun = new Sun(35, 41.2, 0, 139, 45.9, 0, 135, 0, 0);

          //トレンド成分標準偏差
          sdevTDT = sdevTDT_Tokyo;
          sdevTHR = sdevTHR_Tokyo;
          sdevTAT = sdevTAT_Tokyo;

          //ARモデル係数
          iDTCof_Tokyo.CopyTo(iDTCof, 0);
          iHRCof_Tokyo.CopyTo(iHRCof, 0);
          iATCof = iATCof_Tokyo;
          iDTSD = iDTSD_Tokyo;
          iHRSD = iHRSD_Tokyo;
          iATSD = iATSD_Tokyo;

          //振幅変調・シフト係数
          shiftDT = shiftDT_Tokyo;
          swingDTa = swingDTa_Tokyo;
          swingDTb = swingDTb_Tokyo;
          shiftHR = -shiftHR_Tokyo;

          //年周期フーリエ係数
          acaDT_Tokyo.CopyTo(acaDT, 0);
          bcaDT_Tokyo.CopyTo(bcaDT, 0);
          acaHR_Tokyo.CopyTo(acaHR, 0);
          bcaHR_Tokyo.CopyTo(bcaHR, 0);
          acaAT_Tokyo.CopyTo(acaAT, 0);
          bcaAT_Tokyo.CopyTo(bcaAT, 0);
          acFF_Tokyo.CopyTo(acFF, 0);
          bcFF_Tokyo.CopyTo(bcFF, 0);
          acCC_Tokyo.CopyTo(acCC, 0);
          bcCC_Tokyo.CopyTo(bcCC, 0);
          acDTSD_Tokyo.CopyTo(acDTSD, 0);
          bcDTSD_Tokyo.CopyTo(bcDTSD, 0);
          acHRSD_Tokyo.CopyTo(acHRSD, 0);
          bcHRSD_Tokyo.CopyTo(bcHRSD, 0);

          //日周期フーリエ係数
          accDT_Tokyo.CopyTo(accDT, 0);
          bccDT_Tokyo.CopyTo(bccDT, 0);
          accHR_Tokyo.CopyTo(accHR, 0);
          bccHR_Tokyo.CopyTo(bccHR, 0);
          acCLD_Tokyo.CopyTo(acCLD, 0);
          bcCLD_Tokyo.CopyTo(bcCLD, 0);
          acFAR_Tokyo.CopyTo(acFAR, 0);
          bcFAR_Tokyo.CopyTo(bcFAR, 0);

          //最小相対湿度
          minimumRelativeHumidity = minimumRelativeHumidity_Tokyo;
          break;
      }
    }

    /// <summary>日別データをSpline補間して時刻別データを作成する</summary>
    /// <param name="dailyData">日別データ</param>
    /// <returns>時刻別データ</returns>
    private static double[] interpolateDailyData(double[] dailyData)
    {
      double[] dd = new double[dailyData.Length + 2];
      dailyData.CopyTo(dd, 1);
      dd[0] = dd[dd.Length - 2];
      dd[dd.Length - 1] = dd[1];
      double[] sx = new double[dailyData.Length + 2];
      double[] hx = new double[dailyData.Length * 24];
      for (int i = 0; i < sx.Length; i++) sx[i] = -12 + 24 * i;
      for (int i = 0; i < hx.Length; i++) hx[i] = i;
      double[] cDD = CubicSpline.GetParameters(sx, dd);
      return CubicSpline.Interpolate(sx, dd, cDD, hx);
    }

    /// <summary>渡辺の手法で水平面全天日射[W/m2]を計算する</summary>
    /// <param name="aTransmissivity">大気透過率[-]</param>
    /// <param name="sun">計算地点の太陽</param>
    private static double getGlobalHorizontalRadiation
      (double aTransmissivity, ImmutableSun sun)
    {
      double sinAltitude = Math.Sin(sun.Altitude);
      double exRadiation = sun.GetExtraterrestrialRadiation();
      double ps = Math.Pow(aTransmissivity, 1 / sinAltitude);
      double directNormalRadiation = exRadiation * ps;
      double q = (0.8672 + 0.7505 * sinAltitude) *
        Math.Pow(aTransmissivity, 0.421 / sinAltitude) *
        Math.Pow(1d - Math.Pow(aTransmissivity, 1 / sinAltitude), 2.277);
      double diffuseHorizontalRadiation = exRadiation * sinAltitude * (q / (1 + q));
      return directNormalRadiation * sinAltitude + diffuseHorizontalRadiation;
    }

    #endregion

  }
}
