﻿/* Roots.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.Numerics
{
    /// <summary>一次元関数の求根を行うクラス</summary>
    public static class Roots
    {

        /// <summary>機械イプシロン</summary>
        private const double MECH_EPS = 1.0e-7;

        /// <summary>誤差関数</summary>
        /// <param name="x">入力値</param>
        /// <returns>出力値</returns>
        public delegate double ErrorFunction(double x);

        /// <summary>二分法で求根する</summary>
        /// <param name="a">求根範囲1</param>
        /// <param name="b">求根範囲2</param>
        /// <param name="errorTolerance">許容誤差</param>
        /// <param name="eFnc">誤差関数</param>
        /// <returns>解</returns>
        public static double Bisection
            (double a, double b, double errorTolerance, ErrorFunction eFnc)
        {
            const int MAX_ITER = 100;

            double fa = eFnc(a);
            double fb = eFnc(b);

            //根が囲い込めていない場合はエラー
            if (0 < fa * fb) throw new Exception("Bisection Error");

            int iterNum = 0;
            while (true)
            {
                //中間点を計算
                double c = 0.5 * (a + b);
                double fc = eFnc(c);
                //入替え処理
                if (Math.Sign(fc) == Math.Sign(fa))
                {
                    fa = fc;
                    a = c;
                }
                else
                {
                    fb = fc;
                    b = c;
                }
                if(Math.Abs(fc) < errorTolerance) return c;
                iterNum++;
                if (MAX_ITER < iterNum) throw new Exception("Bisection Iteration Error");
            }
        }

        /// <summary>Brent法で求根する</summary>
        /// <param name="a">求根範囲1</param>
        /// <param name="b">求根範囲2</param>
        /// <param name="errorTolerance">許容誤差</param>
        /// <param name="eFnc">誤差関数</param>
        /// <returns>解</returns>
        public static double Brent
            (double a, double b, double errorTolerance, ErrorFunction eFnc)
        {
            const int MAX_ITER = 100;

            double e = 0;
            double d = 0;
            double fa = eFnc(a);
            double fb = eFnc(b);
            double c = a;
            double fc = fa;

            int iterNum = 0;
            while (true)
            {
                //f(b)とf(c)が同一符号となった場合は入替え
                if ((0 < fb && 0 < fc) || (fb < 0 && fc < 0))
                {
                    c = a;
                    fc = fa;
                    e = d = b - a;
                }

                if (Math.Abs(fc) < Math.Abs(fa))
                {
                    a = b;
                    b = c;
                    c = a;
                    fa = fb;
                    fb = fc;
                    fc = fa;
                }

                //収束判定
                double tol = 2.0 * MECH_EPS * Math.Abs(b) + errorTolerance;
                double mid = 0.5 * (c - b);
                if (Math.Abs(mid) < tol || fb == 0.0) return b;

                if ((Math.Abs(e) < tol) || Math.Abs(fa) <= Math.Abs(fb))
                {
                    //区間幅の減少が小さい場合は二分法で処理
                    d = mid;
                    e = d;
                }
                else
                {
                    //逆二次補間を試す
                    double p, q, r;
                    double s = fb / fa;
                    if (a == c)
                    {
                        p = 2.0 * mid * s;
                        q = 1.0 - s;
                    }
                    else
                    {
                        q = fa / fc;
                        r = fb / fc;
                        p = s * (2.0 * mid * q * (q - r) - (b - a) * (r - 1.0));
                        q = (q - 1.0) * (r - 1.0) * (s - 1.0);
                    }
                    if (0 < p) q = -q;
                    p = Math.Abs(p);
                    double min1 = 3.0 * mid * mid * q - Math.Abs(tol * q);
                    double min2 = Math.Abs(e * q);
                    if (2.0 * p < Math.Min(min1, min2))
                    {
                        e = d;
                        d = p / q;
                    }
                    else
                    {
                        d = mid;
                        e = d;
                    }
                }
                a = b;
                fa = fb;
                if (tol < Math.Abs(d)) b += d;
                else b += Math.Sign(mid) * tol;
                fb = eFnc(b);

                iterNum++;
                if (MAX_ITER < iterNum) throw new Exception("Brent Iteration Error");
            }
        }

    }
}
