﻿/* Sun.cs
 * 
 * Copyright (C) 2008 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using System.Collections.Generic;

using Popolo.Numerics;

namespace Popolo.Weather
{
  /// <summary>太陽クラス</summary>
  /// <remarks>
  /// 数値計算で学ぶ光と熱の建築環境学, pp.20, 丸善, 宿谷昌則, 1993
  /// パソコンによる空気調和計算法, 宇田川光弘, 1986
  /// </remarks>
  public class Sun : ImmutableSun
  {

    #region 定数

    /// <summary>太陽定数[W/m2]</summary>
    public const double SOLAR_CONSTANT = 1367d;

    /// <summary>DegreeをRadianに変換する係数</summary>
    private const double DEG_TO_RAD = Math.PI / 180d;

    #endregion

    #region 列挙型

    /// <summary>直散分離の手法</summary>
    public enum SeparationMethod
    {
      /// <summary>Berlage</summary>
      /// <remarks>Berlage,Von H.P.:Zur Theorie der Beleuchtung einer horizontalen Flache durch Tageslicht,Meteorologische Zeitschrift, May 1928,pp.174-180</remarks>
      Berlage,
      /// <summary>松尾</summary>
      /// <remarks>松尾陽:日本建築学会論文報告集,快晴時の日射について 日射量に関する研究2,pp.21-24,1960</remarks>
      Matsuo,
      /// <summary>永田</summary>
      /// <remarks>永田忠彦:晴天空による水平面散乱の日射の式の試案,日本建築学会学術講演梗概集,1978</remarks>
      Nagata,
      /// <summary>Liu-Jordan</summary>
      /// <remarks>Liu,B.Y.H-Jordan,R.C:The interrelationship and characteristic distribution of direct, diffuse and total solar radiation, solar energy, Vol.4, No.3, 1960</remarks>
      LiuAndJordan,
      /// <summary>宇田川</summary>
      /// <remarks>宇田川光弘,木村建一:水平面全天日射量観測値よりの直達日射量の推定,日本建築学会論文報告集,No.267,pp.83-90,19780530</remarks>
      Udagawa,
      /// <summary>渡辺</summary>
      /// <remarks>渡辺俊行:水平面全天日射量の直散分離と傾斜面日射量の推定,日本建築学会論文報告集,No.330,pp.96-108,19830830</remarks>
      Watanabe,
      /// <summary>赤坂</summary>
      /// <remarks>H.Akasaka:Model of circumsolar radiation and diffuse sky radiation including cloudy sky, ISES, Solar World Congress, 1991</remarks>
      Akasaka,
      /// <summary>三木</summary>
      /// <remarks>三木信博:標準気象データの日射直散分離に関する研究 その6 日射直散分離法の提案,日本建築学会学術講演梗概集,pp.857-858,1991</remarks>
      Miki,
      /// <summary>Erbs</summary>
      /// <remarks>D.G. Erbs, S.A. Klein and J.A. Duffie: Estimation of the Diffuse Radiation Fraction for Hourly, Daily and Monthly-Average Global Radiation, Solar Energy, Vol.28, No.4, pp.293-302, 1982</remarks>
      Erbs
    }

    /// <summary>都市</summary>
    public enum City
    {
      /// <summary></summary>
      Aberdeen,
      /// <summary></summary>
      Algiers,
      /// <summary>アムステルダム</summary>
      Amsterdam,
      /// <summary>アンカラ</summary>
      Ankara,
      /// <summary></summary>
      Asuncion,
      /// <summary></summary>
      Athens,
      /// <summary></summary>
      Auckland,
      /// <summary>バンコク</summary>
      Bangkok,
      /// <summary></summary>
      Barcelona,
      /// <summary>北京</summary>
      Beijing,
      /// <summary></summary>
      Belem,
      /// <summary></summary>
      Belfast,
      /// <summary></summary>
      Belgrade,
      /// <summary></summary>
      Berlin,
      /// <summary></summary>
      Birmingham,
      /// <summary></summary>
      Bogota,
      /// <summary></summary>
      Bordeaux,
      /// <summary></summary>
      Bremen,
      /// <summary></summary>
      Brisbane,
      /// <summary></summary>
      Bristol,
      /// <summary></summary>
      Brussels,
      /// <summary></summary>
      Bucharest,
      /// <summary></summary>
      Budapest,
      /// <summary></summary>
      BuenosAires,
      /// <summary></summary>
      Cairo,
      /// <summary></summary>
      Canton,
      /// <summary></summary>
      CapeTown,
      /// <summary></summary>
      Caracas,
      /// <summary></summary>
      Cayenne,
      /// <summary></summary>
      Chihuahua,
      /// <summary></summary>
      Chongqing,
      /// <summary></summary>
      Copenhagen,
      /// <summary></summary>
      Cordoba,
      /// <summary></summary>
      Dakar,
      /// <summary></summary>
      Djibouti,
      /// <summary>ダブリン</summary>
      Dublin,
      /// <summary></summary>
      Durban,
      /// <summary></summary>
      Edinburgh,
      /// <summary></summary>
      Frankfurt,
      /// <summary></summary>
      Georgetown,
      /// <summary>グラスゴー</summary>
      Glasgow,
      /// <summary></summary>
      GuatemalaCity,
      /// <summary></summary>
      Guayaquil,
      /// <summary></summary>
      Hamburg,
      /// <summary></summary>
      Hammerfest,
      /// <summary>ハバナ</summary>
      Havana,
      /// <summary>ヘルシンキ</summary>
      Helsinki,
      /// <summary></summary>
      Hobart,
      /// <summary>香港</summary>
      HongKong,
      /// <summary></summary>
      Iquique,
      /// <summary>イルクーツク</summary>
      Irkutsk,
      /// <summary></summary>
      Jakarta,
      /// <summary></summary>
      Johannesburg,
      /// <summary></summary>
      Kingston,
      /// <summary></summary>
      Kinshasa,
      /// <summary></summary>
      KualaLumpur,
      /// <summary></summary>
      LaPaz,
      /// <summary></summary>
      Leeds,
      /// <summary>リマ</summary>
      Lima,
      /// <summary>リスボン</summary>
      Lisbon,
      /// <summary></summary>
      Liverpool,
      /// <summary>ロンドン</summary>
      London,
      /// <summary></summary>
      Lyons,
      /// <summary>マドリッド</summary>
      Madrid,
      /// <summary>マンチェスター</summary>
      Manchester,
      /// <summary></summary>
      Manila,
      /// <summary></summary>
      Marseilles,
      /// <summary></summary>
      Mazatlan,
      /// <summary>メッカ</summary>
      Mecca,
      /// <summary>メルボルン</summary>
      Melbourne,
      /// <summary>メキシコシティ</summary>
      MexicoCity,
      /// <summary></summary>
      Milan,
      /// <summary></summary>
      Montevideo,
      /// <summary></summary>
      Moscow,
      /// <summary></summary>
      Munich,
      /// <summary>長崎</summary>
      Nagasaki,
      /// <summary>名古屋</summary>
      Nagoya,
      /// <summary>ナイロビ</summary>
      Nairobi,
      /// <summary>南京</summary>
      Nanjing,
      /// <summary></summary>
      Naples,
      /// <summary></summary>
      NewcastleOnTyne,
      /// <summary></summary>
      Odessa,
      /// <summary>大阪</summary>
      Osaka,
      /// <summary></summary>
      Oslo,
      /// <summary></summary>
      PanamaCity,
      /// <summary></summary>
      Paramaribo,
      /// <summary>パリ</summary>
      Paris,
      /// <summary></summary>
      Perth,
      /// <summary></summary>
      Plymouth,
      /// <summary></summary>
      PortMoresby,
      /// <summary></summary>
      Prague,
      /// <summary></summary>
      Reykjavík,
      /// <summary></summary>
      RioDeJaneiro,
      /// <summary>ローマ</summary>
      Rome,
      /// <summary></summary>
      Salvador,
      /// <summary></summary>
      Santiago,
      /// <summary></summary>
      StPetersburg,
      /// <summary></summary>
      SaoPaulo,
      /// <summary></summary>
      Shanghai,
      /// <summary>シンガポール</summary>
      Singapore,
      /// <summary></summary>
      Sofia,
      /// <summary>ストックホルム</summary>
      Stockholm,
      /// <summary>シドニー</summary>
      Sydney,
      /// <summary></summary>
      Tananarive,
      /// <summary>東京</summary>
      Tokyo,
      /// <summary></summary>
      Tripoli,
      /// <summary>ヴェネチア</summary>
      Venice,
      /// <summary></summary>
      Veracruz,
      /// <summary></summary>
      Vienna,
      /// <summary></summary>
      Vladivostok,
      /// <summary></summary>
      Warsaw,
      /// <summary></summary>
      Wellington,
      /// <summary>チューリッヒ</summary>
      Zurich
    }

    #endregion

    #region クラス変数

    /// <summary>都市-位置対応リスト</summary>
    private static Dictionary<City, double[]> cities = new Dictionary<City, double[]>();

    #endregion

    #region インスタンス変数

    /// <summary>法線面直達日射量[W/m2]</summary>
    private double directNormalRadiation;

    /// <summary>水平面天空日射量[W/m2]</summary>
    private double diffuseHorizontalRadiation;

    /// <summary>水平面全天日射量[W/m2]</summary>
    private double globalHorizontalRadiation;

    #endregion

    #region プロパティ

    /// <summary>太陽高度[radian]を取得する</summary>
    public double Altitude { private set; get; }

    /// <summary>太陽方位角[radian]を取得する</summary>
    public double Orientation { private set; get; }

    /// <summary>計算地点の緯度（北が正）[degree]を取得する</summary>
    public double Latitude { private set; get; }

    /// <summary>計算地点の経度（東が正）[degree]を取得する</summary>
    public double Longitude { private set; get; }

    /// <summary>標準時を規定する地点の経度（東が正）[degree]を取得する</summary>
    public double StandardLongitude { private set; get; }

    /// <summary>法線面直達日射量[W/m2]を設定・取得する</summary>
    public double DirectNormalRadiation
    {
      get { return directNormalRadiation; }
      set { directNormalRadiation = Math.Max(0, value); }
    }

    /// <summary>水平面天空（散乱）日射量[W/m2]を設定・取得する</summary>
    public double DiffuseHorizontalRadiation
    {
      get { return diffuseHorizontalRadiation; }
      set { diffuseHorizontalRadiation = Math.Max(0, value); }
    }

    /// <summary>水平面全天日射量[W/m2]を設定・取得する</summary>
    public double GlobalHorizontalRadiation
    {
      get { return globalHorizontalRadiation; }
      set { globalHorizontalRadiation = Math.Max(0, value); }
    }

    /// <summary>現在の日時を取得する</summary>
    public DateTime CurrentDateTime { private set; get; }

    #endregion

    #region コンストラクタ

    /// <summary>静的コンストラクタ</summary>
    static Sun()
    {
      cities.Add(City.Aberdeen, new double[] { 57.15, -2.15, 0 });
      cities.Add(City.Algiers, new double[] { 36.83, 3, 15 });
      cities.Add(City.Amsterdam, new double[] { 52.37, 4.88, 15 });
      cities.Add(City.Ankara, new double[] { 39.92, 32.92, 30 });
      cities.Add(City.Asuncion, new double[] { -25.25, -57.67, -60 });
      cities.Add(City.Athens, new double[] { 37.97, 23.72, 30 });
      cities.Add(City.Auckland, new double[] { -36.87, 174.75, 180 });
      cities.Add(City.Bangkok, new double[] { 13.75, 100.5, -75 });
      cities.Add(City.Barcelona, new double[] { 41.38, 2.15, 15 });
      cities.Add(City.Beijing, new double[] { 39.92, 116.42, 120 });
      cities.Add(City.Belem, new double[] { -1.47, -48.48, -45 });
      cities.Add(City.Belfast, new double[] { 54.62, -5.93, 0 });
      cities.Add(City.Belgrade, new double[] { 44.87, 20.53, 15 });
      cities.Add(City.Berlin, new double[] { 52.5, 13.42, 15 });
      cities.Add(City.Birmingham, new double[] { 52.42, -1.92, 0 });
      cities.Add(City.Bogota, new double[] { 4.53, -74.25, -75 });
      cities.Add(City.Bordeaux, new double[] { 44.83, -0.52, 15 });
      cities.Add(City.Bremen, new double[] { 53.08, 8.82, 15 });
      cities.Add(City.Brisbane, new double[] { -27.48, 153.13, 150 });
      cities.Add(City.Bristol, new double[] { 51.47, -2.58, 0 });
      cities.Add(City.Brussels, new double[] { 50.87, 4.37, 15 });
      cities.Add(City.Bucharest, new double[] { 44.42, 26.12, 30 });
      cities.Add(City.Budapest, new double[] { 47.5, 19.08, 15 });
      cities.Add(City.BuenosAires, new double[] { -34.58, -58.37, -45 });
      cities.Add(City.Cairo, new double[] { 30.03, 31.35, 30 });
      cities.Add(City.Canton, new double[] { 23.12, 113.25, 120 });
      cities.Add(City.CapeTown, new double[] { -33.92, 18.37, 30 });
      cities.Add(City.Caracas, new double[] { 10.47, -67.03, -60 });
      cities.Add(City.Cayenne, new double[] { 4.82, -52.3, -45 });
      cities.Add(City.Chihuahua, new double[] { 28.62, -106.08, -105 });
      cities.Add(City.Chongqing, new double[] { 29.77, 106.57, 120 });
      cities.Add(City.Copenhagen, new double[] { 55.67, 12.57, 15 });
      cities.Add(City.Cordoba, new double[] { -31.47, -64.17, -45 });
      cities.Add(City.Dakar, new double[] { 14.67, -17.47, 0 });
      cities.Add(City.Djibouti, new double[] { 11.5, 43.05, 45 });
      cities.Add(City.Dublin, new double[] { 53.33, -6.25, 0 });
      cities.Add(City.Durban, new double[] { -29.88, 30.88, 30 });
      cities.Add(City.Edinburgh, new double[] { 55.92, -3.17, 0 });
      cities.Add(City.Frankfurt, new double[] { 50.12, 8.68, 15 });
      cities.Add(City.Georgetown, new double[] { 6.75, -58.25, -60 });
      cities.Add(City.Glasgow, new double[] { 55.83, -4.25, 0 });
      cities.Add(City.GuatemalaCity, new double[] { 14.62, -90.52, -90 });
      cities.Add(City.Guayaquil, new double[] { -2.17, -79.93, -75 });
      cities.Add(City.Hamburg, new double[] { 53.55, 10.03, 15 });
      cities.Add(City.Hammerfest, new double[] { 70.63, 23.63, 15 });
      cities.Add(City.Havana, new double[] { 23.13, -82.38, -75 });
      cities.Add(City.Helsinki, new double[] { 60.17, 25, 30 });
      cities.Add(City.Hobart, new double[] { -42.87, 147.32, 150 });
      cities.Add(City.HongKong, new double[] { 22.33, 114.18, 120 });
      cities.Add(City.Iquique, new double[] { -20.17, -70.12, -60 });
      cities.Add(City.Irkutsk, new double[] { 52.5, 104.33, 120 });
      cities.Add(City.Jakarta, new double[] { -6.27, 106.8, 105 });
      cities.Add(City.Johannesburg, new double[] { -26.2, 28.07, 30 });
      cities.Add(City.Kingston, new double[] { 17.98, -76.82, -75 });
      cities.Add(City.Kinshasa, new double[] { -4.3, 15.28, 15 });
      cities.Add(City.KualaLumpur, new double[] { 3.13, 101.7, 120 });
      cities.Add(City.LaPaz, new double[] { -16.45, -68.37, -60 });
      cities.Add(City.Leeds, new double[] { 53.75, -1.5, 0 });
      cities.Add(City.Lima, new double[] { -12, -77.03, -75 });
      cities.Add(City.Lisbon, new double[] { 38.73, -9.15, 0 });
      cities.Add(City.Liverpool, new double[] { 53.42, -3, 0 });
      cities.Add(City.London, new double[] { 51.53, -0.08, 0 });
      cities.Add(City.Lyons, new double[] { 45.75, 4.83, 15 });
      cities.Add(City.Madrid, new double[] { 40.43, -3.7, 15 });
      cities.Add(City.Manchester, new double[] { 53.5, -2.25, 0 });
      cities.Add(City.Manila, new double[] { 14.58, 120.95, 120 });
      cities.Add(City.Marseilles, new double[] { 43.33, 5.33, 15 });
      cities.Add(City.Mazatlan, new double[] { 23.2, -106.42, -105 });
      cities.Add(City.Mecca, new double[] { 21.48, 39.75, 45 });
      cities.Add(City.Melbourne, new double[] { -37.78, 144.97, 150 });
      cities.Add(City.MexicoCity, new double[] { 19.43, -99.12, -90 });
      cities.Add(City.Milan, new double[] { 45.45, 9.17, 15 });
      cities.Add(City.Montevideo, new double[] { -34.88, -56.17, -45 });
      cities.Add(City.Moscow, new double[] { 55.75, 37.6, 45 });
      cities.Add(City.Munich, new double[] { 48.13, 11.58, 15 });
      cities.Add(City.Nagasaki, new double[] { 32.8, 129.95, 135 });
      cities.Add(City.Nagoya, new double[] { 35.12, 136.93, 135 });
      cities.Add(City.Nairobi, new double[] { -1.42, 36.92, 45 });
      cities.Add(City.Nanjing, new double[] { 32.05, 118.88, 120 });
      cities.Add(City.Naples, new double[] { 40.83, 14.25, 15 });
      cities.Add(City.NewcastleOnTyne, new double[] { 54.97, -1.62, 0 });
      cities.Add(City.Odessa, new double[] { 46.45, 30.8, 30 });
      cities.Add(City.Osaka, new double[] { 34.53, 135.5, 135 });
      cities.Add(City.Oslo, new double[] { 59.95, 10.7, 15 });
      cities.Add(City.PanamaCity, new double[] { 8.97, -79.53, -75 });
      cities.Add(City.Paramaribo, new double[] { 5.75, -55.25, -45 });
      cities.Add(City.Paris, new double[] { 48.8, 2.33, 15 });
      cities.Add(City.Perth, new double[] { -31.95, 115.87, 120 });
      cities.Add(City.Plymouth, new double[] { 50.42, -4.08, 0 });
      cities.Add(City.PortMoresby, new double[] { -9.42, 147.13, 150 });
      cities.Add(City.Prague, new double[] { 50.08, 14.43, 15 });
      cities.Add(City.Reykjavík, new double[] { 64.07, -21.97, 0 });
      cities.Add(City.RioDeJaneiro, new double[] { -22.95, -43.2, -45 });
      cities.Add(City.Rome, new double[] { 41.9, 12.45, 15 });
      cities.Add(City.Salvador, new double[] { -12.93, -38.45, -45 });
      cities.Add(City.Santiago, new double[] { -33.47, -70.75, -60 });
      cities.Add(City.StPetersburg, new double[] { 59.93, 30.3, 45 });
      cities.Add(City.SaoPaulo, new double[] { -23.52, -46.52, -45 });
      cities.Add(City.Shanghai, new double[] { 31.17, 121.47, 120 });
      cities.Add(City.Singapore, new double[] { 1.23, 103.92, 120 });
      cities.Add(City.Sofia, new double[] { 42.67, 23.33, 30 });
      cities.Add(City.Stockholm, new double[] { 59.28, 18.05, 15 });
      cities.Add(City.Sydney, new double[] { -34, 151, 150 });
      cities.Add(City.Tananarive, new double[] { -18.83, 47.55, 45 });
      cities.Add(City.Tokyo, new double[] { 35.67, 139.75, 135 });
      cities.Add(City.Tripoli, new double[] { 32.95, 13.2, 30 });
      cities.Add(City.Venice, new double[] { 45.43, 12.33, 15 });
      cities.Add(City.Veracruz, new double[] { 19.17, -96.17, -90 });
      cities.Add(City.Vienna, new double[] { 48.23, 16.33, 15 });
      cities.Add(City.Vladivostok, new double[] { 43.17, 132, 150 });
      cities.Add(City.Warsaw, new double[] { 52.23, 21, 15 });
      cities.Add(City.Wellington, new double[] { -41.28, 174.78, 180 });
      cities.Add(City.Zurich, new double[] { 47.35, 8.52, 15 });
    }

    /// <summary>コンストラクタ</summary>
    /// <param name="latitude">計算地点の緯度（北が正）[degree]</param>
    /// <param name="longitude">計算地点の経度（東が正）[degree]</param>
    /// <param name="standardLongitude">標準時を規定する地点の経度[degree]</param>
    public Sun(double latitude, double longitude, double standardLongitude)
    {
      Latitude = latitude;
      Longitude = longitude;
      StandardLongitude = standardLongitude;
    }

    /// <summary>コンストラクタ</summary>
    /// <param name="latitudeDeg">計算地点の緯度（北が正）[degree]</param>
    /// <param name="latitudeMin">計算地点の緯度（北が正）[minute]</param>
    /// <param name="latitudeSec">計算地点の緯度（北が正）[second]</param>
    /// <param name="longitudeDeg">計算地点の経度（東が正）[degree]</param>
    /// <param name="longitudeMin">計算地点の経度（東が正）[minute]</param>
    /// <param name="longitudeSec">計算地点の経度（東が正）[second]</param>
    /// <param name="standardLongitudeDeg">標準時を規定する地点の経度[degree]</param>
    /// <param name="standardLongitudeMin">標準時を規定する地点の経度[minute]</param>
    /// <param name="standardLongitudeSec">標準時を規定する地点の経度[second]</param>
    public Sun(double latitudeDeg, double latitudeMin, double latitudeSec,
      double longitudeDeg, double longitudeMin, double longitudeSec,
      double standardLongitudeDeg, double standardLongitudeMin, double standardLongitudeSec)
    {
      Latitude = latitudeDeg + latitudeMin / 60d + latitudeSec / 3600d;
      Longitude = longitudeDeg + longitudeMin / 60d + longitudeSec / 3600d;
      StandardLongitude = standardLongitudeDeg + standardLongitudeMin / 60d + standardLongitudeSec / 3600d; 
    }

    /// <summary>コンストラクタ</summary>
    /// <param name="city">都市</param>
    public Sun(City city)
    {
      double[] loc = cities[city];
      Latitude = loc[0];
      Longitude = loc[1];
      StandardLongitude = loc[2];
    }

    #endregion

    #region publicメソッド

    /// <summary>太陽位置等を更新する</summary>
    /// <param name="dateTime">日時</param>
    public void Update(DateTime dateTime)
    {
      CurrentDateTime = dateTime;
      double al, or;
      GetSunPosition(Latitude, Longitude, StandardLongitude, dateTime, out al, out or);
      Altitude = al;
      Orientation = or;
    }

    /// <summary>大気圏外日射量[W/m2]を計算する</summary>
    /// <returns>大気圏外日射量[W/m2]</returns>
    public double GetExtraterrestrialRadiation()
    {
      return GetExtraterrestrialRadiation(CurrentDateTime.DayOfYear);
    }

    /// <summary>法線面直達日射[W/m2]を計算する</summary>
    /// <param name="atmosphericTransmissivity">大気透過率[-]</param>
    /// <returns>法線面直達日射[W/m2]</returns>
    public double GetDirectNormalRadiation
      (double atmosphericTransmissivity)
    {
      return GetDirectNormalRadiation
        (Altitude, atmosphericTransmissivity, CurrentDateTime.DayOfYear);
    }

    /// <summary>直散分離を行う</summary>
    /// <param name="method">直散分離の手法</param>
    /// <param name="globalHorizontalRadiation">水平面全天日射[W/m2]</param>
    public void SeparateGlobalHorizontalRadiation
      (double globalHorizontalRadiation, SeparationMethod method)
    {
      GlobalHorizontalRadiation = globalHorizontalRadiation;
      SeparateGlobalHorizontalRadiation(globalHorizontalRadiation,
        Latitude, Longitude, StandardLongitude, CurrentDateTime, method,
        out directNormalRadiation, out diffuseHorizontalRadiation);
    }

    /// <summary>法線面直達日射[W/m2]を計算して設定する</summary>
    /// <param name="globalHorizontalRadiation">全天日射[W/m2]</param>
    /// <param name="diffuseHorizontalRadiation">天空日射[W/m2]</param>
    public void SetDirectNormalRadiation
      (double globalHorizontalRadiation, double diffuseHorizontalRadiation)
    {
      GlobalHorizontalRadiation = globalHorizontalRadiation;
      DiffuseHorizontalRadiation = diffuseHorizontalRadiation;
      DirectNormalRadiation = GetDirectNormalRadiation
        (GlobalHorizontalRadiation, DiffuseHorizontalRadiation, Altitude);
    }

    /// <summary>天空日射[W/m2]を計算して設定する</summary>
    /// <param name="directNormalRadiation">法線面直達日射[W/m2]</param>
    /// <param name="globalHorizontalRadiation">全天日射[W/m2]</param>
    public void SetDiffuseHorizontalRadiation
      (double directNormalRadiation, double globalHorizontalRadiation)
    {
      GlobalHorizontalRadiation = globalHorizontalRadiation;
      DirectNormalRadiation = directNormalRadiation;
      DiffuseHorizontalRadiation = GetDiffuseHorizontalRadiation
        (DirectNormalRadiation, GlobalHorizontalRadiation, Altitude);
    }

    /// <summary>全天日射[W/m2]を計算して設定する</summary>
    /// <param name="diffuseHorizontalRadiation">天空日射[W/m2]</param>
    /// <param name="directNormalRadiation">法線面直達日射[W/m2]</param>
    public void SetGlobalHorizontalRadiation
      (double diffuseHorizontalRadiation, double directNormalRadiation)
    {
      DirectNormalRadiation = directNormalRadiation;
      DiffuseHorizontalRadiation = diffuseHorizontalRadiation;
      GlobalHorizontalRadiation = GetGlobalHorizontalRadiation
        (DiffuseHorizontalRadiation, DirectNormalRadiation, Altitude);
    }

    /// <summary>日の出時刻を計算する</summary>
    /// <returns>日の出時刻</returns>
    public DateTime GetSunRiseTime()
    {
      return GetSunRiseTime(Latitude, Longitude, StandardLongitude, CurrentDateTime);
    }

    /// <summary>日没時刻を計算する</summary>
    /// <returns>日没時刻</returns>
    public DateTime GetSunSetTime()
    {
      return GetSunSetTime(Latitude, Longitude, StandardLongitude, CurrentDateTime);
    }

    #endregion

    #region 太陽位置関連のstaticメソッド

    /// <summary>太陽位置を計算する</summary>
    /// <param name="latitude">計算地点の緯度（北緯が正）[degree]</param>
    /// <param name="longitude">計算地点の経度（北緯が正）[degree]</param>
    /// <param name="standardLongitude">標準時地点の経度（東経が正）[degree]</param>
    /// <param name="dTime">日時</param>
    /// <param name="altitude">太陽高度[rad]</param>
    /// <param name="orientation">太陽方位角[rad]</param>
    /// <example>
    /// 東京における太陽高度および方位の計算法を示す。
    /// ただし、東京は北緯35.7°、東経139.8°
    /// 日本標準時を決定する明石市は東経135°に位置する
    /// <code>
    /// double al, az;
    /// Sun.GetPosition
    /// (35.7, 139.8, 135, new DateTime(2004, 6, 22, 12, 0, 0, 0), out al, out az);
    /// </code>
    /// </example>
    public static void GetSunPosition
      (double latitude, double longitude, double standardLongitude, DateTime dTime,
      out double altitude, out double orientation)
    {
      //緯度をRadianに変換
      double phi = DEG_TO_RAD * latitude;

      double b = (360d * (dTime.DayOfYear - 81) / 365d) * DEG_TO_RAD;
      double sd = 0.397949 * Math.Sin(b);
      double cd = Math.Sqrt(1 - sd * sd);
      double e = 0.1645 * Math.Sin(2 * b) - 0.1255 * Math.Cos(b) - 0.025 * Math.Sin(b);
      double taum = dTime.Hour + dTime.Minute / 60d + dTime.Second / 3600d
          + e + (longitude - standardLongitude) / 15d;
      double t = ((taum - 12) * 15) * DEG_TO_RAD;

      //Sin太陽高度を計算
      double sp = Math.Sin(phi);
      double cp = Math.Cos(phi);
      double sh = sp * sd + cp * cd * Math.Cos(t);

      //日の出前、日没後の場合
      if (sh <= 0)
      {
        altitude = 0;
        orientation = 0;
      }
      else
      {
        altitude = Math.Asin(sh);
        double ch = Math.Sqrt(1.0 - sh * sh);
        double ca = (sh * sp - sd) / (ch * cp);
        orientation = Math.Acos(ca);
        orientation *= Math.Sign(t);
      }
    }

    /// <summary>太陽高度[radian]を計算する</summary>
    /// <param name="latitude">緯度[degree]</param>
    /// <param name="longitude">計算地点の経度[degree]</param>
    /// <param name="standardLongitude">標準時地点の経度（東経が正）[degree]</param>
    /// <param name="dTime">日時</param>
    /// <returns>太陽高度[radian]</returns>
    public static double GetSunAltitude
      (double latitude, double longitude, double standardLongitude, DateTime dTime)
    {
      double altitude, azimuth;
      GetSunPosition
        (latitude, longitude, standardLongitude, dTime, out altitude, out azimuth);
      return altitude;
    }

    /// <summary>太陽方位角[radian]を計算する</summary>
    /// <param name="latitude">緯度[degree]</param>
    /// <param name="longitude">計算地点の経度[degree]</param>
    /// <param name="standardLongitude">標準時地点の経度（東経が正）[degree]</param>
    /// <param name="dTime">日時</param>
    /// <returns>太陽方位角[radian]</returns>
    public static double GetSunAzimuth
      (double latitude, double longitude, double standardLongitude, DateTime dTime)
    {
      double altitude, azimuth;
      GetSunPosition
        (latitude, longitude, standardLongitude, dTime, out altitude, out azimuth);
      return azimuth;
    }

    /// <summary>日没の時刻を求める</summary>
    /// <param name="latitude">計算地点の緯度[degree]</param>
    /// <param name="longitude">計算地点の経度[degree]</param>
    /// <param name="standardLongitude">標準時地点の経度（東経が正）[degree]</param>
    /// <param name="dTime">日時</param>
    /// <returns>日没の時刻</returns>
    public static DateTime GetSunSetTime
      (double latitude, double longitude, double standardLongitude, DateTime dTime)
    {
      return getSunRiseAndSunSetTime(latitude, longitude, standardLongitude, dTime, true);
    }

    /// <summary>日の出の時刻を求める</summary>
    /// <param name="latitude">計算地点の緯度[degree]</param>
    /// <param name="longitude">計算地点の経度[degree]</param>
    /// <param name="standardLongitude">標準時地点の経度（東経が正）[degree]</param>
    /// <param name="dTime">日時</param>
    /// <returns>日の出の時刻</returns>
    public static DateTime GetSunRiseTime
      (double latitude, double longitude, double standardLongitude, DateTime dTime)
    {
      return getSunRiseAndSunSetTime(latitude, longitude, standardLongitude, dTime, false);
    }

    /// <summary>日の出または日没時刻を求める</summary>
    /// <param name="latitude">計算地点の緯度[degree]</param>
    /// <param name="longitude">計算地点の経度[degree]</param>
    /// <param name="standardLongitude">標準時地点の経度（東経が正）[degree]</param>
    /// <param name="dTime">日時</param>
    /// <param name="isSunSet">計算対象は日没か否か</param>
    /// <returns>日の出の時刻</returns>
    private static DateTime getSunRiseAndSunSetTime
      (double latitude, double longitude, double standardLongitude,
      DateTime dTime, bool isSunSet)
    {
      //緯度をRadianに変換
      double phi = DEG_TO_RAD * latitude;

      double b = (360d * (dTime.DayOfYear - 81) / 365d) * DEG_TO_RAD;
      double sd = 0.397949 * Math.Sin(b);
      double tptd = -Math.Tan(phi) * Math.Tan(Math.Asin(sd));
      //白夜・極夜の場合
      if (tptd < -1 || 1 < tptd)
        return new DateTime(dTime.Year, dTime.Month, dTime.Day, 0, 0, 0);

      double cd = Math.Sqrt(1 - sd * sd);
      double e = 0.1645 * Math.Sin(2 * b) - 0.1255 * Math.Cos(b) - 0.025 * Math.Sin(b);
      double tsr = -Math.Acos(tptd) / DEG_TO_RAD;
      if (isSunSet) tsr *= -1;
      double tstsr = (tsr - longitude + standardLongitude) / 15 + 12 - e;

      int hour = (int)Math.Truncate(tstsr);
      tstsr = 60 * (tstsr - hour);
      int minute = (int)Math.Truncate(tstsr);
      tstsr = 60 * (tstsr - minute);
      int sec = (int)Math.Truncate(tstsr);
      return new DateTime(dTime.Year, dTime.Month, dTime.Day, hour, minute, sec);
    }



    /// <summary>太陽赤緯[degree]を返す</summary>
    /// <param name="dTime">日時</param>
    /// <returns>太陽赤緯[degree]</returns>
    public static double GetSunDeclination(DateTime dTime)
    {
      double dDeg = 2d * Math.PI * dTime.DayOfYear / 365d;
      return DEG_TO_RAD * (0.3622133 - 23.24763 * Math.Cos(dDeg + 0.153231) - 0.3368908 *
          Math.Cos(2.0 * dDeg + 0.2070988) - 0.1852646 * Math.Cos(3.0 * dDeg + 0.6201293));
    }

    /// <summary>均時差を返す</summary>
    /// <param name="dTime">日時</param>
    /// <returns>均時差</returns>
    public static double GetEquationOfTime(DateTime dTime)
    {
      double dDeg = 2d * Math.PI * dTime.DayOfYear / 365d;
      return 60.0 * (-0.0002786409 + 0.1227715 * Math.Cos(dDeg + 1.498311) - 0.1654575 *
          Math.Cos(2.0 * dDeg - 1.261546) - 0.00535383 * Math.Cos(3.0 * dDeg - 1.1571));
    }

    /// <summary>時角[°]を返す</summary>
    /// <param name="equationOfTime">均時差</param>
    /// <param name="longitude">計算地点の経度[degree]</param>
    /// <param name="standardLongitude">標準時地点の経度（東経で正）[degree]</param>
    /// <param name="dTime">日時</param>
    /// <returns>時角[°]</returns>
    /// <remarks>宿谷：数値計算で学ぶ光と熱の建築環境学 p.20</remarks>
    public static double GetHourAngle
      (double equationOfTime, double longitude, double standardLongitude, DateTime dTime)
    {
      double ts = (double)dTime.Hour + (double)dTime.Minute / 60.0d;
      return (15.0 * (ts - 12.0) + longitude - standardLongitude + 0.25 * equationOfTime) * DEG_TO_RAD;
    }

    #endregion

    #region 日射相互変換に関するstaticメソッド

    /// <summary>法線面直達日射[W/m2]を計算する</summary>
    /// <param name="globalHorizontalRadiation">全天日射[W/m2]</param>
    /// <param name="diffuseHorizontalRadiation">天空日射[W/m2]</param>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <returns>法線面直達日射[W/m2]</returns>
    public static double GetDirectNormalRadiation
      (double globalHorizontalRadiation, double diffuseHorizontalRadiation, double altitude)
    {
      return (globalHorizontalRadiation - diffuseHorizontalRadiation) / Math.Sin(altitude);
    }

    /// <summary>天空日射[W/m2]を計算する</summary>
    /// <param name="directNormalRadiation">法線面直達日射[W/m2]</param>
    /// <param name="globalHorizontalRadiation">全天日射[W/m2]</param>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <returns>天空日射[W/m2]</returns>
    public static double GetDiffuseHorizontalRadiation
      (double directNormalRadiation, double globalHorizontalRadiation, double altitude)
    {
      return globalHorizontalRadiation - Math.Sin(altitude) * directNormalRadiation;
    }

    /// <summary>全天日射[W/m2]を計算する</summary>
    /// <param name="diffuseHorizontalRadiation">天空日射[W/m2]</param>
    /// <param name="directNormalRadiation">法線面直達日射[W/m2]</param>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <returns>全天日射[W/m2]</returns>
    public static double GetGlobalHorizontalRadiation
      (double diffuseHorizontalRadiation, double directNormalRadiation, double altitude)
    {
      return directNormalRadiation * Math.Sin(altitude) + diffuseHorizontalRadiation;
    }

    #endregion

    #region 直散分離に関するstaticメソッド

    /// <summary>水平面全天日射[W/m2]の直散分離を行う</summary>
    /// <param name="globalHorizontalRadiation">水平面全天日射[W/m2]</param>
    /// <param name="latitude">緯度[degree]</param>
    /// <param name="longitude">計算地点の経度[degree]</param>
    /// <param name="standardLongitude">標準時地点の経度（東経で正）[degree]</param>
    /// <param name="dTime">日時</param>
    /// <param name="directSolarRadiation">法線面直達日射[W/m2]</param>
    /// <param name="method">直散分離の手法</param>
    /// <param name="diffuseHorizontalRadiation">天空日射[W/m2]</param>
    public static void SeparateGlobalHorizontalRadiation
      (double globalHorizontalRadiation, double latitude, double longitude,
      double standardLongitude, DateTime dTime, SeparationMethod method,
      out double directSolarRadiation, out double diffuseHorizontalRadiation)
    {
      //太陽高度の正弦の計算
      double h = GetSunAltitude(latitude, longitude, standardLongitude, dTime);
      if (h <= 0 || globalHorizontalRadiation <= 0)
      {
        directSolarRadiation = diffuseHorizontalRadiation = 0;
        return;
      }
      h = Math.Max(0.05, h);  //誤差が拡大するため3度以上とする
      double sinH = Math.Sin(h);

      //大気圏外日射[w/m2]の計算
      double io = GetExtraterrestrialRadiation(dTime.DayOfYear);
      double dn, dff;
      //宇田川の手法
      if (method == SeparationMethod.Udagawa)
      {
        double ktt = globalHorizontalRadiation / (io * sinH);
        double ktc = io * sinH * (0.5163 + sinH * (0.333 + 0.00803 * sinH));
        if (ktc <= globalHorizontalRadiation) dn = (-0.43 + 1.43 * ktt) * io;
        else dn = (2.277 + sinH * (-1.258 + 0.2396 * sinH)) * Math.Pow(ktt, 3) * io;
        dff = globalHorizontalRadiation - dn * sinH;
      }
      //Erbsの手法
      else if (method == SeparationMethod.Erbs)
      {
        double ktt = globalHorizontalRadiation / (io * sinH);
        if (ktt < 0.22) dff = globalHorizontalRadiation * (1.0 - 0.09 * ktt);
        else if (ktt < 0.8) dff = globalHorizontalRadiation * (0.9511 + ktt * (-0.1604
          + ktt * (4.388 + ktt * (-16.638 + 12.336 * ktt))));
        else dff = 0.1651 * globalHorizontalRadiation;
        dn = (globalHorizontalRadiation - dff) / sinH;
      }
      //三木の手法
      else if (method == SeparationMethod.Miki)
      {
        double lkt = Math.Min(1, globalHorizontalRadiation / (io * sinH));
        double skt = (lkt - 0.15 - 0.2 * sinH) / 0.6;
        double skd;
        if (skt <= 0) skd = 0;
        else skd = skt * skt * (3 - 2 * skt);
        double lkd = Math.Min(skd * lkt, Math.Pow(0.8, (7 + sinH) / (1 + 7 * sinH)));
        double lks = Math.Max(lkt - lkd, 0.005);
        dn = lkd * io;
        dff = globalHorizontalRadiation - dn * sinH;
      }
      //数値計算による方法
      else
      {
        //大気透過率=0で観測値を上回る場合
        getDirectAndDiffuseRadiation(0, sinH, io, method, out dn, out dff);
        if (globalHorizontalRadiation < dn * sinH + dff)
        {
          directSolarRadiation = 0;
          diffuseHorizontalRadiation = globalHorizontalRadiation;
          return;
        }

        //大気透過率=1で観測値を下回る場合
        getDirectAndDiffuseRadiation(1, sinH, io, method, out dn, out dff);
        if (dn * sinH + dff < globalHorizontalRadiation)
        {
          double rate = globalHorizontalRadiation / (dn * sinH + dff);
          directSolarRadiation = dn * sinH * rate;
          diffuseHorizontalRadiation = dff * rate;
          return;
        }

        //大気透過率を二分法で収束計算
        //誤差関数を定義
        Roots.ErrorFunction eFnc = delegate(double atmTrans)
        {
          getDirectAndDiffuseRadiation(atmTrans, sinH, io, method, out dn, out dff);
          return globalHorizontalRadiation - (dn * sinH + dff);
        };
        Roots.Bisection(0, 1, 0.001, eFnc);
      }
      //0以上とする
      directSolarRadiation = Math.Max(0, dn);
      diffuseHorizontalRadiation = Math.Max(0, dff);
    }

    /// <summary>大気圏外日射[W/m2]を計算する</summary>
    /// <param name="daysOfYear">通日(1月1日=1, 12月31日=365)</param>
    /// <returns>大気圏外日射[W/m2]</returns>
    public static double GetExtraterrestrialRadiation(int daysOfYear)
    {
      return SOLAR_CONSTANT * (1d + 0.033 * Math.Cos(2d * Math.PI * daysOfYear / 365d));
    }

    /// <summary>法線面直達日射[W/m2]と水平面天空日射[W/m2]を計算する</summary>
    /// <param name="aTransmissivity">大気透過率[-]</param>
    /// <param name="sinAltitude">太陽高度の正弦</param>
    /// <param name="exRadiation">大気圏外日射[W/m2]</param>
    /// <param name="method">直散分離手法</param>
    /// <param name="directNormalRadiation">法線面直達日射[W/m2]</param>
    /// <param name="diffuseHorizontalRadiation">水平面天空日射[W/m2]</param>
    private static void getDirectAndDiffuseRadiation
      (double aTransmissivity, double sinAltitude, double exRadiation,
      SeparationMethod method,
      out double directNormalRadiation, out double diffuseHorizontalRadiation)
    {
      double ps = Math.Pow(aTransmissivity, 1 / sinAltitude);
      directNormalRadiation = exRadiation * ps;
      double shi = sinAltitude * (exRadiation - directNormalRadiation);
      diffuseHorizontalRadiation = 0;

      switch (method)
      {
        case SeparationMethod.Akasaka:
          diffuseHorizontalRadiation =
            shi * 0.95 * Math.Pow(aTransmissivity, 1 / (0.5 + 2.5 * sinAltitude)) *
            Math.Pow(1 - aTransmissivity, 2d / 3d);
          break;
        case SeparationMethod.Berlage:
          diffuseHorizontalRadiation = shi * 0.5 / (1 - 1.4 * Math.Log(aTransmissivity));
          break;
        case SeparationMethod.LiuAndJordan:
          diffuseHorizontalRadiation = sinAltitude * exRadiation * (0.271 - 0.2939 * ps);
          break;
        case SeparationMethod.Matsuo:
          diffuseHorizontalRadiation =
            shi * (1 - aTransmissivity) * 1.2 / (1 - 1.4 * Math.Log(aTransmissivity));
          break;
        case SeparationMethod.Nagata:
          diffuseHorizontalRadiation = shi * (0.66 - 0.32 * sinAltitude) *
            (0.5 + (0.4 - 0.3 * aTransmissivity) * sinAltitude);
          break;
        case SeparationMethod.Watanabe:
          double qq = (0.9013 + 1.123 * sinAltitude) * Math.Pow
            (aTransmissivity, 0.489 / sinAltitude) * Math.Pow(1 - ps, 2.525);
          diffuseHorizontalRadiation = sinAltitude * exRadiation * qq / (1 + qq);
          break;
      }
    }

    /// <summary>法線面直達日射[W/m2]を計算する</summary>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <param name="atmosphericTransmissivity">大気透過率[-]</param>
    /// <param name="dayOfYear">通日(1月1日=1, 12月31日=365)</param>
    /// <returns>法線面直達日射[W/m2]</returns>
    public static double GetDirectNormalRadiation
      (double altitude, double atmosphericTransmissivity, int dayOfYear)
    {
      double sinh = Math.Sin(altitude);
      if (sinh <= 0) return 0;
      double exrad = GetExtraterrestrialRadiation(dayOfYear);
      return exrad * Math.Pow(atmosphericTransmissivity, 1 / sinh);
    }

    #endregion

  }

  #region 読み取り専用の太陽interface

  /// <summary>読み取り専用の太陽</summary>
  public interface ImmutableSun
  {
    /// <summary>太陽高度[radian]を取得する</summary>
    double Altitude { get; }

    /// <summary>太陽方位角[radian]を取得する</summary>
    double Orientation { get; }

    /// <summary>法線面直達日射量[W/m2]を取得する</summary>
    double DirectNormalRadiation { get; }

    /// <summary>水平面天空（散乱）日射量[W/m2]を取得する</summary>
    double DiffuseHorizontalRadiation { get; }

    /// <summary>水平面全天日射量[W/m2]を取得する</summary>
    double GlobalHorizontalRadiation { get; }

    /// <summary>計算地点の緯度（北が正）[degree]を取得する</summary>
    double Latitude { get; }

    /// <summary>計算地点の経度（東が正）[degree]を取得する</summary>
    double Longitude { get; }

    /// <summary>標準時を規定する地点の経度（東が正）[degree]を取得する</summary>
    double StandardLongitude { get; }

    /// <summary>現在の日時を取得する</summary>
    DateTime CurrentDateTime { get; }

    /// <summary>大気圏外日射量[W/m2]を計算する</summary>
    /// <returns>大気圏外日射量[W/m2]</returns>
    double GetExtraterrestrialRadiation();
  }

  #endregion

}
