/* Water.cs
 * 
 * Copyright (C) 2007 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.ThermophysicalProperty
{
    /// <summary>水の物性値を計算するためのstaticメソッドを提供するクラス</summary>
    /// <remarks>
    /// HVACSIM+(J)より移植
    /// 1)  Irvine, T.F.Jr., and Liley, P.E., "Steam and Gas Tables
    ///     with Computer Equations," Academic Press, Inc., 1984.
    /// 2)  Van Wylen, G.J., and Sonntag, R.E., Fundamentals of Classical
    ///     Thermodynamics (SI Version).  New York: John Wiley and Sons.
    /// 3)  Chapman, A.J.  Heat Transfer (3rd Edition).  New York:
    ///     Macmillan  Publishing Co. (1974).
    /// 4)  Chi, J.
    /// 5)  Karlekar, B.V., and Desmond, R.M., Engineering Heat Transfer.
    ///     St. Paul: West Publishing Co., 1977.
    /// 6)  CRC Handbook of Chemistry and Physics, 61st Ed. (1980-1981).
    /// </remarks>
    public static class Water
    {

        #region 定数宣言

        /// <summary>絶対温度と摂氏との変換定数</summary>
        private const double CONVERT_C_TO_K = 273.15;

        /// <summary>臨界点での比エンタルピー[kJ/kg]</summary>
        private const double TEMPERATURE_AT_CRITICAL_POINT = 647.096;

        /// <summary>臨界点での比エンタルピー[kJ/kg]</summary>
        private const double ENTHALPY_AT_CRITICAL_POINT = 2099.3;

        /// <summary>臨界点での比体積[m3/kg]</summary>
        private const double SPECIFIC_VOLUME_AT_CRITICAL_POINT = 0.003155;

        /// <summary>臨界点での比エントロピー[kJ/kgK]</summary>
        private const double ENTROPY_AT_CRITICAL_POINT = 4.4289;

        /// <summary>臨界点での飽和水蒸気圧[kPa]</summary>
        private const double PRESSURE_AT_CRITICAL_POINT = 22089;

        /// <summary>三重点での蒸発潜熱[kJ/kg]</summary>
        private const double VAPORIZATION_HEAT_AT_TRIPLE_POINT = 2500.9;

        #endregion

        #region privateメソッド
        
        /// <summary>臨界温度からの離れを求める</summary>
        /// <param name="saturationTemperature">飽和温度[C]</param>
        /// <returns>臨界温度からの離れ</returns>
        private static double getTR(double saturationTemperature)
        {
            return (TEMPERATURE_AT_CRITICAL_POINT -
                (saturationTemperature + CONVERT_C_TO_K)) /
                TEMPERATURE_AT_CRITICAL_POINT;
        }

        #endregion

        #region 飽和水蒸気圧と飽和温度の計算

        /// <summary>飽和温度[℃]から飽和水蒸気圧[kPa]を求める</summary>
        /// <param name="saturationTemperature">飽和温度[℃]</param>
        /// <returns>飽和水蒸気圧[kPa]</returns>
        public static double GetSaturationPressure
            (double saturationTemperature)
        {
            const double P_CONVERT = 0.001d;

            const double C1 = -5.6745359e3d;
            const double C2 = 6.3925247d;
            const double C3 = -9.6778430e-3d;
            const double C4 = 6.2215701e-7d;
            const double C5 = 2.0747825e-9d;
            const double C6 = -9.4840240e-13d;
            const double C7 = 4.1635019d;

            const double N1 = 0.11670521452767e4d;
            const double N2 = -0.72421316703206e6d;
            const double N3 = -0.17073846940092e2d;
            const double N4 = 0.12020824702470e5d;
            const double N5 = -0.32325550322333e7d;
            const double N6 = 0.14915108613530e2d;
            const double N7 = -0.4823265731591e4d;
            const double N8 = 0.40511340542057e6d;
            const double N9 = -0.23855557567849e0d;
            const double N10 = 0.65017534844798e3d;

            double ts = saturationTemperature + CONVERT_C_TO_K;

            //-100~0.01C//三重点以下はwexler-hylandの式
            if (saturationTemperature < 0.01)
            {
                return Math.Exp(C1 / ts + C2 
                    + C3 * ts + C4 * Math.Pow(ts, 2) 
                    + C5 * Math.Pow(ts, 3) + C6 * Math.Pow(ts, 4) 
                    + C7 * Math.Log(ts)) * P_CONVERT;
            }
            //~647.096K//臨界温度まではIAPWS-IF97実用国際状態式
            else
            {
                double alpha = ts + N9 / (ts - N10);
                double a2 = alpha * alpha;
                double A = a2 + N1 * alpha + N2;
                double B = N3 * a2 + N4 * alpha + N5;
                double C = N6 * a2 + N7 * alpha + N8;
                return Math.Pow(2 * C / (-B + Math.Pow(B * B - 4 * A * C, 0.5)), 4) / P_CONVERT;
            }
        }
        
        /// <summary>飽和水蒸気圧[kPa]から飽和温度[C]を求める</summary>
        /// <param name="saturationPressure">飽和水蒸気圧[kPa]</param>
        /// <returns>飽和温度[C]</returns>
        public static double GetSaturationTemperature
            (double saturationPressure)
        {
            const double P_CONVERT = 0.001d;

            const double D1 = -6.0662e1d;
            const double D2 = 7.4624e0d;
            const double D3 = 2.0594e-1d;            
            const double D4 = 1.6321e-2d;
            
            const double N1 = 0.11670521452767e4d;
            const double N2 = -0.72421316703206e6d;
            const double N3 = -0.17073846940092e2d;
            const double N4 = 0.12020824702470e5d;
            const double N5 = -0.32325550322333e7d;
            const double N6 = 0.14915108613530e2d;
            const double N7 = -0.4823265731591e4d;
            const double N8 = 0.40511340542057e6d;
            const double N9 = -0.23855557567849e0d;
            const double N10 = 0.65017534844798e3d;

            //~0C//wexler-hylandの計算値を近似した式
            if (saturationPressure < 0.611213)
            {
                double y = Math.Log(saturationPressure / P_CONVERT);
                return D1 + y * (D2 + y * (D3 + y * D4));
            }
            //0C~//臨界圧力まではIAPWS-IF97実用国際状態式
            else
            {
                double ps = saturationPressure * P_CONVERT;
                double beta = Math.Pow(ps, 0.25);
                double b2 = beta * beta;
                double E = b2 + N3 * beta + N6;
                double F = N1 * b2 + N4 * beta + N7;
                double G = N2 * b2 + N5 * beta + N8;
                double D = 2 * G / (-F - Math.Pow(F * F - 4 * E * G, 0.5));
                return (N10 + D - Math.Pow(Math.Pow(N10 + D, 2) -
                    4 * (N9 + N10 * D), 0.5)) / 2d - CONVERT_C_TO_K;
            }
        }

        #endregion

        #region 蒸発潜熱の計算

        /// <summary>飽和温度[℃]から蒸発潜熱[kJ/kg]を求める</summary>
        /// <param name="saturationTemperature">飽和温度[℃]</param>
        /// <returns>蒸発潜熱[kJ/kg]</returns>
        public static double GetVaporizationLatentHeat
            (double saturationTemperature)
        {
            const double E1 = -3.87446;
            const double E2 = 2.94553;
            const double E3 = -8.06395;
            const double E4 = 11.5633;
            const double E5 = -6.02884;
            const double B = 0.779221;
            const double C = 4.62668;
            const double D = -1.07931;

            //0度以上とする
            saturationTemperature = Math.Max(0, saturationTemperature);

            double tr = getTR(saturationTemperature);
            if (tr < 0.0) return 0.0;
            double y = B * Math.Pow(tr, 1.0 / 3.0) +
                C * Math.Pow(tr, 5.0 / 6.0) + D * Math.Pow(tr, 0.875);
            y += tr * (E1 + tr * (E2 + tr * (E3 + tr * (E4 + tr * E5))));
            return y * VAPORIZATION_HEAT_AT_TRIPLE_POINT;
        }

        #endregion

        #region 飽和水の物性値

        /// <summary>飽和温度[℃]から飽和水の比体積[m3/kg]を求める</summary>
        /// <param name="saturatedTemperature">飽和温度[℃]</param>
        /// <returns>飽和水の比体積[m3/kg]</returns>
        public static double GetSaturatedLiquidSpecificVolume
            (double saturatedTemperature)
        {
            const double A = 1.0;
            const double B = -1.9153882;
            const double C = 12.015186;
            const double D = -7.8464025;
            const double E1 = -3.8886414;
            const double E2 = 2.0582238;
            const double E3 = -2.0829991;
            const double E4 = 0.82180004;
            const double E5 = 0.47549742;

            double tr = getTR(saturatedTemperature);
            double y = A + B * Math.Pow(tr, 1.0 / 3.0) + 
                C * Math.Pow(tr, 5.0 / 6.0) + D * Math.Pow(tr, 0.875);
            y += tr * (E1 + tr * (E2 + tr * (E3 + tr * (E4 + tr * E5))));

            return y * SPECIFIC_VOLUME_AT_CRITICAL_POINT;
        }

        /// <summary>飽和温度[℃]から飽和水のエンタルピー[kJ/kg]を求める</summary>
        /// <param name="saturatedTemperature">飽和温度[℃]</param>
        /// <returns>飽和水のエンタルピー[kJ/kg]</returns>
        public static double GetSaturatedLiquidEnthalpy
            (double saturatedTemperature)
        {
            //273.16<Ts<300の係数
            const double E11 = 624.698837;
            const double E21 = -2343.85369;
            const double E31 = -9508.12101;
            const double E41 = 71628.7928;
            const double E51 = -163535.221;
            const double E61 = 166531.093;
            const double E71 = -64785.4585;
            //300<Ts<600の係数
            const double A2 = 0.8839230108;
            const double E12 = -2.67172935;
            const double E22 = 6.22640035;
            const double E32 = -13.1789573;
            const double E42 = -1.91322436;
            const double E52 = 68.793763;
            const double E62 = -124.819906;
            const double E72 = 72.1435404;
            //600<Tsの係数
            const double A3 = 1.0;
            const double B3 = -0.441057805;
            const double C3 = -5.52255517;
            const double D3 = 6.43994847;
            const double E13 = -1.64578795;
            const double E23 = -1.30574143;

            double tk = saturatedTemperature + CONVERT_C_TO_K;
            double tr = getTR(saturatedTemperature);
            double y;
            if (tk < 300.0) y = tr * (E11 + tr * (E21 +
                tr * (E31 + tr * (E41 + tr * (E51 + tr * (E61 + tr * E71))))));
            else if (tk < 600.0) y = tr * (E12 + tr * (E22 + tr * 
                (E32 + tr * (E42 + tr * (E52 + tr * (E62 + tr * E72)))))) + A2;
            else y = A3 + B3 * Math.Pow(tr, 1.0 / 3.0) + 
                C3 * Math.Pow(tr, 5.0 / 6.0) + D3 * Math.Pow(tr, 0.875) + 
                tr * (E13 + tr * E23);

            return y * ENTHALPY_AT_CRITICAL_POINT;
        }

        /// <summary>飽和温度[℃]から飽和水のエントロピー[kJ/kgK]を求める</summary>
        /// <param name="saturatedTemperature">飽和温度[℃]</param>
        /// <returns>飽和水のエントロピー[kJ/kgK]</returns>
        public static double GetSaturatedLiquidEntropy
            (double saturatedTemperature)
        {
            //273.16<Ts<300の係数
            const double E11 = -1836.92956;
            const double E21 = 14706.6352;
            const double E31 = -43146.6046;
            const double E41 = 48606.6733;
            const double E51 = 7997.5096;
            const double E61 = -58333.9887;
            const double E71 = 33140.0718;
            //300<Ts<600の係数
            const double A2 = 0.912762917;
            const double E12 = -1.75702956;
            const double E22 = 1.68754095;
            const double E32 = 5.82215341;
            const double E42 = -63.3354786;
            const double E52 = 188.076546;
            const double E62 = -252.344531;
            const double E72 = 128.058531;
            //600<Tsの係数
            const double A3 = 1.0;
            const double B3 = -0.324817650;
            const double C3 = -2.990556709;
            const double D3 = 3.2341900;
            const double E13 = -0.678067859;
            const double E23 = -1.91910364;

            double tk = saturatedTemperature + CONVERT_C_TO_K;
            double tr = getTR(saturatedTemperature);
            double y;
            if (tk < 300.0) y = tr * (E11 + tr * (E21 +
                tr * (E31 + tr * (E41 + tr * (E51 + tr * (E61 + tr * E71))))));
            else if (tk < 600.0) y = tr * (E12 + tr * (E22 + tr *
                (E32 + tr * (E42 + tr * (E52 + tr * (E62 + tr * E72)))))) + A2;
            else y = A3 + B3 * Math.Pow(tr, 1.0 / 3.0) +
                C3 * Math.Pow(tr, 5.0 / 6.0) + D3 * Math.Pow(tr, 0.875) +
                tr * (E13 + tr * E23);

            return y * ENTROPY_AT_CRITICAL_POINT;
        }

        #endregion

        #region 飽和蒸気の物性値

        /// <summary>飽和温度[℃]と飽和圧力[kPa]から飽和蒸気の比体積[m3/kg]を求める</summary>
        /// <param name="saturatedTemperature">飽和温度[℃]</param>
        /// <param name="saturatedPressure">飽和圧力[kPa]</param>
        /// <returns>飽和蒸気の比体積[m3/kg]</returns>
        public static double GetSaturatedVaporSpecificVolume
            (double saturatedTemperature, double saturatedPressure)
        {
            const double A = 1.0d;
            const double B = 1.6351057d;
            const double C = 52.584599d;
            const double D = -44.694653;
            const double E1 = -8.9751114d;
            const double E2 = -0.43845530d;
            const double E3 = -19.179576d;
            const double E4 = 36.765319d;
            const double E5 = -19.462437d;

            double tr = getTR(saturatedTemperature);
            double y = A + B * Math.Pow(tr, 1.0 / 3.0) + 
                C * Math.Pow(tr, 5.0 / 6.0) + D * Math.Pow(tr, 0.875);
            y += tr * (E1 + tr * (E2 + tr * (E3 + tr * (E4 + tr * E5))));

            return y * PRESSURE_AT_CRITICAL_POINT * 
                SPECIFIC_VOLUME_AT_CRITICAL_POINT / saturatedPressure;
        }

        /// <summary>飽和温度[℃]から飽和蒸気のエンタルピー[kJ/kg]を求める</summary>
        /// <param name="saturatedTemperature">飽和温度[℃]</param>
        /// <returns>飽和蒸気のエンタルピー[kJ/kg]</returns>
        public static double GetSaturatedVaporEnthalpy
            (double saturatedTemperature)
        {
            const double E1 = -4.81351884;
            const double E2 = 2.69411792;
            const double E3 = -7.39064542;
            const double E4 = 10.4961689;
            const double E5 = -5.46840036;
            const double A = 1.0;
            const double B = 0.457874342;
            const double C = 5.08441288;
            const double D = -1.48513244;

            double tr = getTR(saturatedTemperature);
            double y = A + B * Math.Pow(tr, 1.0 / 3.0) +
                C * Math.Pow(tr, 5.0 / 6.0) + D * Math.Pow(tr, 0.875);
            y += tr * (E1 + tr * (E2 + tr * (E3 + tr * (E4 + tr * E5))));

            return y * ENTHALPY_AT_CRITICAL_POINT;
        }

        /// <summary>飽和温度[℃]から飽和蒸気のエントロピー[kJ/kgK]を求める</summary>
        /// <param name="saturatedTemperature">飽和温度[℃]</param>
        /// <returns>飽和蒸気のエントロピー[kJ/kgK]</returns>
        public static double GetSaturatedVaporEntropy
            (double saturatedTemperature)
        {
            const double E1 = -4.34839;
            const double E2 = 1.34672;
            const double E3 = 1.75261;
            const double E4 = -6.22295;
            const double E5 = 9.99004;
            const double A = 1.0;
            const double B = 0.377391;
            const double C = -2.78368;
            const double D = 6.93135;

            double tr = getTR(saturatedTemperature);
            double y = A + B * Math.Pow(tr, 1.0 / 3.0) +
                C * Math.Pow(tr, 5.0 / 6.0) + D * Math.Pow(tr, 0.875);
            y += tr * (E1 + tr * (E2 + tr * (E3 + tr * (E4 + tr * E5))));

            return y * ENTROPY_AT_CRITICAL_POINT;
        }

        #endregion

        #region 水の物性値

        /// <summary>温度[C]から水（液体）の密度[kg/m3]を計算する</summary>
        /// <param name="temperature">温度[C]</param>
        /// <returns>水（液体）の密度[kg/m3]</returns>
        public static double GetLiquidDensity
            (double temperature)
        {
            double[] a = new double[] { 9.8811040e2, -1.3273604e3,
                4.7162295e3, -4.1245328e3 };

            double tk = temperature + CONVERT_C_TO_K;
            double tr = (TEMPERATURE_AT_CRITICAL_POINT - tk) /
                TEMPERATURE_AT_CRITICAL_POINT;

            double rho = a[a.Length - 1];
            for (int i = a.Length - 2; 0 <= i; i--)
            {
                rho = a[i] + rho * tr;
            }
            return rho;
        }

        /// <summary>温度[C]から水（液体）の定圧比熱[kJ/(kg·K)]を計算する</summary>
        /// <param name="temperature">温度[C]</param>
        /// <returns>水（液体）の比熱[kJ/(kg·K)]</returns>
        public static double GetLiquidIsobaricSpecificHeat
            (double temperature)
        {
            double[] a = new double[] { 1.0570130, 2.1952960e1,
                -4.9895501e1, 3.6963413e1 };

            double tk = temperature + CONVERT_C_TO_K;
            double tr = (TEMPERATURE_AT_CRITICAL_POINT - tk) /
                TEMPERATURE_AT_CRITICAL_POINT;

            double cpw = a[a.Length - 1];
            for (int i = a.Length - 2; 0 <= i; i--)
            {
                cpw = a[i] + cpw * tr;
            }
            return cpw;
        }

        /// <summary>温度[C]から水（液体）の熱伝導率[W/(m·K)]を計算する</summary>
        /// <param name="temperature">温度[C]</param>
        /// <returns>水（液体）の熱伝導率[W/(m·K)]</returns>
        public static double GetLiquidThermalConductivity
            (double temperature)
        {
            double[] a = new double[] { -1.3734399e-1, 4.2128755,
                -5.9412196, 1.2794890 };

            double tk = temperature + CONVERT_C_TO_K;
            double tr = (TEMPERATURE_AT_CRITICAL_POINT - tk) /
                TEMPERATURE_AT_CRITICAL_POINT;

            double lambda = a[a.Length - 1];
            for (int i = a.Length - 2; 0 <= i; i--)
            {
                lambda = a[i] + lambda * tr;
            }
            return lambda;
        }

        /// <summary>温度[C]から水（液体）の粘性係数[Pa·s]を計算する</summary>
        /// <param name="temperature">温度[C]</param>
        /// <returns>水（液体）の粘性係数[Pa·s]</returns>
        public static double GetLiquidViscosity
            (double temperature)
        {
            double[] a = new double[] { 5.2136906e1, -4.0910405e2, 
                1.3270844e3, -1.9089622e3, 1.0489917e3 };

            double tk = temperature + CONVERT_C_TO_K;
            double tr = (TEMPERATURE_AT_CRITICAL_POINT - tk) /
                TEMPERATURE_AT_CRITICAL_POINT;

            double mu = a[a.Length - 1];
            for (int i = a.Length - 2; 0 <= i; i--)
            {
                mu = a[i] + mu * tr;
            }
            return Math.Exp(mu) * 0.000001;
        }

        /// <summary>温度[C]から水（液体）の動粘性係数[m2/s]を計算する</summary>
        /// <param name="temperature">温度[C]</param>
        /// <returns>水（液体）の動粘性係数[m2/s]</returns>
        public static double GetLiquidDynamicViscosity
            (double temperature)
        {
            double mu = GetLiquidViscosity(temperature);
            double rho = GetLiquidDensity(temperature);
            return mu / rho;
        }

        /// <summary>温度[C]から水（液体）の熱拡散率[m2/s]を計算する</summary>
        /// <param name="temperature">温度[C]</param>
        /// <returns>水（液体）の熱拡散率[m2/s]</returns>
        public static double GetLiquidThermalDiffusivity
            (double temperature)
        {
            double lambda = GetLiquidThermalConductivity(temperature);
            double cp = GetLiquidIsobaricSpecificHeat(temperature);
            double rho = GetLiquidDensity(temperature);
            return lambda / (1000 * cp * rho);
        }

        #endregion

        #region HVACSIM+(J)から移植

        /// <summary>蒸気比体積[m3/kg]を蒸気温度[C]および蒸気圧力[kPa]から計算する</summary>
        /// <param name="pkpa">蒸気圧力[kPa]</param>
        /// <param name="tc">蒸気温度[C]</param>
        /// <returns>蒸気比体積[m3/kg]</returns>
        public static double Vsats(double pkpa, double tc) {
            const double a = 1.0d;
            const double b = 1.6351057d;
            const double c = 52.584599d;
            const double d = -44.694653;
            const double e1 = -8.9751114d;
            const double e2 = -0.43845530d;
            const double e3 = -19.179576d;
            const double e4 = 36.765319d;
            const double e5 = -19.462437d;
            const double tcr = 647.3d;
            const double pcr = 22.089d;
            const double vcr = 3.155e-3d;
            const double tcnv = 23.15d;
            const double pcnv = 0.001d;

            double tr = (tcr - tc - tcnv) / tcr;
            double y = a + b * Math.Pow(tr, 1.0d / 3.0d) + c * Math.Pow(tr, 5.0d / 6.0d) + d * Math.Pow(tr, 0.875d);
            y += tr * (e1 + tr * (e2 + tr * (e3 + tr * (e4 + tr * e5))));
            return y * pcr * vcr / (pkpa * pcnv);
        }

        /// <summary>加熱蒸気の比体積[m3/kg]を圧力[kPa]および温度[C]から求める</summary>
        /// <param name="pkpa">圧力[kPa]</param>
        /// <param name="tc">温度[C]</param>
        /// <returns>加熱蒸気の比体積[m3/kg]</returns>
        public static double Vs(double pkpa, double tc) {
            const double r = 4.61631e-4d;
            const double b1 = 5.27993e-2d;
            const double b2 = 3.75928e-3d;
            const double b3 = 0.022d;
            const double em = 40.0d;
            const double a0 = -3.741378d;
            const double a1 = -4.7838281e-3d;
            const double a2 = 1.5923434e-5d;
            const double tcnv = 273.15d;
            const double a3 = 10.0d;
            const double c1 = 42.6776d;
            const double c2 = -3892.70d;
            const double c3 = -9.48654d;
            const double pcnv = 0.001d;
            const double c4 = -387.592d;
            const double c5 = -12587.5d;
            const double c6 = -15.2578d;

            double p = pkpa * pcnv;
            double t = tc + tcnv;
            double ts = c1 + c2 / (Math.Log(p) + c3);
            if (p >= 12.33d) ts = c4 + c5 / (Math.Log(p) + c6);
            return r * t / p - b1 * Math.Exp(-b2 * t) + (b3 - Math.Exp(a0 + ts * (a1 + ts * a2))) / (a3 * p) * Math.Exp((ts - t) / em);
        }

        /// <summary>加熱蒸気のエンタルピー[kJ/kg]を圧力[kPa]と温度[C]から求める</summary>
        /// <param name="pkpa">圧力[kPa]</param>
        /// <param name="tc">温度[C]</param>
        /// <returns>加熱蒸気のエンタルピー[kJ/kg]</returns>
        public static double Hs(double pkpa, double tc) {
            const double b11 = 2041.21d;
            const double b12 = -40.4002d;
            const double b13 = -0.48095d;
            const double b21 = 1.610693d;
            const double b22 = 5.472051e-2d;
            const double b23 = 7.517537e-4d;
            const double b31 = 3.383117e-4d;
            const double b32 = -1.975736e-5d;
            const double b33 = -2.87409e-7d;
            const double b41 = 1707.82d;
            const double b42 = -16.99419d;
            const double b43 = 6.2746295e-2d;
            const double b44 = -1.0284259e-4d;
            const double b45 = 6.4561298e-8d;
            const double em = 45.0d;
            const double c1 = 42.6776d;
            const double c2 = -3892.70d;
            const double c3 = -9.48654d;
            const double pcnv = 0.001d;
            const double c4 = -387.592d;
            const double c5 = -12587.5d;
            const double c6 = -15.2578d;
            const double tcnv = 273.15d;

            double p = pkpa * pcnv;
            double t = tc + tcnv;
            double ts = c1 + c2 / (Math.Log(p) + c3);
            if (p >= 12.33d) ts = c4 + c5 / (Math.Log(p) + c6);
            double a0 = b11 + p * (b12 + p * b13);
            double a1 = b21 + p * (b22 + p * b23);
            double a2 = b31 + p * (b32 + p * b33);
            double a3 = b41 + ts * (b42 + ts * (b43 + ts * (b44 + ts * b45)));
            return a0 + t * (a1 + t * a2) - a3 * Math.Exp((ts - t) / em);
        }

        /// <summary>加熱蒸気のエントロピー[kJ/kgK]を圧力[kPa]と温度[C]から求める</summary>
        /// <param name="pkpa">圧力[kPa]</param>
        /// <param name="tc">温度[C]</param>
        /// <returns>加熱蒸気のエントロピー[kJ/kgK]</returns>
        public static double Ss(double pkpa, double tc) {
            const double a0 = 4.6162961d;
            const double a1 = 1.039008e-2d;
            const double a2 = -9.873085e-6d;
            const double a3 = 5.4311e-9d;
            const double a4 = -1.170465e-12d;
            const double b1 = -0.4650306d;
            const double b2 = 0.001d;
            const double b3 = 10.0d;
            const double c0 = 1.777804d;
            const double c1 = -1.802468e-2d;
            const double c2 = 6.854459e-5d;
            const double c3 = -1.184434e-7d;
            const double em = 85.0d;
            const double c4 = 8.142201e-11d;
            const double e1 = 42.6776d;
            const double e2 = -3892.70d;
            const double e3 = -9.48654d;
            const double e4 = -387.592d;
            const double e5 = -12587.5d;
            const double e6 = -15.2578d;
            const double tcnv = 273.15d;

            double p = pkpa * b2;
            double t = tc + tcnv;
            double ts = e1 + e2 / (Math.Log(p) + e3);
            if (p >= 12.33d) ts = e4 + e5 / (Math.Log(p) + e6);
            return a0 + t * (a1 + t * (a2 + t * (a3 + t * a4))) + b1 * Math.Log(b2 + p * b3) - Math.Exp((ts - t) / em) * (c0 + ts * (c1 + ts * (c2 + ts * (c3 + ts * c4))));
        }

        /// <summary>加熱蒸気の温度[C]を圧力[kPa]とエントロピー[kJ/kgK]から求める</summary>
        /// <param name="p">圧力[kPa]</param>
        /// <param name="s">エントロピー[kJ/kgK]</param>
        /// <returns>加熱蒸気の温度[C]</returns>
        public static double Tpss(double p, double s) {
            const double e1 = 42.6776d;
            const double e2 = -3892.70d;
            const double e3 = -9.48654d;
            const double pcnv = 0.001d;
            const double e4 = -387.592d;
            const double e5 = -12587.5d;
            const double e6 = -15.2578d;
            const double tabs = 273.15d;

            //compare input entropy with saturation value
            double t0 = e1 - tabs + e2/(Math.Log(p * pcnv) + e3);
            if (p >= 12330.0d) t0 = e4 - tabs + e5 / (Math.Log(p * pcnv) + e6);
            double s0 = GetSaturatedVaporEntropy(t0);
            if (s0 >= s) return t0;

            //Initial guess TA is based on assumption of constant specific heat.
            //Subsequent approximations made by interpolation.
            double ta = (t0 + tabs) * (1.0d + (s - s0) / Cps(t0)) - tabs;
            double sa = Ss(p, ta);
            double t = 0.0d;
            for (int i = 0; i < 10; i++) {
                t = ta + (t0 - ta) * (s - sa) / (s0 - sa);
                if (Math.Abs(t - ta) < 0.05d) break;
                t0 = ta;
                s0 = sa;
                ta = t;
                sa = Ss(p, ta);
                if (i == 9) throw new Exception("FUNCTION WaterProperty.Tpss FAILS TO CONVERGE");
            }
            return t;
        }

        /// <summary>蒸気の比熱[kJ/kg/K]を温度[C]から求める</summary>
        /// <param name="t">温度[C]</param>
        /// <returns>蒸気の比熱[kJ/kg/K]</returns>
        /// <remarks>
        /// Specific heat equation from "Fundamentals of Classical Thermodynamics-SI Version" by Van Wylen and Sonntag Table A.9, pg. 683.
        /// Valid for T between 300-3500 K   max error = .43%
        /// </remarks>
        public static double Cps(double t) {
            const double c1 = 143.05d;
            const double c2 = -183.54d;
            const double c3 = 82.751d;
            const double c4 = -3.6989d;
            const double e1 = 0.25d;
            const double e2 = 0.5d;

            double tk = t + 273.15d;
            if (tk < 300.0d || tk > 3500.0d) throw new Exception("FUNCTION WaterProperty.Cps: T OUT OF RANGE");
            double t1 = tk / 100.0d;
           return (c1 + c2 * Math.Pow(t1, e1) + c3 * Math.Pow(t1, e2) + c4 * t1) / 18.015d;
        }

        /// <summary>比熱[kJ/kg/K]を比体積[m3/kg]と温度[C]から求める</summary>
        /// <param name="v">比体積[m3/kg]</param>
        /// <param name="t">温度[C]</param>
        /// <returns>比熱[kJ/kg/K]</returns>
        public static double Cvs(double v, double t) {
            const double tc = 1165.11d;
            const double tfr = 459.67d;
            const double b1 = 0.0063101d;
            const double a0 = 0.99204818d;
            const double a1 = -33.137211d;
            const double a2 = 416.29663d;
            const double a3 = 0.185053d;
            const double a4 = 5.475d;
            const double a5 = -2590.5815d;
            const double a6 = 113.95968d;

            double tr = 9.0d / 5.0d * t + 32.0d + tfr;
            double ve = (v - b1) / 0.062428d;
            return (a0 + a1 / Math.Sqrt(tr) + a2 / tr - a3 * Math.Pow(a4, 2) * tr / Math.Pow(tc, 2) * Math.Exp(-a4 * tr / tc) * (a5 / ve + a6 / Math.Pow(ve, 2))) * 4.1868;
        }

        /// <summary>飽和水蒸気の動粘性係数[kg/m-s]を圧力[kPa]から求める</summary>
        /// <param name="p">圧力[kPa]</param>
        /// <returns>飽和水蒸気の動粘性係数[kg/m-s]</returns>
        /// <remarks>'Heat Transfer' by Alan J. Chapman, 1974.</remarks>
        public static double Vissv(double p) {
            const double c1 = 0.0314d;
            const double c2 = 2.9675e-5d;
            const double c3 = -1.60583e-8d;
            const double c4 = 3.768986e-12d;

            //Convert pressure from kPa to psi
            double psi = p / 6.894757d;
            double vissv = c1 + c2 * psi + c3 * Math.Pow(psi, 2) + c4 * Math.Pow(psi, 3);
            //Convert viscosity from lbm/ft-hr to kg/m-s
            return vissv * 4.1338e-4d;
        }

        /// <summary>加熱蒸気の動粘性係数[kg/m-s]を温度[C]から求める</summary>
        /// <param name="t">温度[C]</param>
        /// <returns>加熱蒸気の動粘性係数[kg/m-s]</returns>
        /// <remarks>'Heat Transfer' by Alan J. Chapman, 1974. (Note: there is little  variation in viscosity at higher pressures.)</remarks>
        public static double Vissph(double t) {
            const double c1 = 0.0183161d;
            const double c2 = 5.7067e-5d;
            const double c3 = -1.42253e-8d;
            const double c4 = 7.241555e-12d;

            //Convert temperature from C to F
            double tf = t * 1.8 + 32.0d;
            double vissph = c1 + c2 * tf + c3 * Math.Pow(tf, 2) + c4 * Math.Pow(tf, 3);
            //Convert viscosity from lbm/ft-hr to kg/m-s
            return vissph * 4.1338e-4d;
        }

        /// <summary>加熱蒸気の熱伝導率[kW/m-C]を温度[C]から求める</summary>
        /// <param name="t">温度[C]</param>
        /// <returns>加熱蒸気の熱伝導率[kW/m-C]</returns>
        /// <remarks>'Heat Transfer' by Alan J. Chapman, 1974.</remarks>
        public static double Steamk(double t)
        {
            const double c1 = 0.824272d;
            const double c2 = 0.00254627d;
            const double c3 = 9.848539e-8d;

            //Convert temperature from C to F
            double tf = t * 1.8d + 32.0d;
            double steamk = (c1 + c2 * tf + c3 * Math.Pow(tf, 2)) * 0.01d;
            //Convert K from Btu/hr-ft-F to kW/m-C
            return steamk * 0.0017308d;
        }

        /// <summary>1[atm]での水密度[kg/m3]を温度[C]から求める</summary>
        /// <param name="tw">温度[C]</param>
        /// <returns>1[atm]での水密度[kg/m3]</returns>
        /// <remarks>CRC Handbook of Chem. and Phys., 61st Edition (1980-1981), p. F-6.</remarks>
        public static double Wrho(double tw) {
            const double ar0 = 999.83952d;
            const double ar1 = 16.945176d;
            const double ar2 = -0.0079870401d;
            const double ar3 = -46.170461e-6d;
            const double ar4 = 105.56302e-9d;
            const double ar5 = -280.54253e-12d;
            const double ar6 = 0.01687985d;

            return (ar0 + tw * (ar1 + tw * (ar2 + tw * (ar3 + tw * (ar4 + tw * ar5))))) / (1.0d + ar6 * tw);
        }

        /// <summary>1[atm]での水の粘性[kg/m-s]を温度[C]から求める</summary>
        /// <param name="tw">温度[C]</param>
        /// <returns>1[atm]での水の粘性[kg/m-s]</returns>
        /// <remarks>Fit to data from Karlekar and Desmond</remarks>
        public static double Wmu(double tw) {
            const double am0 = -3.30233d;
            const double am1 = 1301.0d;
            const double am2 = 998.333d;
            const double am3 = 8.1855d;
            const double am4 = 0.00585d;
            const double am5 = 1.002d;
            const double am6 = -1.3272d;
            const double am7 = -0.001053d;
            const double am8 = 105.0d;
            const double am10 = 0.68714d;
            const double am11 = -0.0059231d;
            const double am12 = 2.1249e-5d;
            const double am13 = -2.69575e-8d;

            double wmu;
            if (tw < 20.0d) wmu = Math.Pow(10.0d, am0 + am1 / (am2 + (tw - 20.0d) * (am3 + am4 * (tw - 20.0d)))) * 100.0d;
            else if (tw > 100.0d) wmu = am10 + tw * (am11 + tw * (am12 + tw * am13));
            else wmu = am5 * Math.Pow(10.0d, (tw - 20.0d) * (am6 + (tw - 20.0d) * am7) / (tw + am8));
            return 0.001 * wmu;
        }

        /// <summary>1[atm]での蒸気の熱伝導率[kW/m-K]を温度[C]から求める</summary>
        /// <param name="tw">温度[C]</param>
        /// <returns>1[atm]での蒸気の熱伝導率[kW/m-K]</returns>
        public static double Wk(double tw) {
            const double ak0 = 0.560101d;
            const double ak1 = 0.00211703d;
            const double ak2 = -1.05172e-5d;
            const double ak3 = 1.497323e-8d;
            const double ak4 = -1.48553e-11d;

            return 0.001d * (ak0 + tw * (ak1 + tw * (ak2 + tw * (ak3 + tw * ak4))));
        }

        /// <summary>1[atm]での空気の比熱[kJ/kg-C]を温度[C]から求める</summary>
        /// <param name="tw">温度[C]</param>
        /// <returns>1[atm]での空気の比熱[kJ/kg-C]</returns>
        public static double Wcp(double tw) {
            const double acp0 = 4.21534d;
            const double acp1 = -0.00287819d;
            const double acp2 = 7.4729e-5d;
            const double acp3 = -7.79624e-7d;
            const double acp4 = 3.220424e-9d;
            const double acp5 = 2.9735d;
            const double acp6 = 0.023049d;
            const double acp7 = -0.00013953d;
            const double acp8 = 3.092474e-7d;

            if (tw > 100.0d) return acp5 + tw * (acp6 + tw * (acp7 + tw * acp8));
            else return acp0 + tw * (acp1 + tw * (acp2 + tw * (acp3 + tw * acp4)));
        }

        #endregion

    }
}
